;----------------------------------------------------------
; CHIPS.INC
; PROGETTO: B1601
;
; CHIPS I/O
;----------------------------------------------------------

; il file puo' essere incluso solo all'interno di GLOBAL.INC
.IFNDEF		_GLOBAL_INC_
.EXIT		"ERROR: This file cannot be included."
.ENDIF

;----------------------------------------------------------

; registro di controllo
CRBIT0		.EQU	$00FC00
CR0OFF		.EQU	$00FC00
CR0ON		.EQU	$00FC01
CRBIT1		.EQU	$00FC02
CR1OFF		.EQU	$00FC02
CR1ON		.EQU	$00FC03
CRBIT2		.EQU	$00FC04
CR2OFF		.EQU	$00FC04
CR2ON		.EQU	$00FC05
CRFWE		.EQU	$00FC06
CRFWEOFF	.EQU	$00FC06
CRFWEON		.EQU	$00FC07
CRXFE		.EQU	$00FC08
CRXFEOFF	.EQU	$00FC08
CRXFEON		.EQU	$00FC09
CRXME		.EQU	$00FC0A
CRXMEOFF	.EQU	$00FC0A
CRXMEON		.EQU	$00FC0B
CREME		.EQU	$00FC0C
CREMEOFF	.EQU	$00FC0C
CREMEON		.EQU	$00FC0D
CRHIM		.EQU	$00FC0E
CRHIMOFF	.EQU	$00FC0E
CRHIMON		.EQU	$00FC0F


; registri VIA 65C22
VIA0		.EQU	$00FD00		; IRQ 2
VIA1		.EQU	$00FD10		; NMI

; bit I/O VIA 0 - IRQ 2
; PA0 -> XM16 - linea A16 memoria DMA con DMA attivo
; PA1 -> /DME - enable DMA
; PA2 -> /DMR - reset DMA (impulso negativo)
; PA3 -> /FDR - reset FDC (impulso negativo)
; PA4 -> /XTC - impulso TC DMA/FDC (impulso negativo)
; PA5 -> /FFSR - set flip flop reset (impulso negativo) - vedere PIA0-PB4
; PA6 -> RSTRM - enable terminazione RS485 120 ohm su linee RX/TX
; PA7 -> /CLRX - reset macchina (impulso negativo)
; PB0 -> /WD - enable wait chip DMA
; PB1 -> /WF - enable wait chip FDC
; PB2 -> /WS - enable wait chip ACIA
; PB3 -> /WA0 - enable wait slot ATA 0
; PB4 -> /WA1 - enable wait slot ATA 1
; PB5 -> /WV - enable wait chip VDC
; PB6 <- onda quadra 1KHZ per timer T2
; PB7 -> /WT - enable wait chip CTC
; CA1 <- TMF1 (da CTC) fronte negativo - vedere anche PIA0-PB6
; CA2 <- INT da controller tastiera - fronte negativo
; CB1 <- INT da controller FDC - fronte positivo - vedere anche PIA0-PB7
; CB2 <- TMF0 (da CTC) fronte negativo - vedere anche PIA0-PB5

; bit I/O VIA 1 - NMI
; PA0 -> /VDE - enable output RGB da chip VDC
; PA1 -> /S16M - enable dot clock VDC a 16MHZ (CGA)
; PA2 -> HPOL - HSYNC VDC positivo
; PA3 -> VPOL - VSYNC VDC positivo
; PA4 -> /KBRES - reset controller tastiera (impulso negativo)
; PA5 -> PCTS1 - inversione livello /CTS1 su ACIA
; PA6 -> PCTS2 - inversione livello /CTS2 su ACIA
; PA7 -> /RS485 - enable RS485 su canale 2 ACIA
; PB0 -> /KFRW0 - enable write file register KFR sincronizzato con PHI0
; PB1 -> DMA clock select: 0->4MHz - 1->5MHz
; PB2 -> /VDCSX - abilita impulso di selezione su /CS0 o CS1 VDC
;		  altrimenti /CS0 a livello 0 oppure CS1 a livello 1
; PB3 -> /VDCS0 - sincronizza write registri VDC con PHI0 invece di PHI2
; PB4 <- TRIGGER POWER - impulso negativo da pulsante ON/OFF
; PB5 -> /MW0 - enable write RAM/FLASH sincronizzato con PHI0
; PB6 <- onda quadra 1KHZ per timer T2
; PB7 -> onda quadra 1KHZ da timer T1
; CA1 <- impulso ATA 0 INT - fronte positivo (non genera NMI)
; CA2 <- impulso NMI da controller tastiera - fronte negativo
; CB1 <- impulso ATA 1 INT - fronte positivo (non genera NMI)
; CB2 <- TRIGGER POWER - impulso negativo da pulsante ON/OFF

VIAPRB		.EQU	$0
VIAPRA		.EQU	$1
VIADDRB		.EQU	$2
VIADDRA		.EQU	$3
VIAT1CL		.EQU	$4
VIAT1CH		.EQU	$5
VIAT1LL		.EQU	$6
VIAT1LH		.EQU	$7
VIAT2CL		.EQU	$8
VIAT2CH		.EQU	$9
VIASR		.EQU	$A
VIAACR		.EQU	$B
VIAPCR		.EQU	$C
VIAIFR		.EQU	$D
VIAIER		.EQU	$E
VIAPRANH	.EQU	$F

; bit IER/IFR
SETFRB		.EQU	10000000B
T1IFRB		.EQU	01000000B
T2IFRB		.EQU	00100000B
CB1IFRB		.EQU	00010000B
CB2IFRB		.EQU	00001000B
SRIFRB		.EQU	00000100B
CA1IFRB		.EQU	00000010B
CA2IFRB		.EQU	00000001B

;----------------------------------------------------------

.IFDEF		_ATA_INC_

; register address for ata #0
ATADATAL	.EQU	$FD20		; data register low
ATAERROR	.EQU	$FD21		; error reg. READ
ATAFEAT		.EQU	$FD21		; feature reg. WRITE
ATASCTCNT	.EQU	$FD22		; sector count
ATALBAL		.EQU	$FD23		; LBA LOW
ATALBAM		.EQU	$FD24		; LBA MID
ATALBAH		.EQU	$FD25		; LBA HIGH
ATADEV		.EQU	$FD26		; device reg.
ATAST		.EQU	$FD27		; status reg. READ
ATACMD		.EQU	$FD27		; command reg. WRITE
ATADATAH	.EQU	$FD2C		; data register high
ATAALTST	.EQU	$FD2E		; alt. status reg. READ
ATACTRL		.EQU	$FD2E		; device control reg. WRITE

ATAOFS		.EQU	$10		; offset ata #1 address

; bit registro di stato
ATABSY		.EQU	10000000B	; busy bit
ATADRDY		.EQU	01000000B	; ready bit
ATADFB		.EQU	00100000B	; device fault bit
ATADRQ		.EQU	00001000B	; data request bit
ATAERRB		.EQU	00000001B	; error bit

; bit registro di errore
ATA_ABRT	.EQU	00000100B	; flag ABORT

; bit registro $0E - device control (ATACTRL)
CTRLSRST	.EQU	00000100B	; software reset (ATACTRL)

.ENDIF

;----------------------------------------------------------

; registri PIA, PIA 1
PIA0		.EQU	$00FD28
PIA1		.EQU	$00FD38		; LCD

; bit I/O PIA 0
; PA0-PA7 -> MA13-MA20 linee indirizzo memoria estesa (256 pagine di 8K)
; PB0-PB3 -> SA13-SA16 linee indirizzo buffer DMA per macchina 65C02
; PB4 <- /FSR livello 0 al reset (Flip Flop)
; PB5 <- TMF0 (da CTC) fronte negativo
; PB6 <- TMF1 (da CTC) fronte negativo
; PB7 <- INT da FDC fronte negativo
; CA2 -> /ENQ0 - enable buzzer
; CB2 -> /SMODE - enable mode one shot per buzzer

; bit I/O PIA 1 - controller LCD
; PA0 -> RS
; PA1 -> R/W
; PA2 -> E
; PB0-PB7 -> DB0 - DB7

PIAPRA		.EQU	$0
PIADDRA		.EQU	$0
PIACRA		.EQU	$1
PIAPRB		.EQU	$2
PIADDRB		.EQU	$2
PIACRB		.EQU	$3

; NOTA: inizializzare CRA, CRB -> 00000000
; Accesso a porta:    CRA, CRB -> 00000100 

;----------------------------------------------------------

.IFDEF _KFR_INC_
; registri controller tastiera
KBFR		.EQU	$00FD40		; KB file register
KBSTATUS	.EQU	$00FD44		; KB status - read
KBDRQ		.EQU	$00FD44		; write - richiesta servizio
KBCLRNMI	.EQU	$00FD45		; write - azzera flag NMI
KBCLRIRQ	.EQU	$00FD46		; write - azzera flag INT

; bit di KBSTATUS
; 7: DRQ attivo - controller busy
; 6: RA5 da controller
; 5: flag NMI da controller
; 4: flag IRQ da controller
; 3, 2 e 1: 0
; 0; RA4 da controller

.ENDIF

;----------------------------------------------------------

IRQVECTR	.EQU	$00FD47		; registro vettore IRQ (sola lettura)

;----------------------------------------------------------

; registri CTC 82C54
CTC0		.EQU	$00FD48

CTCCNT0		.EQU	0
CTCCNT1		.EQU	1
CTCCNT2		.EQU	2
CTCCTRL		.EQU	3

;----------------------------------------------------------

.IFDEF		_RTC_INC_
; registri RTC DS1687
RTCALE		.EQU	$00FD4C		; registro address (senza /CS)
RTCDATA		.EQU	$00FD4D		; registro data

RTCCTRLA	.EQU	$0A		; registro controllo A
RTCEXTCTRLA	.EQU	$4A		; registro controllo esteso 4A (banco 1)
RTCCTRLB	.EQU	$0B		; registro controllo B
RTCEXTCTRLB	.EQU	$4B		; registro controllo esteso 4B (banco 1)
RTCSTATUS	.EQU	$0C		; registro di stato principale
RTCSTATUS2	.EQU	$0D		; registro di stato secondario

RTCSECS		.EQU	$00		; registro secondi
RTCMIN		.EQU	$02		; registro minuti
RTCHOURS	.EQU	$04		; registro ore
RTCDAY		.EQU	$06		; registro giorno
RTCDDATE	.EQU	$07		; registro data
RTCMONTH	.EQU	$08		; registro mese
RTCYEAR		.EQU	$09		; registro anno
RTCCENTURY	.EQU	$48		; registro secolo (banco 1)

RTCALMSECS	.EQU	$01		; registro allarme secondi
RTCALMMIN	.EQU	$03		; registro allarme minuti
RTCALMHOURS	.EQU	$05		; registro allarme ore
RTCALMDATE	.EQU	$49		; registro allarme data (banco 1)

RTCRAM0E	.EQU	$0E		; registro RAM $0E
RTCRAM0F	.EQU	$0F		; registro RAM $0F

RTCURAM		.EQU	$10		; user RAM da $10 a $3F (48 bytes)
RTCURAMLAST	.EQU	$3F

; saved configuration in RTC ram bank 0 & 1
RTCSHID		.EQU	$3B		; default hidden sector's (for fdisk)
RTCHUSE		.EQU	$3C		; flag use hash in fat driver
RTCHLOG2	.EQU	$3D		; hash hlog2 value (how much sector's)

RTCRDFG		.EQU	$3E		; flag ram disk valid configuration ($55)
					; new device or after clear mem = $FF
RTCRD		.EQU	$3F		; saved configuration for ram disk

RTCURAM0	.EQU	$40		; user RAM da $40 a $7F (64 bytes)
RTCURAM0LAST	.EQU	$7F		; solo banco 0

RTCEXTRAMADDR	.EQU	$50		; registro address EXT RAM (128 bytes) banco 1
RTCEXTRAMDATA	.EQU	$53		; registro data EXT RAM (128 bytes) banco 1

RTCSMI2		.EQU	$4E		; SMI recovery stack ADDR - 2
RTCSMI3		.EQU	$4F		; SMI recovery stack ADDR - 3

RTCTYPE		.EQU	$40		; RTC MODEL ($47 per DS1687) - banco 1
RTCSERIAL	.EQU	$41		; S/N RTC - 6 bytes da $41 a $46 - banco 1
RTCCRC		.EQU	$47		; CRC del serial number

; bit di RTCFlag
RTC_TVAL	.EQU	10000000B	; bit time valido
RTC_VBAT	.EQU	01000000B
RTC_VBAUX	.EQU	00100000B

.ENDIF

;----------------------------------------------------------

; registri MOS8563
VDCAddr		.EQU	$00FD4E
VDCStatus	.EQU	$00FD4E
VDCData		.EQU	$00FD4F

;----------------------------------------------------------

.IFDEF		_ACIA_INC_
; registri ACIA R65C52
ACIA		.EQU	$00FD50

ACIAIER		.EQU	$FD50		; INT enable reg. 1 (write)
ACIAISR		.EQU	$FD50		; INT status reg. 1 (read)
ACIACR		.EQU	$FD51		; Control register 1 (write)
ACIACSR		.EQU	$FD51		; Control status reg. 1 (read)
ACIAFR		.EQU	$FD51		; Format register 1 (write)
ACIACDR		.EQU	$FD52		; Compare data reg. 1 (write)
ACIAACR		.EQU	$FD52		; Aux. Control reg. 1 (write)
ACIATDR		.EQU	$FD53		; TX Data reg. (write)
ACIARDR		.EQU	$FD53		; RX Data reg. (read)

ACIAIER1	.EQU	0		; INT enable reg. 1 (write)
ACIAISR1	.EQU	0		; INT status reg. 1 (read)
ACIACR1		.EQU	1		; Control register 1 (write)
ACIAFR1		.EQU	1		; Format register 1 (write)
ACIACSR1	.EQU	1		; Control status reg. 1 (read)
ACIACDR1	.EQU	2		; Compare data reg. 1 (write)
ACIAACR1	.EQU	2		; Aux. Control reg. 1 (write)
ACIATDR1	.EQU	3		; TX Data reg. 1 (write)
ACIARDR1	.EQU	3		; RX Data reg. 1 (read)
ACIAIER2	.EQU	4		; INT enable reg. 2 (write)
ACIAISR2	.EQU	4		; INT status reg. 2 (read)
ACIACR2		.EQU	5		; Control register 2 (write)
ACIAFR2		.EQU	5		; Format register 2 (write)
ACIACSR2	.EQU	5		; Control status reg. 2 (read)
ACIACDR2	.EQU	6		; Compare data reg. 2 (write)
ACIAACR2	.EQU	6		; Aux. Control reg. 2 (write)
ACIATDR2	.EQU	7		; TX Data reg. 2 (write)
ACIARDR2	.EQU	7		; RX Data reg. 2 (read)

.ENDIF

;----------------------------------------------------------

.IFDEF		_DMA_INC_

; registri UM8388
FDCC		.EQU	$00FD58

; NOTA - INT azzerato con una operazione IO sul CHIP
FDCDOR		.EQU	$00FD5A 	; (solo scrittura)
FDCMSR		.EQU	$00FD5C 	; (solo lettura)
FDCDATA		.EQU	$00FD5D
FDCDIR		.EQU	$00FD5F 	; (solo lettura)
FDCCCR		.EQU	$00FD5F 	; (solo scrittura)

; registri 82C37
DMAC		.EQU	$00FD60
DMAADDR0	.EQU	0
DMACNT0		.EQU	1
DMAADDR1	.EQU	2
DMACNT1		.EQU	3
DMAADDR2	.EQU	4
DMACNT2		.EQU	5
DMAADDR3	.EQU	6
DMACNT3		.EQU	7
DMASR		.EQU	$00FD68		; read STATUS REG
DMAWCMD		.EQU	$00FD68		; write COMMAND REG.
DMARQR		.EQU	$00FD69		; read/write REQUEST REG
DMAWMSKB	.EQU	$00FD6A		; write MASK BIT REG
DMARCMD		.EQU	$00FD6A		; read COMMAND REG.
DMAMODE		.EQU	$00FD6B		; read/write MODE REG.
DMARSETFF	.EQU	$00FD6C		; read - set FF (no data)
DMAWCLRFF	.EQU	$00FD6C		; write  - clear FF (no data)
DMARTMP		.EQU	$00FD6D		; read - read temp. register
DMAWMCLR	.EQU	$00FD6D		; write - MASTER CLEAR (no data)
DMARCNT		.EQU	$00FD6E		; read - clr mode reg. cnt. (no data)
DMAWMASKOFF	.EQU	$00FD6E		; write - clr mask register (no data)
DMAMASK		.EQU	$00FD6F		; read/write - mask register

.ENDIF

;----------------------------------------------------------

.IFDEF		_VBB_INC_

; registri R6545 / HD6445
CRTAddr		.EQU	$00FC70
CRTStatus	.EQU	$00FC70
CRTData		.EQU	$00FC71

PIAVBB		.EQU	$00FC74		; PIA VIDEO BOARD

URLVBB		.EQU	$00FC72		; registro underline write only 4 bit

.ENDIF
