;----------------------------------------------------------
; FAT.INC
; PROGETTO: B1601
;
; Parametri BIOS/ATA
;----------------------------------------------------------

; il file puo' essere incluso solo all'interno di GLOBAL.INC
.IFNDEF		_GLOBAL_INC_
.EXIT		"ERROR: This file cannot be included."
.ENDIF

.IFDEF	_FAT_INC_

	.LIST on

;---------------------------------------------------------------------------	
; struttura blocco parametri comune file system FAT12/16/32
	STRUCT	FATCM
VOLFlagL	.DB		; <7> = 1 -> BS BACKUP FAT32 MISMATCH
VOLFlagH	.DB		; <7> = 0 => FD / = 1 => HD
				; <6> = 0 => FAT16/FAT12  720Kb
				; <6> = 1 => FAT32/FAT12 1440Kb				
FSecPerClus	.DB		; settori per cluster (1,2,4,8,10,20,40)
FClusShift	.DB		; shift cluster (0,1,2,3,4,5,6)
FAT1Size	.DW		; dimensione una FAT in settori
; i 2 campi seguenti sono in comune tra FAT12/16 e FAT32
FRDSec		.DW		; # numero settori FAT12/16 root directory
FRDEnt		.DW		; # numero entry FAT12/16 root directory
F32RDCl		.EQU	FRDSec	; first cluster FAT32 root directory (32 bit)
	ESTRUCT	FATCM

;---------------------------------------------------------------------------	
; struttura blocco parametri file system FAT16/32
	STRUCT	FATD
		.DS	FATCMSIZE
FAT1Lba		.LWORD		; ABS. LBA FAT #1
FAT2Lba		.LWORD		; ABS. LBA FAT #2
FRootDir	.LWORD		; ABS. LBA Root Directory
FDataSec	.LWORD		; ABS. LBA primo settore data (cluster #2)
FClCnt		.LWORD		; # numero cluster volume FAT
FSIFreeCnt	.LWORD		; conteggio cluster liberi FAT32
FSINextFree	.LWORD		; prossimo cluster libero FAT32
	ESTRUCT	FATD


;---------------------------------------------------------------------------	
; struttura blocco parametri file system FAT12 floppy disk
	STRUCT	FAT12D
		.DS	FATCMSIZE
FDFAT1Lba	.DW		; ABS. LBA FAT #1
FDFAT2Lba	.DW		; ABS. LBA FAT #2
FDRootDir	.DW		; ABS. LBA Root Directory
FDDataSec	.DW		; ABS. LBA primo settore data (cluster #2)
FDClCnt		.DW		; # numero cluster volume FAT12
FDFreeCnt	.DW		; conteggio cluster liberi FAT12
FDNextFree	.DW		; prossimo cluster libero FAT12
	ESTRUCT	FAT12D

;---------------------------------------------------------------------------	
; struttura blocco parametri estesi volumi FAT

	STRUCT	XFATD
FATVolID	.LWORD		; ID 32 bit volume FAT
FATMedia	.DB		; media (come da boot sector)
FATDirty	.DB		; dirty flag	<7> = 1 -> volume clean
				;		<6> = 1 -> no I/O error 
BSOEMName	.DS	9	; OEM name
FATVolName	.DS	12	; nome volume
BSSysType	.DS	9	; stringa sistema
	ESTRUCT	XFATD
	
;---------------------------------------------------------------------------	
; struttura blocco parametri estesi volumi FAT32

	STRUCT	XFAT32
		.DS	XFATDSIZE
FSInfoLba	.LWORD		; ABS. LBA FAT32 FSINFO
BSBackLba	.LWORD		; ABS. LBA FAT32 backup boot sector
	ESTRUCT	XFAT32
	
;---------------------------------------------------------------------------

; Pagina 03 => strutture XFATD e XFAT32 per dati estesi volumi FAT
DIRP03:	.SECTION page0, ref_only, common	;Direct-Page 03

	.ABSOLUTE		;; inizia sempre da $00
	.ORG		0x00

XF12Dat0	.DS	XFATDSIZE	; FDD 0 => A
XF12Dat1	.DS	XFATDSIZE	; FDD 1 => B

XFATDat0	.DS	XFAT32SIZE	; HDD 0 => C
XFATDat1	.DS	XFAT32SIZE	; HDD 1 => D
XFATDat2	.DS	XFAT32SIZE	; HDD 2 => E
XFATDat3	.DS	XFAT32SIZE	; HDD 3 => F

; variabili ausiliarie utilizzate durante il test boot sector
DP03Ptr		LP
DP03Len		.DB
DP03SaveX	.DB
DP03xfat	.DB
DP03w		.DW

	.RELATIVE

	.ENDS

;---------------------------------------------------------------------------

DIRP04:	.SECTION page0, ref_only, common	;Direct-Page 04

	.ABSOLUTE		;; inizia sempre da $00
	.ORG		0x00

F12Dat0		.DS	FAT12DSIZE	; FDD 0 => A
F12Dat1		.DS	FAT12DSIZE	; FDD 1 => B

FATDat0		.DS	FATDSIZE	; HDD 0 => C
FATDat1		.DS	FATDSIZE	; HDD 1 => D
FATDat2		.DS	FATDSIZE	; HDD 2 => E
FATDat3		.DS	FATDSIZE	; HDD 3 => F

DEVLPtr		LP
FATCurPtr	.DB		; puntatore FAT corrente
FSFatPtr	.DS	6	; puntatori strutture FATDat
FSXFatPtr	.DS	6	; puntatori strutture XFATDat
FATDDTmp	DD
FATDDTmp1	DD
FATDDTmp2	DD
FATWTmp		.DW
FATWTmp1	.DW
FATWTmp2	.DW
FATWTmp3	.DW
FATWTmp4	.DW
FATWTmp5	.DW
FATWTmp6	.DW
FATWTmp7	.DW
FATWTmp8	.DW
FATWTmp9	.DW

ATACurDev	.DB

	.RELATIVE

	.ENDS

;---------------------------------------------------------------------------

DIRP05:	.SECTION page0, ref_only, common	;Direct-Page 05

	.ABSOLUTE		;; inizia sempre da $00
	.ORG		0x00

storfg		.DB
wildfg		.DB
dotfg		.DB
maxcnt		.DB
cnt		.DB
cnt1		.DB
cntx		.DB
LPtr1		LP
LPtr2		LP
haswild		.DB
FDirIdx		.DB
FPathStr	.DS	2
FDirStr		.DS	78
CDLen		.DS	6
FDrive		.DB
FDefDrv		.DB
lstidx		.DB
FEndDir		.DB
FFileScn	.DB
devfg		.DB
devidx		.DB
tmpfcb		.DS	16
LPtr3		LP

	.RELATIVE

	.ENDS

;---------------------------------------------------------------------------

	.LIST off		
	
.ENDIF
