;;
;; Copyright (c) 2013 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; name: bios.asm	
;; rev.: 2016/03/03
;; bios 65C816 version v 1.0

;; bios sub-system interface

_RTC_INC_	.SET	1

	.INCLUDE	inc\global.inc
	.INCLUDE 	inc\dirp00.inc
	.INCLUDE 	inc\dirp01.inc

.LIST on
	
	.CODEF8

	.EXTERN _ScnCharOut, _ScnCharIn, cinit, LF8Tst

	.LONGA	off
	.LONGI	off

;---------------------------------------------------------------------------
; system bios call interface
;---------------------------------------------------------------------------
	
;---------------------------------------------------------------------------

; tabella funzioni text char output
TXTOTBL:
	.DW	_KeybOut, _SerOut, _SerOut, _ScnOut, _PrnOut, _PrnOut

; tabella funzioni text char input
TXTITBL:
	.DW	_KeybIn, _SerIn, _SerIn, _ScnIn, _PrnIn, _PrnIn

; tabella funzioni text string output
TXTSTBL:
	.DW	_KeybOut, _SerStrOut, _SerStrOut, $0000
	.DW	_PrnStrOut, _PrnStrOut


;---------------------------------------------------------------------------
;---------------------------------------------------------------------------

; char in da schermo (rediretto)	
_ScnIn:
	.LONGA	off
	.LONGI	off
	
	;lda	DflTxtIn
	lda	#TXTD_SCN
	pea	#DP01ADDR	; imposta nuova DP
	pld
	ora	ScnCR		; input rediretto da video
	sta	ScnCR
	lda	ScnRow
	sta	ScnLinTmp
	lda	ScnRight
	sta	ScnInput
	jsr	_ScnCharIn
	sta	STKCR,s	
	clc
	rtl
	
; char in da porta seriale
_SerIn:
	sec
	rtl

; char out su porta seriale
_SerOut:
	.EXTERN _spput0

	ldx	DflTxtOut
	dex
?02:	jsr	_spput0
	bcc	?10
	iny
	bne	?02
?10:	rtl

; string out su porta seriale
_SerStrOut:
	ldx	DflTxtOut
	dex
?02:	lda	[COPPtr]
	beq	?20
?05:	jsr	_spput0
	bcc	?10
	iny
	bne	?05
	rtl
?10:	dec	coptmp
	beq	?20
	inc	COPPtr
	bne	?02
	inc	COPPtr+1
	bra	?02
?20:	rtl

; char in da porta LPT - errore permanente 
_PrnIn:
	CPU08
	sec
	rtl

; char out su porta LPT
_PrnOut:
	CPU08
	sec
	rtl

; string out su porta LPT
_PrnStrOut:
	CPU08
	sec
	rtl

; char out su tastiera - errore permanente
_KeybOut:
	CPU08
	sec
	rtl
	
COP_ScnPrint:
	.GLOBAL COP_ScnPrint
	LONG_OFF
	
	ldy	#0		; indice stringa
	CPU16	
	lda	#DP01ADDR	; imposta nuova DP
	tcd
	lda	STKPCL,s	; ricava ptr stringa
	sta	ScnPrm
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; banco stringa
	sta	ScnPrm+2	
	stz	ScnFlag
	lda	#$40
	trb	ScnAutoIns	; modo output
?01:	lda	[ScnPrm],y	; loop
	beq	?02		; fine stringa
	phy
	INDEX08
	jsr	_ScnCharOut
	INDEX16
	ply
	iny
	bne	?01
?02:	CPU16
	iny
	tya			; ptr terminatore stringa
	clc
	adc	ScnPrm
	sta	STKPCL,s	; update indirizzo di ritorno
RTL1:	
	CPU08
	clc
	rtl

COP_TxtPrint:
	.GLOBAL COP_TxtPrint
	LONG_OFF
	
	lda	DflTxtOut
	cmp	#TXTD_SCN
	beq	COP_ScnPrint
	asl	a
	tax
	ldy	#0		; indice stringa
	ACC16	
	lda	STKPCL,s	; ricava ptr stringa
	sta	COPPtr
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; banco stringa
	sta	COPPtr+2
?01:	lda	[COPPtr],y	; loop
	beq	?02		; fine stringa
	iny
	bne	?01
?02:	CPU16
	iny			; skip final NUL
	tya			; ptr terminatore stringa
	clc
	adc	COPPtr
	sta	STKPCL,s	; update indirizzo di ritorno
	CPU08
	dey			; string length
	beq	RTL1		; exit if empty
	sty	coptmp
	jmp	(TXTSTBL,x)
	
	
; output stringa a video
;In	- A offset stringa (16 bit)
;	  X banco stringa (8 bit)
; registri tutti preservati
COP_ScnStrOut:
	.GLOBAL COP_ScnStrOut
	LONG_OFF

	ldy	#0
	CPU16	
	lda	#DP01ADDR	; imposta nuova DP
	tcd
	lda	STKCR,s		; ricava ptr stringa
	sta	ScnPrm
	ACC08			; A,M -> 8 bit
	lda	STKXR,s		; banco stringa
	sta	ScnPrm+2	
	stz	ScnFlag
	lda	#$40
	trb	ScnAutoIns	; modo output
?01:	lda	[ScnPrm],y	; loop
	beq	?02		; fine stringa
	phy
	INDEX08
	jsr	_ScnCharOut
	INDEX16
	ply
	iny
	bne	?01
?02:	CPU08
	clc
	rtl

COP_TxtStrOut:
	.GLOBAL COP_TxtStrOut
	LONG_OFF
	
	lda	DflTxtOut
	cmp	#TXTD_SCN
	beq	COP_ScnStrOut
	asl	a
	tax
	ldy	#0		; indice stringa
	ACC16	
	lda	STKPCL,s	; ricava ptr stringa
	sta	COPPtr
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; banco stringa
	sta	COPPtr+2
?01:	lda	[COPPtr],y	; loop
	beq	?02		; fine stringa
	iny
	bne	?01
?02:	iny			; skip final NUL
	dey			; string length
	beq	RTL2		; exit if empty
	sty	coptmp
	jmp	(TXTSTBL,x)
	
; Output carattere a schermo
COP_ScnCharOut:
	.GLOBAL COP_ScnCharOut
	LONG_OFF

	pea	#DP01ADDR	; imposta nuova DP
	pld
	lda	#$40
	trb	ScnAutoIns	; modo output
	lda	STKCR,s
COP?01:
	stz	ScnFlag
	jsr	_ScnCharOut
RTL2:
	clc			; no error
	rtl

_ScnOut:
	pea	#DP01ADDR	; imposta nuova DP
	pld
	tax
	lda	#$40
	trb	ScnAutoIns	; modo output
	txa
	bra	COP?01
	
COP_ScnPrChar:
	.GLOBAL COP_ScnPrChar
	LONG_OFF
	
	ACC16	
	lda	#DP01ADDR	; imposta nuova DP
	tcd
	lda	STKPCL,s	; ricava ptr stringa
	sta	ScnPrm
	inc	a		; update return address
	sta	STKPCL,s	
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; banco stringa
	sta	ScnPrm+2	
	lda	#$40
	trb	ScnAutoIns	; modo output
	lda	[ScnPrm]	; char to print
	bra	COP?01

COP_TxtCharOut:
	.GLOBAL COP_TxtCharOut
	LONG_OFF
	
	lda	DflTxtOut
	asl	a
	tax
	lda	STKCR,s
	jmp	(TXTOTBL,x)

COP_DevCharOut:
	.GLOBAL COP_DevCharOut
	LONG_OFF
	
	lda	STKCR+1,s	; # device
	tay
	asl	a
	tax
	lda	STKCR,s		; character
	jmp	(TXTOTBL,x)

COP_TxtPrChar:
	.GLOBAL COP_TxtPrChar
	LONG_OFF
	
	ACC16	
	lda	STKPCL,s	; ricava ptr stringa
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s	
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; banco stringa
	sta	COPPtr+2
	lda	DflTxtOut
	asl	a
	tax	
	lda	[COPPtr]	; char to print
	jmp	(TXTOTBL,x)

COP_ScnPasStr:
	.GLOBAL COP_ScnPasStr
	LONG_OFF

	CPU16	
	lda	#DP01ADDR	; imposta nuova DP
	tcd
	lda	STKCR,s		; ricava ptr stringa
	sta	ScnPrm
	CPU08			; A,M -> 8 bit
	lda	STKXR,s		; banco stringa
	sta	ScnPrm+2	
	lda	#0
	sta	ScnFlag
	xba			; B = 0
	lda	#$40
	trb	ScnAutoIns	; modo output
	lda	[ScnPrm]	; str len
	ldy	#1		; first index
	CPU16
	
_ScnBufOut:
	tax			; buf len
	beq	?02
	ACC08
?01:	lda	[ScnPrm],y	; loop
	phx
	phy
	INDEX08
	jsr	_ScnCharOut
	INDEX16
	ply
	plx
	iny
	dex
	bne	?01
?02:	CPU08
	clc
	rtl

COP_TxtPasStr:
	.GLOBAL COP_TxtPasStr
	LONG_OFF

	lda	DflTxtOut
	cmp	#TXTD_SCN
	beq	COP_ScnPasStr
	sec
	rtl

COP_ScnBufOut:
	.GLOBAL COP_ScnBufOut
	LONG_OFF

	ldy	#0		; first index
	CPU16	
	lda	#DP01ADDR	; imposta nuova DP
	tcd
	lda	STKCR,s		; ricava ptr stringa
	sta	ScnPrm
	CPU08			; A,M -> 8 bit
	lda	STKXR,s		; banco stringa
	sta	ScnPrm+2	
	stz	ScnFlag
	lda	#$40
	trb	ScnAutoIns	; modo output
	CPU16	
	lda	STKYR,s		; buf len
	bra	_ScnBufOut
	LONG_OFF
	
COP_TxtBufOut:
	.GLOBAL COP_TxtBufOut
	LONG_OFF

	lda	DflTxtOut
	cmp	#TXTD_SCN
	beq	COP_ScnBufOut
	sec
	rtl
	
COP_KbCharIn:
	.GLOBAL COP_KbCharIn

; char in da tastiera
_KeybIn:
	.LONGA	off
	.LONGI	off

	pea	#DP01ADDR	; imposta nuova DP
	pld
	lda	ScnCol		; input tastiera - full editor
	sta	ScnColStart
	lda	ScnRow
	sta	ScnRowStart
	jsr	_ScnCharIn
	sta	STKCR,s
	clc
	rtl

COP_TxtCharIn:
	.GLOBAL COP_TxtCharIn
	.LONGA	off
	.LONGI	off

	lda	DflTxtIn
	asl	a
	tax
	jmp	(TXTITBL,x)	

COP_DevCharIn:
	.GLOBAL COP_DevCharIn
	.LONGA	off
	.LONGI	off

	lda	STKCR,s		; device
	tay
	asl	a
	tax
	jmp	(TXTITBL,x)	

COP_ScnXchg:
	.GLOBAL COP_ScnXchg
	LONG_OFF
	
	pea	#DP01ADDR	; imposta nuova DP
	pld
	lda	STKCR,s		; A = 0..1
	bne	?04		; switch to VGA
	lda	VBBRam
	bpl	?10		; already PAL -- exit
	stz	VBBRam
	bra	?08
?04:	bit	VBBRam
	bmi	?10		; already VGA -- exit
	bit	VBBFlag		; VGA installed?
	bpl	?10		; no
	lda	#$80
	sta	VBBRam
?08:	jsr	cinit
?10:	clc
	rtl
	
sys_tasksw:
	
	.PUBLIC sys_tasksw

	ACC16			; retrieve subfunction number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2	
	lda	[COPPtr]	; byte after signature
	beq	?80
	;bcc	?02		; ok
	sec
	lda	#EINV		; error
	rtl

?80:	jsr	LF8Tst
	clc
	rtl


sys_mkfs:
	.PUBLIC sys_mkfs

	clc
	rtl
