;----------------------------------------------------------
; FDC.ASM
; PROGETTO: B1601
;
; Gestione controller floppy - UM8388
;----------------------------------------------------------

_DMA_INC_	.SET	1
_RTC_INC_	.SET	1

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc

.LIST on

BUFDSK1		.EQU	$02
BUFDSK2		.EQU	$03

; costanti di temporizzazione
T1MS		.EQU	(4 * PHI2)	; costante 1ms (timer PHI2)
T10MS		.EQU	(10 * 1000 * PHI2)
T20MS		.EQU	20
SNDTMO		.EQU	2000		; 2000ms (timer PB6) - timeout SendTo

INTTMO		.EQU	2000		; 2000ms (timer PB6) - timeout INT
RESTMO		.EQU	2500		; 2500ms (timer PB6) - timeout res

MOTACL		.EQU	950		; ritardo spin-up (ms)
MOTDCL		.EQU	2600		; ritardo spin-down (ms)

; GAP 3 read/write
RWGPL3DD	.EQU	$1B		; GPL3 -> D.D. standard
RWGPL3DDA	.EQU	$10		; GPL3 -> D.D. 10 tracce
RWGPL3HD	.EQU	$36		; GPL3 -> H.D.
RWGPL3HDA	.EQU	$14		; GPL3 -> D.D. 10 tracce

; GAP 3 format
FMGPL3DD	.EQU	$54		; GPL3 -> D.D. standard
FMGPL3DDA	.EQU	$24		; GPL3 -> D.D. 10 tracce
FMGPL3HD	.EQU	$6C		; GPL3 -> H.D.
FMGPL3HDA	.EQU	$32		; GPL3 -> H.D.

; comandi per UM8388
CMDSPECIFY	.EQU	$03
CMDSENSEST	.EQU	$04
CMDRECAL	.EQU	$07
CMDSENSEINT	.EQU	$08
CMDSEEK		.EQU	$0F
CMDWRITE	.EQU	$45
CMDREADID	.EQU	$4A
CMDFMT		.EQU	$4D
CMDREAD		.EQU	$66

UM_SRTHUT	.EQU	$DF		; SRT = 3ms, HUT = 240ms
UM_HLTND	.EQU	$02		; HLT = 2ms, ND = 0 (DMA MODE)

; costanti per modo DMA (no autoinc, address inc, single mode)
VERFTRASF	.EQU	$42		; verfy transfer: read FDC, dummy MEM
WRITETRASF	.EQU	$46		; write transfer: read FDC, write MEM
READTRASF	.EQU	$4A		; read  transfer: read MEM, write FDC

maxret		.EQU	1		; retries #

; page 0 local var's (declared in bios temp. work area)
_P0BTMP:	.SECTION page0, ref_only, offset bwrktmp	; DP Tmp. BIOS

FDCTMP_START	.EQU	$

ncnt		.DW	; number of bytes for dma
ntrsf		.DW	; number of trasferred bytes
ptmp		.DW	; temp. pointer used by format
bufstart	.DW	; used by ram disk
bufend		.DW	; used by ram disk
bufp		LP	; buffer pointer
verfp		LP	; verify buffer pointer

xmstart		.EQU	ptmp	; ram disk start address
dmabuf		.EQU	bufend	; dma buffer # & needs dma buffers #
bufcnt		.EQU	bufend+1
drvhead		.EQU	bufstart
dtlstp		.EQU	bufstart+1
lpfn		.EQU	verfp	; callback function pointer

fdcres		.DS	7	; store phase reult bytes
drive		.DB	; current drive #
track		.DB	; current track
head		.DB	; current side
sector		.DB	; start sector
sectnum		.DB	; sector's number/track 
maxsect		.DB	; max. number of sects/track + 1
eot		.DB	; end of track
gap3		.DB	; gap 3 length
dmast		.DB	; save dma status reg.
fdop		.DB	; fdc function index
vdop		.DB	; ram disk operation
opfg		.DB	; flag param's

fmtfg		.EQU	opfg
buftst		.EQU	gap3	; used by ram disk

dskfmt		.DB	; disk format
fmtfill		.DB	; fill byte while format
tmpa		.DB
tmpx		.DB
retries		.DB	; retries #

xmbank		.EQU	retries	; bank that hold ram disk
mcand1		.EQU	tmpa	; used by ram disk
mcand2		.EQU	tmpx

FDCTMP_SIZ	.EQU	($-FDCTMP_START)

	.ENDS

	.CODEF8
	.GLOBAL fdinit, vdinit
	;.EXTERN LF8Tst, getctime
	
	.LONGA	off
	.LONGI	off
	
;---------------------------------------------------------------------------
; System interface
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
; sys_fdrw - read/write/verify sectors in floppy disk
;
; prototype: sys_fdrw(bDrive, bTrak, bHead, bSect, bNum, bBuf, bFlag)
;
; Params (7 bytes):
;	bDrive	= drive number (0..1)
;	bTrak 	= track number ($00..$4F)
;	bHead	= head number ($00..$01)
;	bSect	= sector number ($01...)
;	bNum	= number of sectors to be read/write ($01...)
;	bBuf	= buffer number ($00..$FF)
;	bFlag	= flag
;			<7>: verify data after writing
;			<0>: CBM format (otherwise DOS format)
;
; Out:
;	CF = 0 if operation completed successfully
;		C  = number of bytes effectively transferred
;		X  = disk format ($00, $01, $02, $03)
;		Y  = unchanged
;	CF = 1 if error
;		C  = number of bytes effectively transferred
;		X  = disk format ($00, $01, $02, $03, $FF)
;		Y  = error code
;	Status register always preserved except carry

; params offset
bFlag	.SET	STKPRMS
bBuf	.SET	STKPRMS + 1
bNum	.SET	STKPRMS + 2
bSect	.SET	STKPRMS + 3
bHead	.SET	STKPRMS + 4
bTrak	.SET	STKPRMS + 5
bDrive	.SET	STKPRMS + 6

sys_fdrw:
	.GLOBAL	sys_fdrw

	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	stz	ntrsf		; clear bytes transferred
	ACC08			; A,M -> 8 bit
	stz	fdcerr
	stz	dmast
	lda	#$FF		; invalid format
	sta	dskfmt
	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2	
	lda	[COPPtr]	; byte after signature
	cmp	#$03		; right value: 0,1,2
	bcs	?12		; error
	sta	fdop
	lda	bDrive,s	; drive number
	tax
	cpx	#FDMAXDRIVE
	bcs	?14		; error
	sta	drive
	lda	bTrak,s		; track number
	cmp	#$50		; max track number $4F
	bcs	?16		; error
	sta	track
	lda	bHead,s		; head number
	cmp	#$02
	bcs	?16
	sta	head
	sec
	lda	bSect,s		; starting sector
	beq	?16
	sta	sector
	lda	bNum,s		; number of sector
	beq	?16
	sta	sectnum
	lda	bBuf,s		; buffer number
	sta	dmabuf
	lda	bFlag,s		; flag
	and	#$81		; mask bit 7 and bit 0
	sta	opfg
	txa			; phisycal drive?
	beq	?08		; yes
	lda	vdrive		; ram disk formatted?
	bpl	?10		; no
	and	#$03
	cmp	#$03		; format $03?
	beq	?10		; yes, invalid format
	sta	dskfmt		; current ram disk format
	jsr	vdrw		; ram disk operation
	bra	?24
?08:	bit	fdcctl		; drive attached?
	bmi	?22		; yes
	bra	?14
?10:	lda	#FDC_BADFMT
	bra	?20
?12:	lda	#FDC_BADOP
	bra	?20
?14:	lda	#FDC_NODRIVE	; invalid drive
	bra	?20
?16:	lda	#FDC_PARAMS
?20:	sta	fdcerr
	sec
	bra	?24
?22:	jsr	fdrw		; fd operation
?24:	ACC16			; report out data
	lda	ntrsf
	sta	STKCR,s
	ACC08
	lda	dskfmt
	sta	STKXR,s
	lda	#0
	sta	STKXR+1,s
	lda	fdcerr		; error code if CF=1
	rtl
	
;---------------------------------------------------------------------------
; sys_fdtrack - read/write/verify whole track (both side) on floppy disk
;
; prototype: sys_fdtrack(bDrive, bTrak, bBuf, bFlag)
;
; Params (4 bytes):
;	bDrive	= drive number
;	bTrak 	= track number ($00..$4F)
;	bBuf	= buffer number ($00..$FF)
;	bFlag	= flag
;			<7>: verify data after writing
;			<0>: CBM format (otherwise DOS format)
;
; Out:
;	CF = 0 if operation completed successfully
;		C  = number of bytes effectively transferred
;		X  = disk format ($00, $01, $02, $03)
;		Y  = unchanged
;	CF = 1 if error
;		C  = number of bytes effectively transferred
;		X  = disk format ($00, $01, $02, $03, $FF)
;		Y  = error code
;	Status register always preserved except carry

; params offset
bFlag	.SET	STKPRMS
bBuf	.SET	STKPRMS + 1
bTrak	.SET	STKPRMS + 2
bDrive	.SET	STKPRMS + 3

sys_fdtrack:
	.GLOBAL	sys_fdtrack

	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	stz	ntrsf		; clear bytes transferred
	ACC08			; A,M -> 8 bit
	stz	fdcerr
	stz	dmast
	lda	#$FF
	sta	dskfmt
	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2	
	lda	[COPPtr]	; byte after signature
	cmp	#$03		; right value: 0,1,2
	bcs	?12		; error
	sta	fdop
	lda	bDrive,s	; drive number
	tax
	cpx	#FDMAXDRIVE
	bcs	?14		; error
	sta	drive
	lda	bTrak,s		; track number
	cmp	#$50		; max track number $4F
	bcs	?16		; error
	sta	track
	stz	head
	lda	#1
	sta	sector
	lda	bBuf,s		; buffer number
	sta	dmabuf
	lda	bFlag,s		; flag
	and	#$81		; mask bit 7 and bit 0
	ora	#$40		; flag MT
	sta	opfg
	txa			; phisycal drive?
	beq	?08		; yes
	lda	vdrive		; ram disk formatted?
	bpl	?10		; no
	and	#$03
	cmp	#$03		; format $03?
	beq	?10		; yes, invalid format
	sta	dskfmt		; current ram disk format
	jsr	vdrw		; ram disk operation
	bra	?24
?08:	bit	fdcctl		; drive attached?
	bmi	?22		; yes
	bra	?14
?10:	lda	#FDC_BADFMT
	bra	?20
?12:	lda	#FDC_BADOP
	bra	?20
?14:	lda	#FDC_NODRIVE
	bra	?20
?16:	lda	#FDC_PARAMS
?20:	sta	fdcerr
	sec
	bra	?24
?22:	jsr	fdrw		; fd drive operation
?24:	ACC16			; report out data
	lda	ntrsf
	sta	STKCR,s
	ACC08
	lda	dskfmt
	sta	STKXR,s
	lda	#0
	sta	STKXR+1,s
	lda	fdcerr		; error code if CF=1
	rtl

;---------------------------------------------------------------------------
; sys_fdfmt - low level format whole floppy disk
;
; prototype: sys_fdfmt(bDrive, bFlag, bFill, lpCall)
;
; Params (6 bytes):
;	bDrive	= drive number
;	bFlag 	= format flag
;		<7>: valid callback function in lpCall
;		<6>: verify track after format one
;		<1>: select H.D. format (otherwise D.D. format)
;		<0>: select CBM format (otherwise standard DOS format)
;	bFill 	= fill byte for all sectors
;	lpCall	= long pointer to callback user function while formatting
;
; Out:
;	CF = 0 if operation completed successfully
;		A,X,Y preserved
;	CF = 1 if error
;		A, X preserved
;		Y = error code
;	Status register always preserved except carry

; params offset
lpCall	.SET	STKPRMS
bFill	.SET	STKPRMS + 3
bFlag	.SET	STKPRMS + 4
bDrive	.SET	STKPRMS + 5

sys_fdfmt:
	.GLOBAL	sys_fdfmt

	lda	bDrive,s	; recover params from stack
	sta	drive		; drive number
	tay
	lda	bFlag,s
	sta	fmtfg		; format flag
	tax
	and	#00000011B
	sta	dskfmt		; required format
	cmp	#00000011B	; format 3 is invalid
	bne	?01
	lda	#FDC_BADFMT
	sec
	bra	?05		; error
?01:	lda	bFill,s
	sta	fmtfill		; fill byte
	txa
	bpl	?04		; no callback
	lda	lpCall+2,s	; bank that hold callback function
	sta	lpfn+2
	bne	?02		; callback function cannot stay in bank $00		
	lda	#$80		; invalid callback
	trb	fmtfg		; no callback
	bra	?04
?02:	CPU16
	lda	lpCall,s	; address of callback function
	sta	lpfn
	CPU08
?04:	lda	#0
	tax
	sta	fdcerr		; clear error
	sta	track
	sta	head
	sta	bufp		; set dma buffer at $030000
	sta	bufp+1
	inc	a
	sta	sector		; sector = 1
	inc	a
	inc	a
	sta	bufp+2
	dex
	stx	dtlstp		; DTL/STP = $FF
	ldx	dskfmt
	lda	>MAXSECT,x
	sta	maxsect		; max. numbers of sectors/track + 1
	dec	a
	sta	sectnum		; sectors/track 
	sta	eot		; end of track
	cpy	#FDMAXDRIVE
	bcc	?06
	lda	#FDC_NODRIVE
?05:	sta	fdcerr
	rtl
?06:	tyx			; phisycal drive?
	beq	?08		; yes
	jsr	vdfmt		; ram disk formattation
	bra	?10
?08:	jsr	fdfmt		; format floppy disk
?10:	lda	fdcerr		; report error if one
	rtl

;---------------------------------------------------------------------------
; sys_fdverf - verify integrity of low level format of whole floppy disk
;
; prototype: sys_fdverf(bDrive, lpCall)
;
; Params (4 bytes):
;	bDrive	= drive number
;	lpCall	= long pointer to callback user function while verify
;		  if lpCall = null the callback function is not called
;		  callback function ignored for ram disk
;
; Out:
;	CF = 0 if operation completed successfully
;		A,X preserved
;		Y = disk format
;	CF = 1 if error
;		A, X preserved
;		Y = error code
;	Status register always preserved except carry

; params offset
lpCall	.SET	STKPRMS
bDrive	.SET	STKPRMS + 3

sys_fdverf:
	.GLOBAL	sys_fdverf

	stz	fmtfg		; no callback
	stz	fdcerr
	lda	bDrive,s	; recover params from stack
	sta	drive		; drive number
	tay
	lda	lpCall+2,s	; bank that hold callback function
	sta	lpfn+2
	beq	?02		; callback function cannot stay in bank $00		
	CPU16
	lda	lpCall,s	; address of callback function
	sta	lpfn
	CPU08
	lda	#$80		; call callback function
	tsb	fmtfg
?02:	cpy	#FDMAXDRIVE
	bcc	?06
	lda	#FDC_NODRIVE	; error
?04:	sta	fdcerr
	rtl
?06:	tyx			; phisycal drive?
	beq	?10		; yes
	lda	vdrive		; ram disk formatted?
	bmi	?08		; yes
	lda	#FDC_BADFMT
	sec
	bra	?04
?08:	and	#00000011
	clc
	bra	?12
?10:	jsr	fdverf		; verify floppy disk format
	bcs	?14
	lda	dskfmt		; get format
?12:	sta	STKYR,s
	lda	#0
	sta	STKYR+1,s
?14:	lda	fdcerr		; report error if one
	rtl

;---------------------------------------------------------------------------
; sys_fdctl - control function 
;
; prototype: sys_fdctl(X=drive number if required)
;
; Functions:
;	$00: 	reset FDC controller
;			input  reg.: none
;			output reg.: unchanged
;	$01: 	POST routine (init FDC controller)
;			input  reg.: none
;			output reg.: unchanged
;	$02:	get drive status
;			input  reg.: X = drive number
;			output reg.: Y = drive status (ST3), A,X unchanged
;			output reg.: A,X,Y unchanged
;	$03:	check if disk inserted into drive
;			input  reg.: X = drive number
;			output reg.: unchanged
;	$04:	get format diskette if inserted into drive
;			input  reg.: X = drive number
;			output reg.: Y = diskette format, A,X unchanged
;	$05:	set write protect bit on ram disk (no effect for phis. drive)
;			input  reg.: X = drive number
;			output reg.: unchanged
;	$06:	reset write protect bit on ram disk (no effect for phis. drive)
;			input  reg.: X = drive number
;			output reg.: unchanged
; Out:
;	CF = 0 if operation completed successfully
;	CF = 1 if error
;		A, X preserved
;		Y = error code
;	Status register always preserved except carry
sys_fdctl:
	.GLOBAL	sys_fdctl

	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	ACC08			; A,M -> 8 bit
	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2	
	lda	STKXR,s
	sta	drive		; get drive # (X reg.)	
	lda	[COPPtr]	; byte after signature
	cmp	#MXCTLFN
	bcc	?04
	lda	#FDC_BADOP
	sta	fdcerr
	rts
?04:	stz	fdcerr	
	asl	a
	tax			; function index
	jsr	(CTLFN,x)
	lda	fdcerr		; return error code if any
	rtl

; ctl function 0: reset controller	
ctl_fn0:
	jmp	fdcrst

; ctl function 1: POST routine (init)	
ctl_fn1:
	jmp	fdinit

; ctl function 2: get drive status	
ctl_fn2:
	jsr	ctl_tst		; check drive #
	tax			; phisycal drive?
	beq	?10		; yes
	lda	vdrive		; bit 4: wp on bit
	asl	a
	asl	a		; shift to bit 6
	and	#01000000B
	ora	#00100000B	; ready bit
	clc
?05:	sta	STKYR+2,s		; report status
	lda	#0
	sta	STKYR+3,s
	rts
?10:	jsr	ctl_dr		; check drive attached
	jsr	drvst
	bcc	?05
	rts
	
; ctl function 3: check disk ready
ctl_fn3:
	jsr	ctl_tst		; check drive #
	tax			; phisycal drive?
	beq	?10		; yes
	bit	vdrive		; test bit 6: change disk bit
	clc
	bvc	?05		; disk not changed
	lda	#$40
	trb	vdrive		; reset change disk bit
	lda	#FDC_CHANGE
	sta	fdcerr
	sec
?05:	rts
?10:	jsr	ctl_dr		; check drive attached
	jsr	chkdsk		; change disk line test
	jmp	stop

; ctl function 4: get disk format
ctl_fn4:
	jsr	ctl_tst		; check drive #
	tax			; phisycal drive?
	beq	?10		; yes
	lda	vdrive		; test bit 7: ram disk formatted?
	bmi	?02		; yes
	lda	#FDC_BADFMT
	sta	fdcerr
	sec
	rts
?02:	and	#00000011B	; report current format
	sta	STKYR+2,s
	lda	#0
	sta	STKYR+3,s	
	clc
	rts
?10:	stz	fdcdrv		; get disk format - invalidate current format
	jsr	chkfmt		; try to check format
	bcs	?20		; error
	lda	fdcdrv		; update format
	and	#01111100B
	ora	dskfmt
	ora	#$80		; recognized format
	sta	fdcdrv
	lda	dskfmt		; report format in Y reg.
	sta	STKYR+2,s
	lda	#0
	sta	STKYR+3,s	
?20:	jmp	stop

; ctl function 5: set write protect bit on ram disk
ctl_fn5:
	jsr	ctl_tst		; check drive #
	tax			; phisycal drive?
	beq	?10		; yes, done
	lda	#$10		; set bit 4
	tsb	vdrive
?10:	clc	
	rts

; ctl function 6: reset write protect bit on ram disk
ctl_fn6:
	jsr	ctl_tst		; check drive #
	tax			; phisycal drive?
	beq	?10		; yes, done
	lda	#$10		; reset bit 4
	trb	vdrive
?10:	clc	
	rts

; test drive number
ctl_tst:
	lda	drive
	cmp	#FDMAXDRIVE
	bcc	?10
	lda	#FDC_NODRIVE	; error: invalid drive
	sta	fdcerr
	pla			; skip return address
	pla
?10:	rts

; test if drive is attached to controller
ctl_dr:
	bit	fdcctl
	bmi	?10
	lda	#FDC_NODRIVE	; error: invalid drive
	sta	fdcerr
	pla			; skip return address
	pla
?10:	rts

;---------------------------------------------------------------------------
; fdc read/write & format routines
;---------------------------------------------------------------------------

; execute read/verf/write track/sector's on floppy disk
; param's:
;	fdop=fdc operation (0=read,1=fake verify,2=write)
;	track=fd track
;	sector=fd first sector
;	head=fd side
;	eot=fd last sector
;	gap3=gap3 length for read/write
;	dtlstp=fd DTL/STP param
;	ncnt=bytes # to transfer
;	opfg<7>=verify buffer with disk
;	opfg<6>=MT command (multi track)
;	bufp=buffer pointer
;	verfp=verify buffer pointer
;
; out:	CF=1 if error, fdcerr hold error code
fdrw:
	jsr	chkfmt		; start motor and test disk format
	bcc	?00
	jmp	stop		; stop motor
?00:	jsr	chkparms	; check params
	bcc	?01
	jmp	stop		; stop motor
?01:	lda	#maxret		; retries count
	sta	retries
	bit	opfg		; will verify buffer?
	bpl	?02		; no
	jsr	savebuf		; save dma buffer for verify	
?02:	ldy	track		; seek track
	jsr	seek
	bcs	?14		; error: retry?
	ldx	fdop
	lda	>CMDTAB,x	; FDC command
	pha
	lda	>DMATAB,x	; DMA mode
	xba			; B=dma mode
	ldx	bufp
	ldy	bufp+1
	lda	bufp+2
	jsr	dmaset		; dma setup
	pla			; FDC command
	bit	opfg
	bvc	?04		; no MT
	ora	#$80		; MT operation
?04:	jsr	fdcmd		; execute command
	bcs	?14		; error: retry?
	bit 	opfg		; will verify data?
	bpl	?16		; no -- no error
	lda	#WRITETRASF	; read disk into verify buffer
	xba			; dma mode (read data)
	ldx	verfp
	ldy	verfp+1
	lda	verfp+2
	jsr	dmaset		; dma setup
	lda	#CMDREAD
	bit	opfg
	bvc	?06		; no MT
	ora	#$80		; MT operation
?06:	jsr	fdcmd		; execute command
	bcs	?14		; error: retry?
	ldy	#0
	INDEX16			; verify buffer's
?08:	lda	[bufp],y
	cmp	[verfp],y
	bne	?10		; unmatch (ZF=0)
	iny
	cpy	ncnt
	bcc	?08
?10:	INDEX08			; here ZF=1 if match, ZF=0 if not
	clc			; assume no error
	beq	?16		; match
	lda	#FDC_VER	; verify error
	sec
?14:	dec	retries		; try again?
	bne	?02		; yes
	bra	?18		; CF=1, A=error code	
?16:	lda	fdcdrv		; set valid format
	and	#01111100B
	ora	dskfmt		; update format in drive
	ora	#$A0		; current valid format&trust bit
	sta	fdcdrv
?18:	bit 	opfg		; will restore dma buffer after verify?
	bpl	done		; no
	jsr	restbuf		; restore dma buffer

; end fdc operation: reset controller if any error
; stop motor & set fdcerr if error
; in:	CF=1 if error, A=error code
done:
	bcc	stop2
	pha			; save error code
	jsr	fdcrst		; reset controller after error
	pla			; restore error code
	sec

; stop motor & set error code if any
; in:	CF=1 if error, A=error code
stop:
	bcc	stop2
	sta	fdcerr		; set error code
stop2:
	jmp	stopmtr

; format whole floppy disk
; out:	CF=1 if error, fdcerr hold error code
fdfmt:
	bit	fdcctl		; drive attached?
	bmi	?02		; yes
	lda	#FDC_NODRIVE
	sta	fdcerr
	sec
	rts
?02:	jsr	chkdsk		; start motor and check if disk in drive
	bcc	?04		; ok - ready
	bvc	stop		; not ready, stop motor&exit
?04:	lda	#$10		; test disk write protect on
	bit	fdcdrv
	beq	?06
	lda	#FDC_WP		; wp on error
	bra	stop		; stop motor&exit
?06:	stz	fdcerr		; clear error
	lda	fdcdrv		; reset format
	and	#01011100B
	sta	fdcdrv
	bit	fdcdrv		; need to call seek0 double step?
	bvs	?08		; no
	jsr	seek0		; double step seek track 0
	bcs	stop		; error
?08:	lda	dskfmt		; set FDC rate	
	jsr	setrate
	lda	#$01		; reset bit 0...
	trb	fmtfg		; ...for first call to callback
	ACC16
	tsc			; make room for a local buffer in stack
	sec
	sbc	#80
	tcs
	inc	a		; pointer to temp. buffer
	sta	ptmp
	ACC08
	ldy	#79		; save buffer dma in temp. buffer
?09:	lda	[bufp],y
	sta	(ptmp),y
	dey
	bpl	?09
	iny			; Y=0
	tyx			; X=0
?10:	sty	track		; format this track...
?11:	stx	head		; ...in this side
	lda	#maxret	
	sta	retries		; retries count
?12:	ldy	#0		; fill buffer with data for current track/head
	ldx	#1		; X=sector
?14:	lda	track 
	sta	[bufp],y	; C=track
	iny
	lda	head
	sta	[bufp],y	; H=head
	iny
	txa
	sta	[bufp],y	; R=sector
	iny
	lda	#2
	sta	[bufp],y	; N=size of sector
	iny
	inx
	cpx	maxsect		; loop
	bcc	?14
	sty	ncnt		; DMA count of bytes
	stz	ncnt+1	
	jsr	tstrdy		; test disk ready
	bcs	?20		; error - no disk or disk changed
	lda	#READTRASF	; read transfer
	xba			; B=dma mode
	ldx	bufp
	ldy	bufp+1
	lda	bufp+2
	jsr	dmaset
	bit	fmtfg
	bpl	?16		; no callback
	phk			; JSL simulation
	pea	#?16-1		; return address
	lda	eot		; params for callback
	xba			; B = sector/track
	lda	fmtfg
	lsr	a		; bit 0 -> carry -> 0 if first call
	lda	drive		; A = drive
	ldy	track		; Y = track
	ldx	head		; X = head
	jmp	[lpfn]
?16:	; return from notify callback
	ldy	track		; track seek
	jsr	seek
	bcs	?20		; seek error
	jsr	fdftcmd		; format track
	bcc	?18		; no error
	dec	retries
	bne	?12		; try again
	bra	?20
?18:	lda	#$01
	tsb	fmtfg		; bit 0 = 1 -> for callback
	ldx	head		; next side
	inx
	cpx	#2
	bcc	?11		; format same track in next side
	ldx	#0		; restart side=0
	ldy	track		; next track
	iny
	cpy	#80
	bcc	?10		; format next track
	clc			; format end without error
?20:	php			; save carry
	plx			; save carry in X bit 0
	ldy	#79		; restore buffer dma
?22:	lda	(ptmp),y
	sta	[bufp],y
	dey
	bpl	?22
	ACC16CLC
	tsc			; restore stack frame
	adc	#80
	tcs
	ACC08
	txa			; A = saved status
	lsr	a		; CF=1 if error
	bcs	?24		; error	
	lda	fdcdrv		; set valid format
	and	#01111100B
	ora	dskfmt		; update format in selected drive
	ora	#$80		; current valid format (but not trust bit)
	sta	fdcdrv
?24:	jmp	done		; stop motor - reset controller if error

; verify low level format of floppy disk
; out:	CF=1 if error, fdcerr hold error code
fdverf:
	bit	fdcctl		; drive attached?
	bmi	?02		; yes
	lda	#FDC_NODRIVE
	sta	fdcerr
	sec
	rts
?02:	stz	fdcdrv		; get disk format - invalidate current format
	jsr	chkfmt		; try to check format
	bcc	?06
?04:	jmp	stop		; error
?06:	lda	#0
	tax
	sta	bufp		; set dma buffer at $030000
	sta	bufp+1		; fake transfer
	inc	a
	sta	sector		; sector = 1
	inc	a
	inc	a
	sta	bufp+2
	dex
	stx	dtlstp		; DTL/STP = $FF
	ldx	dskfmt
	lda	>MAXSECT,x
	sta	maxsect		; max. numbers of sectors/track + 1
	dec	a
	sta	sectnum		; sectors/track 
	sta	eot		; end of track
	asl	a		; num. of bytes
	sta	ncnt+1
	stz	ncnt
	ldy	#0		; Y=0
	tyx			; X=0
?10:	sty	track		; check this track...
?11:	stx	head		; ...in this side
	lda	#maxret	
	sta	retries		; retries count
?12:	jsr	tstrdy		; test disk
	bcs	?04		; error - no disk or disk changed	
	lda	#VERFTRASF	; fake transfer, only check track
	xba			; B=dma mode
	ldx	bufp		; buffer pointer
	ldy	bufp+1
	lda	bufp+2
	jsr	dmaset
	bit	fmtfg
	bpl	?14		; no callback
	phk			; JSL simulation
	pea	#?14-1		; return address
	lda	dskfmt
	xba			; B = sector/track
	lda	fmtfg
	lsr	a		; bit 0 -> carry -> 0 if first call
	lda	drive		; A = drive
	ldy	track		; Y = track
	ldx	head
	jmp	[lpfn]
?14:	; return from notify callback
	ldy	track		; track seek
	jsr	seek
	bcs	?04		; seek error
	lda	#CMDREAD	; fake read whole track	in current side
	jsr	fdcmd
	bcc	?18
	dec	retries
	bne	?12
	bra	?20
?18:	lda	#$01
	tsb	fmtfg		; bit 0 = 1 -> for callback
	ldx	head		; next side
	inx
	cpx	#2
	bcc	?11		; same track, side=1
	ldx	#0		; clear head
	ldy	track		; next track
	iny
	cpy	#80
	bcc	?10		; next track, side=0
	lda	fdcdrv		; set valid format
	and	#01111100B
	ora	dskfmt		; update format (but not trut bit)
	ora	#$80		; current valid format
	sta	fdcdrv
	clc
?20:	jmp	done

;---------------------------------------------------------------------------
; fdc basic read/write/format command's
;---------------------------------------------------------------------------

; fd format track command
; out:	CF=1 if error, A=error code
; A,X,Y destroyed
fdftcmd:
	stz	ntrsf		; clear numbers of transferred bytes
	stz	ntrsf+1
	lda	dskfmt
	and	#$03		; disk format [0..2]
	tax			; X=format
	lda	>FGPLTBL,x	; GAP 3 length
	sta	gap3
	lda	>HEADTBL,x	; head inversion for CBM format
	eor	head
	asl	a		; head -> bit 2
	asl	a
	sta	drvhead		; head + drive (=0)	
	lda	#CMDFMT		; format track command
	jsr	sendto
	lda	drvhead		; head + drive
	jsr	sendto
	lda	#2		; sector size = 512 bytes
	jsr	sendto
	lda	eot		; SC = sector/track
	jsr	sendto
	lda	gap3		; GPL 3 length
	jsr	sendto
	lda	fmtfill		; D = fill byte
	jsr	sendto
	bit	VIA0+VIAPRB	; clear CB1 flag	
	lda	#$02		; enable DMA bus
	trb	VIA0+VIAPRANH	; PA1 -> /DME = 0
	jsr	chkres		; wait result
	jsr	chkdma
	bcs	?10		; error in phase result
	bit	fmtfg		; verf track ?
	bvc	?10		; no
	lda	eot
	asl	a		; number of bytes
	sta	ncnt+1
	stz	ncnt
	stz	ntrsf		; clear numbers of transferred bytes
	stz	ntrsf+1
	lda	#VERFTRASF	; fake transfer, just check track
	xba			; B=dma mode
	ldx	bufp		; buffer pointer
	ldy	bufp+1
	lda	bufp+2
	jsr	dmaset
	lda	#CMDREAD	; fake read whole track	in current side
	jsr	fdcmd
?10:	rts

; fd read/write/verf command
; in:	A=command
; out:	CF=1 if error, A=error code
; A,X,Y destroyed
fdcmd:
	;stz	ntrsf		; clear numbers of transferred bytes
	;stz	ntrsf+1
	tay			; Y=command
	lda	dskfmt
	and	#$03		; disk format [0..2]
	tax			; X=format
	lda	>RWGPLTBL,x	; GAP 3 length for required format
	sta	gap3
	lda	>HEADTBL,x	; head inversion for CBM format
	eor	head
	asl	a		; head -> bit 2
	asl	a
	sta	drvhead		; head + drive (=0)
	tya			; A=command
	jsr	sendto
	lda	drvhead		; head + drive
	jsr	sendto
	lda	track		; track
	jsr	sendto
	lda	head		; head
	jsr	sendto
	lda	sector		; sector
	jsr	sendto
	lda	#2		; sector size = 512 bytes
	jsr	sendto
	lda	eot		; EOT (last sector)
	jsr	sendto
	lda	gap3		; GPL 3 length
	jsr	sendto
	lda	dtlstp		; DTL/STP
	jsr	sendto
	bit	VIA0+VIAPRB	; clear CB1 flag (FDC interrupt)
	lda	#$02		; enable DMA bus control
	trb	VIA0+VIAPRANH	; PA1 -> /DME = 0
	jsr	chkres		; wait for transfer ending and CPU bus enable

; read from dma count the effective transferred bytes count
chkdma:
	pha			; save error code if any
	php			; save carry for pending error
	lda	!DMASR		; save DMA status & clear status reg
	sta	dmast
	bcs	?30		; error pending
?10:	sta	!DMAWCLRFF	; loop for read dma reg.
	nop
	nop
	lda	!DMAC+DMACNT2	; low count DMA residue
	xba
	nop
	lda	!DMAC+DMACNT2	; high count DMA residue
	xba
	nop
	sta	!DMAWCLRFF
	cmp	!DMAC+DMACNT2
	nop
	bne	?10
	xba
	nop
	cmp	!DMAC+DMACNT2
	nop
	bne	?10
	xba			; C = residual bytes # 
	ACC16
	inc	a		; DMA count residue
	sta	ntrsf
	sec
	lda	ncnt		; # transferred bytes
	sbc	ntrsf
	sta	ntrsf
	cmp	ncnt
	ACC08
	bne	?20		; error: no match
	lda	dmast		; check dma status
	bit	#$04		; premature cycle ending?
	bne	?30		; no
?20:	lda	$01,s		; set CF in saved status
	ora	#PCFLAG		; set error flag	
	sta	$01,s
	lda	#FDC_DLOSS	; data loss error
	sta	$02,s		; report error
?30:	plp			; return CF
	pla			; return error code
	rts

;---------------------------------------------------------------------------
; disk check routines
;---------------------------------------------------------------------------

; test if disk ready while motor is spinning
; out:	CF=0 if disk ready
;	CF=1 if disk not ready or changed, A=FDC_NOTREADY
; A destroyed, X&Y preserved
tstrdy:
	clc
	lda	!FDCDIR		; read disk change line
	and	#$80		; bit 7 will be 0
	beq	?10		; ready - exit with CF = 0
	lda	#FDC_NOTREADY	; not ready
	sec			; error
?10:	rts

; check disk format - if the actual disk was not never checked it will be
; tested for the format
; out:	CF=1 if error, A=error code
; A&Y destroyed, X preserved
chkfmt:
	jsr	chkdsk		; start motor and check if disk ready
	bcc	?06		; disk ready and not changed
	bvc	?20		; not ready - exit with CF = 1
	stz	fdcerr		; disk changed
?06:	bit	fdcdrv		; need to call seek0 double step?
	bvs	?08		; no
	jsr	seek0		; double step seek track 0
	bcs	?20		; error
?08:	lda	fdcdrv		; bit 7 -> know format ?
	bmi	?30		; yes, so check and compare	
	lda	#$02		; start with H.D.
?10:	sta	dskfmt
	jsr	setrate		; setup rate
	jsr	?50		; little delay
	jsr	readid		; read ID track=0 head=0
	bcc	?12		; ok
	lda	dskfmt
	beq	?16		; error - unrecognized format
	lda	#0		; try D.D. rate
	bra	?10
?12:	jsr	?14		; check track & head
	bcs	?20
	lda	#$80
	tsb	fdcdrv		; set know format bit
	rts
?14:	lda	fdcres+3	; check track
	cmp	fdctrk
	beq	?18		; will be the same
?16:	lda	#FDC_BADFMT	; unknow format
?17:	sec
	rts
?18:	lda	fdcres+4	; head is inverted if CBM format
	tsb	dskfmt		; this set eventually bit 0 for CBM format
	clc
?20:	rts
?30:	bit	#$20		; trust format?
	bne	?34		; yes
	and	#$02		; mask rate	
	sta	dskfmt		; here bit 0 = 0
	tay
	eor	fdcctl		; compare with current rate
	and	#$02
	beq	?31		; same rate
	tya
	jsr	setrate
	jsr	?50		; little delay
?31:	lda	#$80
	trb	fdcdrv		; clear know format bit
	jsr	readid		; read first ID in current track
	bcs	?20		; error
	jsr	?14		; check track & head
	bcs	?20		; error
	lda	fdcdrv
	eor	dskfmt		; compare computed format with the stored one
	and	#$03		; will be 0
	beq	?32
	lda	#FDC_FMT	; mismatch format
	bra	?17
?32:	lda	#$80
	tsb	fdcdrv		; set know format bit
	clc
	rts
?34:	and	#$03		; trust format
	sta	dskfmt
	clc
	rts

; 1ms delay
?50:	lda	#$20		; T2 count PHI2 pulses
	trb	VIA0+VIAACR
	lda	#T1MS
	stz	VIA0+VIAT2CL
	sta	VIA0+VIAT2CH
?55:	lda	VIA0+VIAIFR
	bit	#T2IFRB
	beq	?55
	rts

; check disk in drive - if disk is changed reset controller and recalibrate
; drive & move head in track 0 & aquire wp status on disk
; out:	CF=0 if disk ready
;	CF=1 if disk not ready or changed
;	VF=1 if disk changed else VF=0 if not ready
;	A=error code FDC_NOTREADY or FDC_CHANGE or others
; A destroyed, X&Y preserved
chkdsk:
	phy
	stz	fdcerr
	jsr	startmtr	; start motor
	clc
	lda	!FDCDIR		; read disk change line
	and	#$80		; bit 7 will be 0
	beq	?20		; ready so aquire wp status
	lda	#$01		; clear ready bit
	trb	fdcctl
	jsr	fdcrst		; aquire new disk change line status
	bcs	?30		; error
	jsr	seek0		; recalibrate after reset and seek track 0
	bcs	?30		; error
	ldy	#FDC_CHANGE	; disk change?
	sep	#PVFLAG		; VF = 1 if disk change
	lda	!FDCDIR		; read line
	and	#$80		; bit 7 = 0 if disk inserted into drive
	beq	?10		; so disk changed
	ldy	#FDC_NOTREADY	; not ready
	clv			; VF flag = 0 if not ready
?10:	sec			; error
	tya			; A=error code
	bvc	?30		; not ready
?20:	php			; save CF&VF
	pha			; save error code if any
	jsr	chkwp		; aquire wp disk status
	pla
	plp
?30:	ply	
	rts

; set write protect on bit and disk ready bit 
chkwp:
	lda	#$10
	trb	fdcdrv		; clear wp on bit
	lda	#$01		; set ready bit
	tsb	fdcctl	
	lda	#CMDSENSEST	; sense status
	jsr	sendto
	lda	#0		; drive
	jsr	sendto
	jsr	waitres		; get result
	bit	fdcres		; ST3 bit 6 is wp on bit
	bvc	?10		; no wp
	lda	#$10
	tsb	fdcdrv		; set bit 4: wp on
?10:	rts

; sense phisycal drive #0 status
; out:	A=ST3 status if no error
;	CF=1 if error, fdcerr=error code
drvst:
	jsr	startmtr
	jsr	?20
	jmp	stop

?20:	lda	#CMDSENSEST	; sense status
	jsr	sendto
	lda	#0		; drive
	jsr	sendto
	jsr	waitres		; get result
	lda	fdcres		; get ST3
	clc
	rts

; read first valid ID in current track, head = 0
; if current track is not valid ($FF) seek to track 0
; out:	CF=1 if error, A=error code
; A&Y destroyed, X preserved
readid:
	lda	fdctrk
	cmp	#$FF
	bne	?02
	ldy	#0
	jsr	seek
	bcs	chk
?02:	lda	#CMDREADID
	jsr	sendto
	lda	#0		; head + drive (head = 0)
	jsr	sendto
	
; wait phase result after read/write/format/read ID and check result
; out:	CF=1 if error, A=error code
; A&Y destroyed, X preserved
chkres:
	jsr	waitint		; wait interrupt
	jsr	waitres		; get result
	lda	fdcres		; test ST0
	and	#11000000B
	beq	chk		; here CF = 0 - result ok
	cmp	#01000000B
	bne	?10		; bad controller
	clc			; no error
	lda	fdcres+1	; test ST1
	bmi	chk		; 7 -> end of track, no error
	asl	a		; skip bit 7
	asl	a		; 6 -> skip
	asl	a
	ldy	#FDC_DATACRC
	bcs	?20		; 5 -> CRC error
	asl	a
	ldy	#FDC_DLOSS
	bcs	?20		; 4 -> data loss
	asl	a
	asl	a
	ldy	#FDC_NOID
	bcs	?20		; 2 -> ID not found		
	asl	a
	ldy	#FDC_WP
	bcs	?20		; 1 -> write protect
	asl	a
	ldy	#FDC_MARK
	bcs	?20		; 0 -> mark address
?10:	ldy	#FDC_BAD	; bad controller
	sec
?20:	tya			; A=error code
chk:	rts

;---------------------------------------------------------------------------
; drive seek routines
;---------------------------------------------------------------------------

; recalibrate drive if need and move head in track 0 with a double step seek
; this routine should be called if bit 6 of fdcdrv is 0 as after a reset
; out:	CF=1 if error, A=error code
; A&Y destroyed, X preserved
seek0:
	ldy	#1		; seek track 1
	jsr	seek		; and recalibrate if need
	bcs	?10		; error
	ldy	#0		; seek track 0
	jsr	seek
	bcs	?10
	lda	#$40		; set bit 6 of fdcdrv
	tsb	fdcdrv
?10:	rts

; move head in track Y & recalibrate drive if need
; in:	Y=track
; A destroyed, X&Y preserved
seek:
	bit	fdcctl		; test bit 6: need restore?
	bvc	?04		; no
	jsr	restore		; recalibrate drive
	bcs	?10		; error - exit
	stz	fdctrk		; store current track (=0)
	lda	#$40
	trb	fdcctl		; reset flag restore: done
	cpy	#0		; requested track=0 ?
	beq	?06		; yes, exit after a while
?04:	cpy	fdctrk		; already in the desidered track ?
	beq	?06		; yes, exit
	lda	#CMDSEEK	; send seek command
	jsr	sndto2
	bcs	?08
	lda	#0		; drive
	jsr	sndto2
	bcs	?08	
	tya			; track
	sta	fdctrk
	jsr	sndto2
	bcs	?08
	lda	#FDC_SEEK	; error seek ?
	jsr	wseek		; wait result
	bcs	?08
?06:	clc	
?08:	jsr	?20
	bcc	?12
?10:	sta	tmpa	
	lda	#$FF		; invalidate current track
	sta	fdctrk
	lda	tmpa
?12:	rts

; 10ms delay
?20:	php
	pha
	lda	#$20		; T2 count PHI2 pulses
	trb	VIA0+VIAACR
	lda	#<T10MS
	sta	VIA0+VIAT2CL
	lda	#>T10MS
	sta	VIA0+VIAT2CH
?22:	TASKSW	
	lda	VIA0+VIAIFR
	bit	#T2IFRB
	beq	?22
	pla
	plp
	rts

; causes the read/write head within the FDD to retract in the track 0 position
; out:	CF=1 if error, A=error code
; A destroyed, X&Y preserved
restore:
	jsr	?10		; ricalibrate: first step
	bcc	wsk		; no double step need
				; if error, need one step more
?10:	lda	#CMDRECAL	; ricalibrate command
	jsr	sndto2
	bcs	wsk		; error
	lda	#0		; drive
	jsr	sndto2
	bcs	wsk		; error
	lda	#FDC_FAULT	; restore error ?

; wait result phase after seek/recalibrate command
; in:	A=error code FDC_FAULT or FDC_SEEK
; out:	CF=1 if error, A=error code
; A,X,Y preserved
wseek:
	sta	tmpa		; save error code
	jsr	?20		; sense interrupt
	bcs	wsk		; return pending error
	lda	fdcres		; test result status
	and	#01100000B
	cmp	#01100000B
	sec
	beq	?10		; restore/seek error
	clc			; no error
?10:	lda	tmpa
	rts

; wait result phase after sensing interrupt
?20:	jsr	waitint		; wait interrupt within timeout
	lda	#CMDSENSEINT	; sense interrupt
	jsr	sendto
	jsr	waitres		; wait result phase
wsk:	rts

;---------------------------------------------------------------------------
; low level controller routines
;---------------------------------------------------------------------------

; dma setup on channel 2 for a transfer cycle
; in:	X=transfer buffer low pointer
;	Y=transfer buffer high pointer
;	A=transfer buffer bank
;	B=dma mode transfer
;	ncnt=bytes to transfer count #
dmaset:
	pha
	lda	#$01
	trb	VIA0+VIAPRANH	; DMA X16=0 (bank 2)
	pla
	and	#$01		; mask bit 0
	beq	?02		; transfer buffer is in bank 2
	lda	#$01
	tsb	VIA0+VIAPRANH	; DMA X16=1 (bank 3)
?02:	lda	!DMASR		; clear status flag
	nop
	nop
	lda	!DMASR
	lda	#$06
	nop
	sta	!DMAWMSKB	; mask off bit channel 2
	nop
	xba			; A=transfer mode
	sta	!DMAMODE	; set DMA transfer mode
	nop
?06:	sta	!DMAWCLRFF	; clear F/F
	nop
	nop
	stx	DMAC+DMAADDR2	; store low buffer address
	nop
	nop
	sty	DMAC+DMAADDR2	; store high buffer address
	nop
	sta	!DMAWCLRFF	; clear F/F
	nop
	nop
	cpx	DMAC+DMAADDR2	; cmp low buffer address
	nop
	bne	?06
	cpy	DMAC+DMAADDR2	; cmp high buffer address
	nop
	bne	?06
	ACC16
	lda	ncnt
	dec	a		; DMA count=ncnt-1
	pha
	ACC08
	plx			; X=low count
	ply			; Y=high count
?08:	sta	!DMAWCLRFF	; clear F/F	
	nop
	nop
	stx	DMAC+DMACNT2	; store low count
	nop
	nop
	sty	DMAC+DMACNT2	; store high count
	nop
	sta	!DMAWCLRFF	; clear F/F	
	nop
	nop
	cpx	DMAC+DMACNT2	; cmp low count
	nop
	bne	?08
	cpy	DMAC+DMACNT2	; cmp high count
	nop
	bne	?08
	sta	!DMAWCLRFF	; clear F/F	
	lda	#$02
	nop
	nop
	sta	DMAWMSKB	; un-mask bit channel 2
	lda	#$08
	tsb	fdcctl		; set bit 3
	rts

; reset controller FDC after boot or for recovering after error
; motor status as in current fdcctl bit 4
; after reset restore flag (bit 6) in fdcctl are set
; and drive status will be cleared
; out:	CF=1 if error, A=error code
;	fdcctl<6:5> set/reset
; registers: A&Y destroyed, X preserved
fdcrst:
	php			; save I status
	sei			; disable interrupt
	bit	VIA0+VIAPRB	; reset flag CB1
	stz	fdcerr
	lda	#$20
	trb	fdcctl		; reset flag FDC
	lda	#$40
	tsb	fdcctl		; force restore
	stz	fdcdrv		; invalidate drive
	lda	fdcctl
	and	#00010000B	; mask on motor bit	
	ora	#00001000B	; enable FDC interrupt&dma, reset FDC 
	sta	!FDCDOR		; reset controller
	nop
	nop
	nop
	nop
	bit	VIA0+VIAPRB	; reset flag CB1
	nop
	nop
	nop
	nop
	ora	#00001100B	; enable controller
	sta	!FDCDOR		; this write cause interrupt from FDC
	plp			; restore I flag
	jsr	waitint		; wait interrupt within timeout
	
	; controller reply to sense interrupt command
	; with 4 status bytes: $C0, $C1, $C2, $C3
	
	ldy	#$C0		; first expected status
?01:	lda	#CMDSENSEINT
	jsr	sendto		; sense interrupt
	jsr	waitres		; result phase (normally 2 bytes)
	cpy	fdcres		; check first byte result
	beq	?03		; ok, as expected
	sec			; controller error
	lda	#FDC_BAD	; bad controller error
	rts
?03:	iny
	cpy	#$C4		; last expected result is $C3
	bcc	?01		; loop
	lda	#$20
	tsb	fdcctl		; flag controller FDC OK
	lda	fdcctl
	jsr	setrate		; select current rate
	lda	#CMDSPECIFY
	jsr	sendto		; setup standard controller
	lda	#UM_SRTHUT
	jsr	sendto
	lda	#UM_HLTND	; send command to FDC

; like sendto but always return to the caller, even if error
sndto2:
	jsr	sendto
	rts

; send command to controller within SNDTMO timeout
; in:	A=command
; out:	CF=1 if error, A=error code
; A,X,Y preserved (if error A will be destroyed)
; if error, return address from stack will be pulled out and discarded
sendto:
	sta	tmpa
	jsr	wait16us	; 16uS delay safety
	lda	#$20		; T2 count PB6 pulses (1ms)
	tsb	VIA0+VIAACR
	lda	#<SNDTMO	; timeout send byte
	sta	!VIA0+VIAT2CL
	lda	#>SNDTMO
	sta	!VIA0+VIAT2CH
?02:	lda	!FDCMSR		; read FDC status
	and	#$C0		; mask bit 7,6
	cmp	#$80		; bit 6 will be 0
	beq	?03		; FDCcan accept data within timeout
	lda	#T2IFRB
	bit	VIA0+VIAIFR	; test timeout
	beq	?02		; no timeout	
	pla			; pull out return address...
	pla			; ...and discard it
	lda	#FDC_SNDTOUT	; timeout error
	sec			; error
	rts
?03:	jsr	wait16us	; 16uS delay before to access data register
	lda	tmpa
	sta	!FDCDATA	; send byte to controller
	clc
	rts
	
; wait result phase from FDC within RESTMO timeout
; out:	CF=1 if result phase error, A=error code
; A destroyed, X&Y preserved
; if error, return address from stack will be pulled out and discarded
waitres:
	stx	tmpx		; save X
	jsr	wait16us	; safety delay of 16uS
	ldx	#0		; index for store 7 bytes of result
	lda	#$20		; T2 count PB6 pulses (1ms)
	tsb	VIA0+VIAACR
?00:	lda	#<RESTMO	; setup result phase timeout
	sta	!VIA0+VIAT2CL
	lda	#>RESTMO
	sta	!VIA0+VIAT2CH
?01:	lda	!FDCMSR		; read status reg. of FDC
	and	#$C0		; mask bit 7,6
	cmp	#$C0		; bit 6 will be 1 when CPU can read FDC
	beq	?03		; data ready within timeout
	lda	#T2IFRB
	bit	VIA0+VIAIFR	; test timeout
	beq	?01		; no timeout
	lda	#FDC_RESTOUT	; set timeout result phase error
?02:	plx			; pull out return address...
	plx			; ...and discard it
	ldx	tmpx		; restore X
	sec			; error
	rts			; exit
?03:	jsr	wait16us	; 16uS delay before to read data register
	lda	!FDCDATA	; read FDC data register 
	sta	fdcres,x	; store byte
	jsr	wait16us	; 16uS delay before to read status register
	lda	!FDCMSR		; test bit 4 (busy)
	bit	#00010000B
	beq	?05		; no more bytes
	inx
	cpx	#7		; max. 7 bytes in result phase
	bcc	?00		; loop next byte
	lda	#FDC_BADRES	; too many bytes	
	bne	?02		; error
?05:	ldx	tmpx	
	clc			; OK
	rts

; wait interrupt from UM8388 within INTTMO timeout
; out:	CF=1 if timeout error, A=error code
; A destroyed, X&Y preserved
; if error, return address from stack will be pulled out and discarded
waitint:
	lda	#$20		; T2 count PB6 pulses (1ms)
	tsb	VIA0+VIAACR
	lda	#<INTTMO	; interrupt timeout
	sta	!VIA0+VIAT2CL
	lda	#>INTTMO
	sta	!VIA0+VIAT2CH
?02:	lda	#CB1IFRB	; CB1 -> positive edge INT from FDC	
	bit	VIA0+VIAIFR
	bne	?04		; INT within timeout
	lda	#T2IFRB		; test timeout
	bit	VIA0+VIAIFR
	beq	?02		; no timeout
	sec			; error
	pla			; pull out return address...
	pla			; ...and discard it
	lda	#FDC_INTTOUT	; set timeout error	
	rts
?04:	bit	VIA0+VIAPRB	; clear flag CB1
	lda	#$02		; enable CPU bus
	tsb	VIA0+VIAPRANH	; PA1 -> /DME = 1
	lda	#$08
	trb	fdcctl		; reset dma active bit
	clc			; no error
	rts

; setup clock rate in controller
; in:	A=$00(DD),=$02(HD)
; A destroyed, X&Y preserved
setrate:
	and	#00000010B	; mask bit 1
	beq	?02		; select DD
	tsb	fdcctl		; select HD
	lda	#0		; rate 500Kb/s
	beq	?04		; set rate
?02:	lda	#00000010B	; rate 250Kb/s
	trb	fdcctl			
?04:	sta	!FDCCCR		; set rate
	rts

; enable motor and wait for spin-up timer expiration
; A destroyed, X&Y preserved
startmtr:
	sei			; disable IRQ
	lda	#00110000B	; CNT0 mode 0 - 2 bytes
	sta	!CTC0+CTCCTRL	; CTC 0 stopped
	lda	#CB2IFRB	; spin-down timer stopped 
	sta	VIA0+VIAIER	; clear IRQ CB2 (TMF0) from CNT0 CTC
	sta	VIA0+VIAIFR	; clear flag CB2
	lda	#00010000B	; motor already on?
	bit	fdcctl
	beq	?04		; no, put on
	cli			; yes - exit
	rts
?04:	tsb	fdcctl		; flag motor on
	ora	#00001100B	; enable controller,IRQ,DMA,drive0
	sta	!FDCDOR		; select drive and motor
	cli
	lda	#$20		; T2 count PB6 pulses (1ms)
	tsb	VIA0+VIAACR
	lda	#<MOTACL	; spin-up delay
	sta	!VIA0+VIAT2CL
	lda	#>MOTACL
	sta	!VIA0+VIAT2CH
?06:	TASKSW			; task switch while wait timer expiration	
	lda	!VIA0+VIAIFR	; loop spin-up
	bit	#T2IFRB
	beq	?06
	rts
	
; set up motor spin-down timer
; this routine preserve CF
; A&X&Y preserved
stopmtr:
	sei			; disable IRQ
	lda	fdcctl
	bit	#00010000B	; motor bit
	beq	?10		; motor already off		
	; motor on -> setup spin-down timer
	lda	#00110000B	; CNT0 mode 0 - 2 bytes
	sta	!CTC0+CTCCTRL	; CTC 0 stopped
	lda	#CB2IFRB
	sta	!VIA0+VIAIER	; clear IRQ CB2 (TMF0) from CNT0
	sta	!VIA0+VIAIFR	; clear flag CB2
	lda	#<MOTDCL	; setup spin-down timer
	sta	!CTC0		; CNT0
	lda	#>MOTDCL
	sta	!CTC0	
	lda	#(CB2IFRB.OR.$80) 
	sta	!VIA0+VIAIER	; enable interrupt on CB2
?10:	cli
	rts

; move dma buffer to safe area while verify data
savebuf:
	ldx	.ABS.CRFWEOFF
	sta	!CRFWEON	; enable write in bank $F8 RAM
	lda	verfp+2		; the source bank
	sta	?10+2		; store source bank for mvn istruction
	sta	!CRFWEOFF,x	; restore writing in bank $F8
	phb			; save data bank
	CPU16
	lda	ncnt
	dec	a		; count #
	ldx	verfp		; source address
	ldy	#!DMASAVE	; dest address
?10:	mvn	#0,#1		; move to bank 1
	CPU08	
	plb			; restore data bank
	rts

; restore dma buffer from safe area
restbuf:
	php			; save carry
	pha			; save error code if any
	ldx	.ABS.CRFWEOFF	
	sta	!CRFWEON	; enable write in bank $F8 RAM
	lda	verfp+2		; the source bank
	sta	?10+1		; store dest bank for mvn istruction
	sta	!CRFWEOFF,x	; restore writing in bank $F8
	phb			; save data bank
	CPU16
	lda	ncnt
	dec	a		; count #
	ldx	#!DMASAVE	; source address
	ldy	verfp		; dest address
?10:	mvn	#1,#0		; move from bank 1
	CPU08	
	plb			; restore data bank
	pla
	plp
	rts

; check params for read/write
; out:	CF=1 if error, A=error code
chkparms:
	stz	bufp		; always aligned at 512 bytes
	stz	verfp
	lda	#$FF
	sta	dtlstp		; standard for disk r/w
	ldx	fdop		; operation index
	lda	>VDOPTAB,x
	sta	vdop		; ram disk operation
	bmi	?02		; write operation
	lda	#$80
	trb	opfg		; no verify for read
?02:	ldx	dskfmt
	lda	>MAXSECT,x
	sta	maxsect		; max. number of sectors/track + 1
	lda	opfg		; test required format
	eor	dskfmt		; compare bit 0
	and	#$01		; they will be the same
	bne	?20		; mismatch format
	bit	opfg		; test MT operation
	bvc	?04		; no MT
	stz	head		; MT -> start head = 0
	ldx	#1		; MT -> start sector = 1
	stx	sector
	ldx	maxsect		; MT -> SectNum = max. sectors
	dex
	stx	sectnum
?04:	lda	sector		; start sector will be < MaxSect
	cmp	maxsect
	bcs	?22
	lda	sectnum		; here CF = 0
	tay
	dec	a		; SectNum - 1
	adc	sector		; EOT = end of track
	bcs	?22		; overflow
	cmp	maxsect
	bcs	?22		; EOT > (MaxSect - 1)
	sta	eot		; end of track
	tya			; number of sectors
	bit	opfg
	bvc	?08		; no MT
	asl	a		; MT -> double number of sectors
	tay
?08:	asl	a		; number of pages of 256 bytes - here CF = 0
	sta	ncnt+1		; number of bytes to be trasferred
	stz	ncnt
	sty	bufcnt		; number of required dma buffers
	lda	dmabuf		; test buffer dma
	tay
	and	#$7F		; mask off bit 7
	tax
	adc	bufcnt		; here never carry
	cmp	#$81		; over the limit of the bank?
	bcs	?24		; yes
	txa
	asl	a		; high byte of address of the buffer
	sta	bufp+1		; verfp and bufp equal but in opposite bank
	sta	verfp+1
	ldx	#BUFDSK1	; bank of buffer #0..#127
	tya
	bpl	?12
	inx			; bank of buffer #128..#255
?12:	stx	bufp+2
	ldy	#BUFDSK1	; bank of verify buffer
	cpx	#BUFDSK2	; always different from other buffer
	beq	?16
	iny
?16:	sty	verfp+2
	clc
	rts
?20:	lda	#FDC_FMT	; format mismatch
	sec
	rts
?22:	lda	#FDC_PARAMS	; invalid params
	sec
	rts
?24:	lda	#FDC_BADBUF	; invalid buffer
	sec
	rts

; 16us delay for safe access to FDC register's
wait16us:

.IF	PHI2 = 4
??CNT	.EQU	9
.ELSE
??CNT	.EQU	5
.ENDIF

; #cycles = 5*N + 13 + M(2 * nop) = 64/32 cycles
; 4MHz -> N = 9, M = 6
; 2MHz -> N = 5, M = 4

	lda	#??CNT
?01:	dec	a
	bne	?01
.IF	PHI2 = 4
	nop
.ENDIF
	nop
	nop
	rts

;---------------------------------------------------------------------------
; ram disk managment
;---------------------------------------------------------------------------

; ram disk format whole disk (X = drive)
; always return CF=0
; any time ram disk is formatted, change disk bit (bit 6) is set
vdfmt:
	lda	vdrive
	bit	#$10		; ram disk write protect on?
	beq	?01		; no
	lda	#FDC_WP		; wp on error
	sta	fdcerr
	sec
	rts
?01:	and	#01111100B	; reset format
	sta	vdrive
	lda	#$01		; reset bit 0
	trb	fmtfg		; for first call to callback
	lda	#XMBANK		; bank that old X mem
	sta	bufp+2
	ldx	CRXME		; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	ldy	#0		; start track
	tyx			; start head
?02:	sty	track		; format this track...
?04:	stx	head		; ...in this side
	jsr	getlba		; set ram disk address
	sta	!PIA0+PIAPRA	; set ram disk bank
	bit	fmtfg
	bpl	?06		; no callback
	phk			; JSL simulation
	pea	#?06-1		; return address
	lda	eot		; params for callback
	xba			; B = sector/track
	lda	fmtfg
	lsr	a		; bit 0 -> carry -> 0 if first call
	lda	#1		; A = drive
	ldy	track		; Y = track
	ldx	head		; X = head
	jmp	[lpfn]
?06:	; return from callback
	ldx	sectnum		; number of sector per track
	INDEX16
	ldy	xmstart		; start address of X mem
?08:	sty	bufp		; sector's loop
	lda	fmtfill		; fill byte
	ldy	#$01FF		; size of sector
?10:	sta	[bufp],y	; fill one sector
	dey
	bpl	?10
	dex
	beq	?12		; end sector's loop
	CPU16CLC
	lda	bufp
	adc	#$0200		; next sector pointer
	tay
	ACC08
	cpy	#XMSTOP		; end of X mem
	bcc	?08		; fill next block
	inc	!PIA0+PIAPRA	; increment X mem bank
	ldy	#XMSTART	; start address of X mem
	bra	?08		; fill next block
?12:	INDEX08	
	lda	#$01
	tsb	fmtfg		; bit 0 = 1 -> for callback
	ldx	head		; next side
	inx
	cpx	#2
	bcc	?04		; same track, next side
	ldx	#0		; clear head
	ldy	track		; next track
	iny
	cpy	#80
	bcc	?02		; next track, side=0
	lda	vdrive		; set valid format
	and	#01111100B
	ora	dskfmt		; update format in selected drive
	ora	#$C0		; current valid format
	sta	vdrive		; set bit 6: change disk
	plx			; restore X mem setting
	sta	!CRXME,x
	ldx	#RTCRDFG	; validate ram disk configuration
	stx	RTCALE
	lda	#$55
	sta	RTCDATA
	inx			; save configuration in RTC ram
	stx	RTCALE
	lda	vdrive
	sta	RTCDATA		; save ram disk
	clc
	rts

; execute read/write on ram disk
; in:	vdop=command ($00=read, $80=write, $40=fake verify)
;	opfg<7>=1 if read verify buffer
; out:	CF=1 if error (just verify error in ram disk)
; A,X,Y destroyed
vdrw:
	jsr	chkparms	; check params
	bcc	?02		; ok
?01:	sta	fdcerr
	rts			; return error
?02:	bit	vdop		; write operation?
	bpl	?04		; no
	lda	#$10
	bit	vdrive		; write protect?
	beq	?04		; no
	lda	#FDC_WP		; yes...
	sec			; ...error
	bra	?01
?04:	bit	opfg		; verify buffer?
	bpl	?06		; no
	jsr	savebuf		; save verify buffer		
?06:	jsr	getlba		; set ram disk address
	sta	!PIA0+PIAPRA	; set ram disk bank
	stz	buftst		; now using bufp buffer
	jsr	vdcmd		; execute command (here always CF = 0)
	lda 	opfg
	bpl	?20		; no verify, exit	
	sta	buftst		; now using verfp buffer
	stz	vdop		; read for verify
	lda	xmbank
	sta	!PIA0+PIAPRA	; set ram disk bank
	jsr	vdcmd		; execute command
	ldy	#0		; verify data
	INDEX16
?08:	lda	[bufp],y
	cmp	[verfp],y
	bne	?10		; unmatch: ZF=0
	iny
	cpy	ncnt
	bcc	?08		; when finish ZF=1
?10:	INDEX08			; here ZF=1 if match
	clc			; assume no error
	beq	?12
	lda	#FDC_VER	; verify error
	sta	fdcerr
?12:	jsr	restbuf		; restore verify buffer
?20:	rts

; ram disk read/write command
; in:	vdop=command ($00=read, $80=write, $40=fake verify)
;	buftst=0 if using bufp, =$80 if using verfp
; out:	CF=0 always
; A,X,Y destroyed
vdcmd:
	ldx	.ABS.CRXME	; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	ldx	.ABS.CRFWEOFF
	phx
	sta	!CRFWEON	; enable write in bank $F8 RAM
	INDEX16
	ldy	ncnt
	bit	vdop		; fake verify operation?
	bvs	?24		; yes, no operation in ram disk
	ldx	bufp		; buffer start for read/write
	lda	bufp+2
	bit	buftst		; read verify buffer?
	bpl	?02		; no, so use bufp buffer
	ldx	verfp		; yes, so use verfp verify buffer
	lda	verfp+2
?02:	stx	bufstart	; set start buffer for ram disk access
	sta	tmpa		; source or dest bank		
	ACC16CLC		; calc end of transfer buffer
	tya			; count #
	adc	bufstart
	sta	bufend		; end address of transfer
	ACC08
	lda	#$FF
	xba			; B = $FF
	bit	vdop		; check read/write flag
	bpl	?14		; read operation
	; write operation
	lda	tmpa		; the source bank
	sta	?10+2		; store source bank for mvn istruction
	ldx	bufstart	; start address of source block
	phb			; save current danta bank
	ldy	xmstart		; start address of current h&t&&s
?08:	lda	#$01		; size of block - 1 (512 bytes) 	
	xba			; before B = $FF, A = $01 -> C = $01FF
?10:	mvn	#0,#1		; move sector from bank 2 or 3 to bank 1
	cpx	bufend		; X point past the last moved byte
	bcs	?22		; end of move
	cpy	#XMSTOP		; end of X mem?
	bcc	?08		; no, move next block
	plb			; restore data bank
	inc	PIA0+PIAPRA	; increment X mem bank
	phb			; save current danta bank
	ldy	#XMSTART	; start address from beginning of X mem
	bra	?08		; move next block
?14:	; read operation
	lda	tmpa		; dest bank
	ldy	bufstart	; start address of dest. block
	sta	?20+1		; store dest. bank for mvn istruction
	phb			; save current danta bank
	ldx	xmstart		; start address of current h&t&s
?18:	lda	#$01		; size of block - 1 (512 bytes) 	
	xba			; before B = $FF, A = $01 -> C = $01FF
?20:	mvn	#1,#0		; move sector from bank 1 to bank 2 or 3
	cpy	bufend		; Y point past the last moved byte
	bcs	?22		; end of move
	cpx	#XMSTOP		; end of X mem?
	bcc	?18		; no, move next block
	plb			; restore data bank
	inc	PIA0+PIAPRA	; increment X mem bank
	phb
	ldx	#XMSTART	; start address from beginning of X mem
	bra	?18		; move next block
?22:	plb			; restore data bank
	ldy	ncnt
?24:	sty	ntrsf		; store numbers of moved bytes
	INDEX08
	plx
	sta	!CRFWEOFF,x	; restore writing in bank $F8
	plx			; restore X mem setting
	sta	!CRXME,x
	clc
	rts
	
; translate the tuplet (T,H,S) into LBA (Linear Block Address)
; LBA = [(2*track + head)]*P + sector - 1
; where P = sectors/track
getlba:
	jsr	lbabase		; get LBA base
	ldx	sector
	stx	mcand1		; store S in word
	stz	mcand2
	ACC16CLC		; A/M 16 bit + CLC
	adc	mcand1
	dec	a		; LBA
	ror	a		; divide by 16 to get xmem bank
	ror	a
	ror	a
	ror	a
	ACC08
	sta	xmbank		; ram disk bank
	xba			; B = bank, A = modulus
	ror	a		; shift in last carry
	lsr	a		; calc the modulo
	lsr	a
	lsr	a		; MSB of xmem
	clc
	adc	#>XMSTART	; add the physical start of xmem
	sta	xmstart+1
	stz	xmstart
	xba			; return in A the ram disk bank
	rts
	
; get base of Linear Block Address (LBA)
; LBA = [(2*track + head)]*P + sector - 1
; where P = sectors/track
; this function calc the LBA base = [(2*track + head)]*P
; the result is stored in C
lbabase:
	lda	maxsect
	dec	a
	sta	mcand1		; multiplicand #1
	lda	track
	asl	a		; 2*track (here always CF = 0)
	adc	head
	sta	mcand2		; 2*track + head
	lda	#0		; MSB result of (macnd1 * mcand2)
	ldx	#9		; loop 9 bit
	clc
?02:	ror	a
	ror	mcand1		; LSB result of multiplication
	bcc	?04
	clc
	adc	mcand2
?04:	dex
	bne	?02
	xba			; B = MSB of result
	lda	mcand1		; A = LSB of result
	rts			; C = 16 bit result

;---------------------------------------------------------------------------
; POST routine
;---------------------------------------------------------------------------

; this routine is invoked within boot phase
fdinit:
	lda	#11111011B
	trb	fdcctl		; reset all bit's but not dma flag
	lda	#00000010B
	tsb	fdcctl		; select HD rte at beginning
	jsr	startmtr	; start motor
	jsr	fdcrst		; reset FDC controller
	lda	#$20		; test bit 5
	bit	fdcctl
	beq	?100		; FDC controller fault
	jsr	?110
	stz	fdcerr
	;jsr	startmtr	; start motor
	jsr	seek0		; recalibrate after reset and seek track 0
	bcs	?30		; fail
	lda	#$80		; set flag drive ok
	tsb	fdcctl
	
	SCNPRINT
	.DB	'pass.', $0D, $00
	
	lda	!FDCDIR		; read disk change line
	and	#$80		; bit 7 will be 0 if disk ready
	bne	?20		; not ready
	jsr	chkwp		; aquire wp status
?20:	clc
	bra	?32
?30:	SCNPRINT
	.DB	'fail.', $0D, $00
	sec
	sta	fdcerr
?32:	jmp	stopmtr
		
?100:	SCNPRINT
	.DB	'ERROR: diskette controller UM8388 fail', $0D, $00
	TASKSW
	jmp	stopmtr		; exit
	
?110:	SCNPRINT
	.DB	'drive #0...', $00
	rts
	
; POST routine for initial ram disk configuration
vdinit:
	jsr	?100
	lda	#$40
	sta	vdrive		; reset unformatted ram disk (disk changed)
	jsr	?110
	bit	Bnk0Flag	; <6> -> warm reset flag
	bvc	?10		; cold reset - unformatted
	ldx	#RTCRDFG	; resume saved configuration
	stx	RTCALE		; from RTC RAM in bank 0 & 1
	lda	RTCDATA		; valid ram disk configuration?
	cmp	#$55
	bne	?10		; no -- make default ram disk
	inx
	stx	RTCALE
	lda	RTCDATA		; get ram disk 0 configuration
	ora	#$40		; always disk ghanged at startup
	sta	vdrive
?10:	rts

?100:	SCNPRINT
	.DB	'drive #1...', $00
	rts

?110:	SCNPRINT
	.DB	'virtual drive (ram disk).', $0D, $00
	rts

;---------------------------------------------------------------------------
; Tables
;---------------------------------------------------------------------------

; ctl function's table
CTLFN:
	.DW	ctl_fn0, ctl_fn1, ctl_fn2, ctl_fn3, ctl_fn4
	.DW	ctl_fn5, ctl_fn6

MXCTLFN		.EQU	($-CTLFN)/2

; max. number of secors/track
MAXSECT
	.DB	10, 11, 19, 21

; GAP 3 lenght for rd/wr command
RWGPLTBL:
	.DB	RWGPL3DD, RWGPL3DDA, RWGPL3HD, RWGPL3HDA

; GAP 3 lenght while format track
FGPLTBL:
	.DB	FMGPL3DD, FMGPL3DDA, FMGPL3HD, FMGPL3HDA
	
; head inversion for MSDOS/CBM format
HEADTBL:
	.DB	$00, $01, $00, $01

; commands for UM8388
CMDTAB:
	.DB	CMDREAD, CMDREAD, CMDWRITE
	
; transfer command for DMA
DMATAB:
	.DB	WRITETRASF, VERFTRASF, READTRASF
	
; disk operation flag (read, verify, write)
VDOPTAB:
	.DB	$00, $40, $80

;---------------------------------------------------------------------------
; end of file
;---------------------------------------------------------------------------
