;;
;; Copyright (c) 2016 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; bios 65C816 version v 1.0

;; rtc.asm
;; rev.: 2016/03/21
;; rtc handler

; include RTC definition
_RTC_INC_	.SET	1

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc

.LIST on

MAXFNC		.EQU	4

; page 0 local var's (declared in bios temp. work area)
_P0BTMP:	.SECTION page0, ref_only, offset bwrktmp	; DP Tmp. BIOS

dmonth		.DS	12	; array of month's day's
cyear		.DW		; full year - 1
ty400		.DW		; full year / 400
ty100		.DW		; full year / 100
dvsor		.DW		; divisor/remainder
quot		.DW		; quotient
date1		.DW		; last march sunday
date2		.DW		; last october sunday
dayofw		.DB		; day of week (1=sunday...)
date		.DB		; date (day of month)
month		.DB
year		.DB
ctry		.DB		; century
jan1		.DB		; first day of the year

.ENDS

	.CODEF8

	.LONGA	off
	.LONGI	off

;---------------------------------------------------------------------------
; system call interface
;---------------------------------------------------------------------------

; handle real time clock
;
; functions:
;
;		$00: set time (X=seconds, Y=minutes, A=hours)
;		$01: set date (X=day, Y=month, A=year, B=century)
;		?02: get time (X=seconds, Y=minutes, A=hours, B=day of week)
;		$03: get date (X=day, Y=month, A=year, B=century)
;
sys_rtc:
	
	.PUBLIC sys_rtc

	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2
	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	lda	[COPPtr]	; byte after signature: function number #
	tax			; X=function number
	ACC08			; A,M -> 8 bit
	cpx	#MAXFNC
	bcs	?10		; invalid function
	txa
	asl	a
	tax			; function index
	jmp	(JMPTBL,x)
?10:	lda	#RTC_BADFNC	
	sec
	rtl

; set rtc time
rtcsett:
	lda	STKXR,s		; X=seconds
	tax
	lda	STKYR,s		; Y=minutes
	tay
	lda	STKCR,s		; A=hours
	jsr	chktim		; check if valid time
	bcc	?10
	lda	#RTC_BADTIME	; return error
	sec
	rtl
?10:	xba			; B=hours	
	lda	#RTCCTRLB	; stop time update
	sta	RTCALE
	lda	#$80
	tsb	RTCDATA
	stz	RTCALE		; seconds register
	stx	RTCDATA		; set seconds
	lda	#RTCMIN
	sta	RTCALE
	sty	RTCDATA		; set minutes
	lda	#RTCHOURS
	sta	RTCALE
	xba
	sta	RTCDATA		; set hours
	lda	#RTCCTRLB	; start time update
	sta	RTCALE
	lda	#$80
	trb	RTCDATA
	clc
	rtl

; get rtc time
rtcgett:
	sei
	ldx	#RTCCTRLA	; set bank 1
	stx	RTCALE
	lda	#$10
	tsb	RTCDATA
	ldx	#RTCEXTCTRLA	; check INCR bit 6
	stx	RTCALE
?02:	bit	RTCDATA
	bvs	?02
	ldx	#0		; seconds
	stx	RTCALE
	lda	RTCDATA
	sta	STKXR,s
	inx
	inx
	stx	RTCALE		; minutes
	lda	RTCDATA
	sta	STKYR,s
	inx
	inx
	stx	RTCALE		; hours
	lda	RTCDATA
	sta	STKCR,s
	inx
	inx
	stx	RTCALE		; day of week
	lda	RTCDATA
	sta	STKCR+1,s	
	cli
	clc
	rtl

; get rtc date
rtcgetd:
	sei
	ldx	#RTCCTRLA	; set bank 1
	stx	RTCALE
	lda	#$10
	tsb	RTCDATA
	ldx	#RTCEXTCTRLA	; check INCR bit 6
	stx	RTCALE
?02:	bit	RTCDATA
	bvs	?02
	ldx	#RTCDDATE	; day
	stx	RTCALE
	lda	RTCDATA
	sta	STKXR,s
	inx
	stx	RTCALE		; month
	lda	RTCDATA
	sta	STKYR,s
	inx
	stx	RTCALE		; year
	lda	RTCDATA
	sta	STKCR,s
	ldx	#RTCCENTURY
	stx	RTCALE		; century
	lda	RTCDATA
	sta	STKCR+1,s
	cli
	clc
	rtl

rtctest:
	.PUBLIC rtctest
	phd
	pea	#0
	pld
	jsr	chkdate
	bcc	?02
	lda	#RTC_BADDATE	; return error
	sec
	pld
	brk
?02:	jsr	getdayw		; get day of week
	jsr	getlmt
	ldx	date1
	ldy	date2
	clc
	pld
	brk

rtcsetd:
	lda	STKXR,s		; X=date (day of month)
	tax
	lda	STKYR,s		; Y=month
	tay
	lda	STKCR+1,s	; B=century
	xba
	lda	STKCR,s		; A=year	
	jsr	chkdate
	bcc	?02
	lda	#RTC_BADDATE	; return error
	sec
	rtl
?02:	jsr	getdayw		; get day of week		
	sei
	ldx	#RTCCTRLA	; set bank 1
	stx	RTCALE
	lda	#$10
	tsb	RTCDATA
	lda	#RTCCTRLB	; stop time update
	sta	RTCALE
	lda	#$80
	tsb	RTCDATA
	ldx	#RTCDAY		; day of week
	stx	RTCALE
	lda	dayofw
	sta	RTCDATA
	inx
	stx	RTCALE		; date (day of week)
	lda	date
	sta	RTCDATA
	inx
	stx	RTCALE		; month
	lda	month
	sta	RTCDATA
	inx
	stx	RTCALE		; year
	lda	year
	sta	RTCDATA
	ldx	#RTCCENTURY
	stx	RTCALE		; century
	lda	ctry
	sta	RTCDATA
	cli
	clc
	rtl

; check time validity
chktim:
	cpx	#60
	bcs	?10
	cpy	#60
	bcs	?10
	cmp	#24
?10:	rts			; return CF=1 if invalid time

; check date validity
chkdate:
	cpy	#1		; check month 1..12
	bcs	?04
?02:	sec
?03:	rts
?04:	cpy	#13
	bcs	?03
	sty	month
	xba			; check century
	cmp	#100
	bcs	?03
	sta	ctry
	xba
	cmp	#100		; check year
	bcs	?03
	sta	year
	cpx	#1		; check date (day of month)
	bcc	?02
	stx	date
	ldx	#11		; build month's array
?05:	lda	>MDAYTBL,x	; max. days
	sta	dmonth,x
	dex
	bpl	?05
	lda	year		; century start?
	bne	?06		; no...so check if year divisible by 4
	lda	ctry		; check if century is divisible by 4
?06:	and	#$03		; year/century divisible by 4?
	bne	?08		; no...so no leap year
	inc	dmonth+1	; leap year so february have 29 days
?08:	dey			; month zero based
	ldx	dmonth,y
	cpx	date
	bcc	?02		; over month limit
	clc
	rts

getdayw:
	jsr	fyear
	CPU16
	lda	#400
	sta	dvsor
	lda	cyear
	jsr	fudiv		; year / 400
	sta	ty400
	lda	#100
	sta	dvsor
	lda	cyear
	jsr	fudiv		; year / 100
	sta	ty100
	lda	cyear
	lsr	a		; year / 4
	lsr	a
	clc
	adc	#35
	clc
	adc	cyear		; add year
	clc
	adc	ty400		; add (year / 400)
	sec
	sbc	ty100		; subtract (year / 100)
	ldx	#7
	stx	dvsor
	jsr	fudiv		; divide by 7...remainder is the first day
	CPU08
	tya			; first day of the year
	sta	jan1		; save for later use
	clc
	adc	date		; add current date
	dec	a		; day number (0 based)
	ldy	#0		; hi byte
	ldx	month		; add days
	dex
	beq	?10		; january...so nothing to add
	dex
?06:	clc	
	adc	dmonth,x	; add days till previous month
	bcc	?08
	iny
?08:	dex
	bpl	?06
?10:	xba
	tya
	xba
	CPU16
	ldx	#7
	stx	dvsor
	jsr	fudiv		; Y -> 0 = monday .. 6 = sunday
	CPU08
	iny
	cpy	#7		; sunday ?
	bne	?15		; no
	ldy	#0		; yes
?15:	iny
	sty	dayofw		; 1=sunday..7=saturday
	rts

getlmt:
	lda	#0
	xba
	lda	jan1
	clc
	adc	dmonth		; get march, 31
	adc	dmonth+1
	adc	dmonth+2
	dec	a
	CPU16
	ldx	#7
	stx	dvsor
	jsr	fudiv		; Y -> 0 = monday .. 6 = sunday
	CPU08
	lda	#31
	cpy	#6		; sunday?
	beq	?10		; yes, found first limit
?05:	dec	a		; back one day
	dey
	bpl	?05		; till previous monday
?10:	sta	date1		; last march sunday
	lda	#3
	sta	date1+1
	lda	#0
	xba
	lda	dmonth+1	; get october, 31
	clc
	adc	jan1
	CPU16CLC
	adc	#276
	dec	a
	ldx	#7
	stx	dvsor
	jsr	fudiv		; Y -> 0 = monday .. 6 = sunday
	CPU08
	lda	#31
	cpy	#6		; sunday?
	beq	?20		; yes, found first limit
?15:	dec	a		; back one day
	dey
	bpl	?15		; till previous monday
?20:	sta	date2		; last october sunday
	lda	#10
	sta	date2+1
	ACC16
	ldx	#0		; not in DST range
	lda	date		; check if date is the DST range
	cmp	date1
	bcc	?25		; not in DST
	cmp	date2
	bcs	?25		; not in DST
	dex			; DST time
?25:	;stx	dstfg
	ACC08
	rts
	
; get full year - 1
fyear:
	lda	year
	sta	ty100
	stz	ty100+1
	lda	ctry	
	sta	cyear
	stz	cyear+1	
	ACC16
	lda	cyear
	asl	a		; x 4
	asl	a
	adc	cyear		; x 5
	asl	a		; x 10
	sta	cyear
	asl	a		; x 40
	asl	a
	adc	cyear		; x 50
	asl	a		; x 100
	adc	ty100
	dec	a
	sta	cyear
	ACC08
	rts

; fast unsigned division 16 bit
;
;	entry:	C = 16 bit dividend
;		dvsor = 16 bit divisor
;
;	exit:	C = 16 bit quotient
;		Y = dvsor = 16 bit remainder
;
;	use:	all
;
;	note:	no check for null divisor
;		must be called with cpu in 16 bit mode
;
;-----
fudiv:
;-----
	.LONGA	on
	.LONGI	on

	ldx	#1		; bit counter
	tay			; Y=dividend
	stz	quot		; init quotient
	lda	dvsor		; C=divisor
?10:	asl	a		; shift divisor: get leftmost bit
	bcs	?20		; go to division
	inx
	cpx	#17		; test all divisor bit's
	bne	?10
?20:	ror	a		; put shifted-out bit back
	sta	dvsor
?30:	tya			; get dividend
	sec
	sbc	dvsor
	bcc	?40		; can't subctract, retain old dividend
	tay			; Y=new dividend
?40:	rol	quot		; shift carry into quotient (1 if division)
	lsr	dvsor		; shift right divisor for next subtract
	dex
	bne	?30
	sty	dvsor		; remainder
	lda	quot		; C=quotient
	rts
	
;---------------------------------------------------------------------------
; tables
;---------------------------------------------------------------------------

; rtc function's jump table
JMPTBL:
	.DW	rtcsett, rtcsetd, rtcgett, rtcgetd

; month's table
MDAYTBL:
	.DB	31,28,31,30,31,30,31,31,30,31,30,31
