; START.ASM
; PROGETTO: B1601
; Compilatore Avocet X65816

; Codice area F8FE00 - F8FFFF rimappato da banco 0

_START_INC_	.SET	1
_KFR_INC_	.SET	1
_RTC_INC_	.SET	1
_DMA_INC_	.SET	1
_ACIA_INC_	.SET	1
_KBD_INC_	.SET	1


	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\dirp01.inc

	;.JMPTF8

	.LIST on

	.CODEF8

;----------------------------------------------------------
; ---- CODICE START & INTERRUPT
;----------------------------------------------------------

	.SYSCODE
	
_KbdCnt		.EQU	(DP01ADDR + KbdCnt)
_KbdToggle	.EQU	(DP01ADDR + KbdToggle)
_KbdSt		.EQU	(DP01ADDR + KbdSt)
_PS2Ctl		.EQU	(DP01ADDR + PS2Ctl)
_KbdITail	.EQU	(DP01ADDR + KbdITail)
_KbdIHead	.EQU	(DP01ADDR + KbdIHead)
_KbdShift	.EQU	(DP01ADDR + KbdShift)
_KbdFlag	.EQU	(DP01ADDR + KbdFlag)
	
	LONG_OFF
	
	.GLOBAL _KbdCnt, _KbdToggle, _KbdSt, _PS2Ctl, _KbdITail
	.GLOBAL _KbdIHead, _KbdShift, _KbdFlag
	.GLOBAL TORAM

TORAM:	

int0sr:
	SPISR	0

int1sr:	
	SPISR	1

	
; gestione IRQ2 - VIA 0 (T1, CA2, CB2, CA1)	
int2sr:
	ldy	VIA0+VIAIFR	; flag IFR VIA
	bpl	?50		; no irq da VIA
	tya			; IFR
	asl	a		; N = bit 6 (T1IFRB)
	bpl	?20		; no T1 IRQ
	tay
	ACC16			; A,M -> 16 bit
	inc	JiffyClk	; incrementa contatore clock 10ms
	bne	?02
	inc	JiffyClk+2
?02:	ACC08			; A,M -> 8 bit
	ldx	#SYSTMRCNT-1	; System Timer (max. 2560ms)
?04:	bit	SysTMF,x	; timer attivo ?
	bpl	?05		; no
	dec	SysTmr,x	; update timer x
	bne	?05
	stz	SysTMF,x	; cancella timer
?05:	dex
	bpl	?04
	lda	VIA0+VIAT1CL	; clear flag T1
?20:	tya			; IFR
	lsr	a		; C  = CA2IFRB
	bcc	?30		; no IRQ CA2 da controller tastiera
	jsr	KbInt		; gestione IRQ tastiera	
	lda	#CA2IFRB	; azzera flag CA2
	sta	VIA0+VIAIFR
	sta	KBCLRIRQ	; azzera linea INT
?30:	tya			; IFR
	and	#CB2IFRB	; CB2-TMF0 abilitato ?
	and	VIA0+VIAIER	;  e Timer motor spirato ?
	beq	?40		; no
	lda	#00010000B
	trb	fdcctl		; flag motor off
	lda	#00001100B
	sta	!FDCDOR		; enable controller, motor off
	lda	#CB2IFRB	; disabilita CB2-TMF0
	sta	VIA0+VIAIER
	sta	VIA0+VIAIFR	; clear flag CB2
?40:
?50:	rts
	
KbInt:
	pea	#DP01ADDR	; DP = pagina 1
	pld
	ldx	KBFR		; primo byte (flag)
	bpl	?20		; bit 7 = 0 -> codice ASCII
	txa
	and	#$B0		; maschera 4 bit alti - ignora release (6)
	cmp	#$80		; codice FLAG ?
	bne	?10		; no
	txa
	and	#$03		; bit 7 e 6 di KbdShift
	lsr	a		; C = bit 6
	ror	a		; C = bit 7
	ror	a
	and	#$C0		; maschera bit 7,6
	sta	KbdShift
	lda	KBFR+1		; flag
	sta	KbdFlag
	rts
?10:	cmp	#$90		; toggle bit ?
	bne	?20		; no
	lda	KBFR+1		; toggle bit
	sta	KbdToggle
	and	#SCROLLLOCKB	; test SCROLL LOCK BIT
	php
	lda	#$80		; SCOLL LOCK bit di ScnScroll
	plp
	beq	?12
	tsb	ScnScroll	; disabilita scroll
	rts
?12:	trb	ScnScroll	; abilita scroll
	rts
?20:	ldy	KbdCnt		; test spazio nel buffer
	cpy	#KBDBUFLEN
	bcs	?90		; no room - scarta tasto
	txa
	bit	#$10		; test 4 bit bassi toggle
	beq	?30		; no toggle
	lda	#$0F		; azzera 4 bit bassi di KbdToggle
	trb	KbdToggle
	txa 
	and	#$0F		; 4 bit bassi di KbdToggle
	tsb	KbdToggle	; imposta 4 bit bassi di KbdToggle
	txa
?30:				; inserisce codice tasto nel buffer
	and	#$C0		; bit 7 -> ASCII(0) oppure CONTROL(1)
				; bit 6 -> MAKE(0) oppure RELEASE(1)
	ldx	KbdITail	; indice coda buffer
	sta	KbdBuf,x	; inserisce flag tasto
	inx
	lda	KBFR+1		; codice tasto
	sta	KbdBuf,x
	inx
	iny			; aggiorna contatore bytes
	iny
	sty	KbdCnt
	cpx	#KBDBUFLEN	; test coda piena
	bcc	?40		; no
	ldx	#0		; coda circolare
?40:	stx	KbdITail	; aggiorna ptr. coda buffer
?90:	rts

	
int3sr:
int4sr:
int5sr:
int6sr:
	rts

; INT from RTC
int7sr:
	ldx	#RTCCTRLA	; setta banco 1 RTC
	stx	RTCALE
	lda	RTCDATA		; salva flag banco
	pha
	ora	#$10		; setta banco 1
	sta	RTCDATA
	ldx	#RTCSMI2	; recupera eventuale ALE
	stx	RTCALE
	lda	RTCDATA
	pha			; ALE da ripristinare
	ldx	#RTCSTATUS
	stx	RTCALE
	lda	RTCDATA		; azzera flag
	bmi	?01
	jmp	?20		; NO IRQ da RTC
?01:	bit	#$20		; flag AF (alarm)
	beq	?10		; no alarm
	ldx	#RTCEXTCTRLA	; check INCR bit 6
	stx	RTCALE
?02:	bit	RTCDATA
	bvs	?02
	ldx	#RTCDAY		; day of week
	stx	RTCALE
	lda	RTCDATA
	cmp	#1		; sunday?
	bne	?10		; no
	inx
	stx	RTCALE
	lda	RTCDATA		; day of month (date)
	cmp	#25		; last sunday of month?
	bcc	?10		; no
	inx
	stx	RTCALE		; month
	lda	RTCDATA
	cmp	#3		; march?
	bne	?04		; no
	ldx	#RTCHOURS	; increment hours
	stx	RTCALE
	inc	RTCDATA
	ldx	#RTCRAM0F
	stx	RTCALE
	stz	RTCDATA		; reset flag for next time	
	bra	?10
?04:	cmp	#10		; october?
	bne	?10		; no
	ldx	#RTCRAM0F
	stx	RTCALE
	lda	RTCDATA		; already updated?
	bpl	?06		; no
	stz	RTCDATA		; yes...so reset flag
	bra	?10
?06:	lda	#$FF
	sta	RTCDATA		; set updated flag
	ldx	#RTCHOURS	; decrement hours
	stx	RTCALE
	dec	RTCDATA
?10:	ldx	#RTCEXTCTRLA	; check others interrupt's
	stx	RTCALE
	lda	RTCDATA
	bit	#$01		; KF ?
	php	
	lda	#$07		; cancella altri IRQ
	trb	RTCDATA
	plp
	beq	?20		; NO KF
	lda	#$01		; TEST KS START
	bit	RTCFlag
	bne	?11		; KS OFF
	tsb	RTCFlag		; segnala KS OFF
	bra	?20
?11:	lda	#$08		; PAB = 1, PWR OFF
	;tsb	RTCDATA
	trb	RTCDATA
	lda	#$01		; CLEAR KS START
	;trb	SysRTC		
?20:	ply			; ALE da ripristinare
	ldx	#RTCCTRLA	; setta banco originale RTC
	stx	RTCALE
	pla			; banco originale
	sta	RTCDATA		; salva flag banco
	sty	RTCALE		; ripristina ALE
	rts
	

	.CODE

	.EXTERN SysStart
	.GLOBAL intsr
		
; vettori ISR
intsr:
	.DW	int0sr, int1sr, int2sr, int3sr
	.DW	int4sr, int5sr, int6sr, int7sr
	
; codice di startup
_RSTEntry:
	sta	CRHIMOFF	; forza start da flash
	sta	CRFWEON		; abilita flash/emulator F00000 - F7FFFF
	sta	CRXFEON		; emulator
	lda	CRBIT0		; test boot
	bne	?02		; boot from emulator
	sta	CRXFEOFF	; boot from flash	
?02:	sei
	cld
	clc
	xce			; CPU in modo nativo
	ldy	#0
	CPU16
	ldx	#STACK_ADDR	; init stack
	txs
	tya			; DP = 0 (page 0)
	tcd
	tyx
	lda	#$FFFF
	mvn	#$F0, #$F8
	mvn	#$F1, #$F9
	mvn	#$F2, #$FA
	mvn	#$F3, #$FB
	mvn	#$F4, #$FC
	mvn	#$F5, #$FD
	mvn	#$F6, #$FE
	mvn	#$F7, #$FF
	CPU08
	phk
	plb			; DBR = PBR = 0
	sta	CRXFEOFF	; disabilita flash F00000 - F7FFFF
	sta	CRFWEOFF
	jml	_SysStart	; long jmp - bootstrap (banco $F8)

_Unexpected:
	rti
	
; software interrupt generated by 'cop' istruction
;
; stack frame
;	---------
;	|  PBR  |	0F
;	---------
;	|  PCH  |	0E
;	---------	
;	|  PCL  |	0D
;	---------
;	|   P   |	0C
;	---------
;	|   B   |	0B
;	---------
;	|   A   |	0A
;	---------
;	|  XH   |	09
;	---------
;	|  XL   |	08
;	---------
;	|  YH   |	07
;	---------
;	|  YL   |	06
;	---------
;	|  DPH  |	05
;	---------
;	|  DPL  |	04
;	---------
;	|  DBR  |	03
;	---------
;	| CNTH  |	02
;	---------          --> CNT parameters bytes count 
;	| CNTL  |	01
;	---------
;
; equates for access stack offset data
STKCNT		.SET	$01
STKYR		.SET	$06
STKXR		.SET	$08
STKCR		.SET	$0A
STKBR		.SET	$0B
STKSR		.SET	$0C
STKPCL		.SET	$0D
STKPBR		.SET	$0F
STKNVAR		.SET	$0F

;USRCOP		.SET	$C0

;EINVAL		.EQU	$01

_COPEntry:
	CPU16			; A/MEM/X/Y -> 16 bit
	pha			; save C in stack
	phx			; save X(16) in stack
	phy			; save Y(16) in stack
	phd			; save DPR in stack
	phb			; save DBR in stack
	lda	#0
	tax			; X = 0
	pha			; params byte count in the stack ( = 0) 
	tcd			; set DPR = $0000
	phk			; set DBR = PBR = $00
	plb
	lda	STKPCL,s	; load PC saved in stack
	dec	a		; pointer to signature byte
	sta	COPPtr		; save long pointer
	CPU08			; A/MEM/X/Y -> 8 bit
	txa
	xba			; B = 0
	lda	STKPBR,s	; PBR bank where COP was executed
	sta	COPPtr+2
	lda	STKSR,s		; saved P in stack
	tax
	and	#~PCFLAG	; clear carry in saved P
	sta	STKSR,s
	bit	#PIFLAG		; check if IRQ was enabled
	bne	?04
	cli			; enable IRQ
?04:	phx			; saved P in stack
	lda	[COPPtr]	; fetch signature byte
	sta	COPIdx
	CPU16			; now C = $00XX
	asl	a		; index * 4	
	asl	a
	tax
	cpx	#$0200
	bcc	?05		; range $00..$7F
	lda	>SYSTBLE_ADDR,x	; address of cop vector range $80..$FF
	tay
	lda	>SYSTBLE_ADDR+2,x ; A = bank of cop vector, B = valid/params
	bra	?06
?05:	ldy	!SYSTBLE_ADDR,x	; address of cop vector range $00..$7F
	lda	!SYSTBLE_ADDR+2,x ; A = bank of cop vector, B = valid/params
?06:	sty	?08+1
	CPU08
	bmi	?30		; B<7> = 1 -> invalid function
	sta	!?08+3
	xba			; A = bytes number of params
	sta	STKCNT+1,s	; take in account the 'phx' in ?04 !!
	pla			; saved P in stack
	lsr	a		; pass carry to function
?08:	jsl	$000000
	bcc	?16		; no error
	sta	STKYR,s		; return error in Y
	lda	#0
	sta	STKYR+1,s
	lda	STKSR,s		; saved P in stack
	ora 	#PCFLAG		; set carry in saved P
	sta	STKSR,s
?16:	; epilogue code	
	CPU16
	lda	STKCNT,s	; number of params bytes in the stack
	beq	?20		; no params -- skip stack cleaning
	clc
	tsc			; C = stack pointer
	adc	#STKNVAR	; add size of stack vars
	tax			; source pointer for data move  
	adc	STKCNT,s	; add params bytes count
	tay			; dest pointer for data move
	lda	#STKNVAR-1	; move bytes count
	mvp	#0, #0		; cleanup stack
	tya			; new stack pointer
	tcs
?20:	pla			; skip STKCNT
	plb			; restore DBR
	pld			; restore DPR
	ply			; restore Y
	plx			; restore X
	pla			; restore C
	rti			; restore P and return

?30:	plx	
	jmp	?16

_BRKEntry:
	CPU16			; salva registri 16 bit
	pha			; salva A
	phx			; salva X
	phy			; salva Y
	phd
	phb			; salva DBR
	phk			; DBR = 0 = PBR(K)
	plb			; DATA BANK 0
	CPU08
	jml	BRKMON
	
; gestione interrupt non mascherabile linea NMI (RTI restore P and register size)
_NMIEntry:
	CPU16			; salva registri 16 bit
	pha			; salva A
	phx			; salva X
	phy			; salva Y
	phd
	phb			; salva DBR
	phk			; DBR = 0 = PBR(K)
	plb			; DATA BANK 0
	lda	#0		; pagina 0
	tcd
	CPU08			; imposta registri 8 bit

	lda	!VIA1+VIAIFR	; flag IFR VIA 1
	bpl	?100		; no irq da VIA 1
	lsr	a		; test CA2 (NMI KEYB)
	bcc	?50		; no CA2
	ldx	KBFR+2
	cpx	#KB_CTRLBREAK
	bne	?02
	tay
	lda	#$80
	tsb	CtrlBrk
	tya
	bra	?40
?02:	cpx	#KB_ALTSYSREQ
	bne	?06
	ldx	#CA2IFRB
	stx	VIA1+VIAIFR
	stx	KBCLRNMI	; azzera linea INT
	tay
	lda	#$40
	tsb	CtrlBrk	
	jml	ALTSYSMON
?06:
?40:	ldx	#CA2IFRB
	stx	VIA1+VIAIFR
	stx	KBCLRNMI	; azzera linea INT
?50:	lsr	a
	lsr	a
	lsr	a
	bcc	?100
	ldx	#CB2IFRB
	stx	VIA1+VIAIFR	
	lda	#$01
	tsb	CtrlBrk	

?100:	
	CPU16			; ripristina registri 16 bit
	plb			; ripristina DBR
	pld
	ply			; ripristina Y
	plx			; ripristina X
	pla			; ripristina A
	rti
	LONG_OFF

; gestione interrupt linea IRQ (RTI restore P and register size)
_IRQEntry:
	CPU16			; salva registri 16 bit
	pha			; salva A
	phx			; salva X
	phy			; salva Y
	phd			; salva DP
	phb			; salva DBR
	phk			; DBR = 0 = PBR(K)
	plb			; DATA BANK 0
	lda	#0
	tcd			; imposta pagina 0
	CPU08			; imposta registri 8 bit
	ldx	IRQVECTR	; legge vettore IRQ (x2)
	jsr 	(intsr,x)	; esegue subroutine IntSr
	CPU16			; ripristina registri 16 bit
	plb			; ripristina DBR
	pld			; ripristina DP
	ply			; ripristina Y
	plx			; ripristina X
	pla			; ripristina A
	rti
	LONG_OFF
	
;----------------------------------------------------------
; ---- VETTORI CPU
;----------------------------------------------------------
	
	.SYSVECTRS
	
	; vettori modo nativo
	.DW	0		; FFE0
	.DW	0		; FFE2
	.DW	_COPEntry	; FFE4 -> COP
	.DW	_BRKEntry	; FFE6 -> BRK
	.DW	_Unexpected	; FFE8 -> ABORT (non usato)
	.DW	_NMIEntry	; FFEA -> NMI
	.DW	_Unexpected	; FFEC -> riservato
	.DW	_IRQEntry	; FFEE -> IRQ

; nel banco $F8
_SysStart:
	jmp	!SysStart
	.DB	$00
	
	; vettori modo emulazione
	.DW	_COPEntry	; FFF4 -> COP
	.DW	_Unexpected	; FFF6 -> riservato
	.DW	_Unexpected	; FFF8 -> ABORT (non usato)
	.DW	_NMIEntry	; FFFA -> NMI
	.DW	_RSTEntry	; FFFC -> RESET
	.DW	_IRQEntry	; FFFE -> IRQ

	
	.END _RSTEntry
