;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; bios 65C816 version v 1.0

;; vprint.asm
;; c-like format function (as printf())

;; This code attempt to implement the "c-like" printf() family function
;; Current implementation is near fully compatible with standard C function's
;; For compatibility comparison see the table below, about format string
;;
;;	FORMAT STRING
;;	'%'[flag][width]['.'precision][modifier]type
;;
;;	flags ::= 	'-' | '+' | ' ' | '#' | '0' | ','
;;	width ::=       '*' | number (8 bit) (if * byte passd by arg)
;;	precision ::=   '*' | number (8 bit) (if * byte passd by arg)
;;	modifier ::=	'l' | 'h' | 'b' | 'H' | 'B'
;;	type ::=	'c' | 's' | 'd' | 'i' | 'u' | 'x' | 'X' | 'p' | 'P' |
;;			'f' | 'F' | 'e' | 'E' | 'g' | 'G'
;;
;;	flag '#'	- alternate form for hex:C-style prefix with 0x or 0X
;;	flag '-'	- left justify (default)
;;	flag '+'	- right justify
;;	flag ' ' or '0' - fill field with blank or '0'
;;	flag ','	- only for d,i,u types - group thousands with 
;;			  ',' as seperator
;;	c		- character (no modifier allowed)
;;	s		- ASCIZ string (NULL terminated string)
;;	Bs		- Pascal/Basic string (not standard)
;;	d,i		- signed integer 32 bit (in decimal)
;;	u		- unsigned integer 32 bit (in decimal)
;;	ld, li		- long signed integer 64 bit (in decimal)
;;	lu		- long unsigned integer 64 bit (in decimal)
;;	hd, hi		- short signed integer 16 bit (in decimal)
;;	hu		- short unsigned integer 16 bit (in decimal)
;;	bd, bi		- byte signed integer (in decimal, not standard)
;;	bu		- byte unsigned integer (in decimal, not standard)
;;	x,X		- unsigned integer 32 bit in hex (nocaps/caps)
;;	lx,lX		- long unsigned integer 64 bit in hex (nocaps/caps)
;;	hx,hX		- short unsigned integer 16 bit in hex (nocaps/caps)
;;	bx,bX		- unsigned byte in hex (nocaps/caps, not standard)
;;	p,P		- long pointer (3 bytes) in hex (nocaps/caps, no std.)
;;	f,F,e,E,g,G	- float point number (see sprintf()) 

;; format function's follow "c" convention for param's, so for example:
;; sprintf(count,dest,format,...)
;; arg's are pushed in stack in that order: last_par,....,format,dest,count
;; count = sum of bytes pushed (count too)

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp02.inc

FMINPREC	.EQU	8	; minimum float precision (if unspecified)
GLPPUT		.EQU	DP02ADDR+FmtLPPut	; long pointer to putter function

	
;----------------------------------------------------------

; spec's class
_dc	.EQU	0	;/* don't care           */
_si	.EQU	1	;/* sign fill +/-        */
_af	.EQU	2	;/* alternate form       */
_ar	.EQU	3	;/* format (width or precision) by argument */
_lj	.EQU	4	;/* left justify         */
_pr	.EQU	5	;/* precision            */
_nu	.EQU	6	;/* numeral              */
_lo	.EQU	7	;/* long                 */
_sh	.EQU	8	;/* short                */
_SH	.EQU	9	;/* byte                 */
_fz	.EQU	10	;/* fill zeros           */
_de	.EQU	11	;/* decimal              */
_un	.EQU	12	;/* unsigned decimal     */
_he	.EQU	13	;/* hexadecimal          */
_pt	.EQU	14	;/* pointer              */
_HE	.EQU	15	;/* HEXADECIMAL		 */
_PT	.EQU	16	;/* POINTER              */
_ch	.EQU	17	;/* char                 */
_st	.EQU	18	;/* string               */
_ff	.EQU	19	;/* float %f/F           */
_fe	.EQU	20	;/* float %e             */
_fg	.EQU	21	;/* float %g             */
_FE	.EQU	22	;/* float %E             */
_FG	.EQU	23	;/* float %G             */
_tg	.EQU	24	;/* thousand group       */
_bc	.EQU	25	;/* BCD float numbers    */
_PS	.EQU	26	;/* pascal/basic string  */
_xf	.EQU	27	;/* hex float            */
_XF	.EQU	28	;/* HEX FLOAT            */

; FmtFlag flag
F_ALTFMT	.EQU	$80
F_LEFTJ		.EQU	$40
F_FILLZ		.EQU	$20
F_GROUP		.EQU	$10
F_WIDTH		.EQU	$08
F_PREC		.EQU	$04

; FmtMod modifier's flag
F2_PSSTRING	.EQU	$80
F2_BCD		.EQU	$40
F2_LONG		.EQU	$20
F2_SHORT	.EQU	$10
F2_BYTE		.EQU	$08
F2_CAPS		.EQU	$04
F2_SIGNED	.EQU	$01

MASKMOD		.EQU	(F2_LONG.OR.F2_SHORT.OR.F2_BYTE.OR.F2_BCD)
; modifier mask '%c'
MASKCHAR	.EQU	(MASKMOD.OR.F2_PSSTRING)
; modifier mask '%s'
MASKSTR		.EQU	(MASKMOD)
; modifier mask '%p'
MASKPTR		.EQU	(MASKCHAR)
; modifier mask float numbers
MASKFLOAT	.EQU	(F2_LONG.OR.F2_SHORT.OR.F2_BYTE.OR.F2_PSSTRING)
; modifier mask integer
MASKINTG	.EQU	(F2_BCD.OR.F2_PSSTRING)

; conversion function index 
T_CHAR		.EQU	0
T_STR		.EQU	(1 * 2)
T_PTR		.EQU	(2 * 2)
T_INT		.EQU	(3 * 2)
T_HEX		.EQU	(4 * 2)
T_FLOAT		.EQU	(5 * 2)

; current stage while scan format
FLAGSTAGE	.EQU	0
FILLZSTAGE	.EQU	1
WIDESTAGE	.EQU	2
DOTSTAGE	.EQU	3
PRECSTAGE	.EQU	4
MODSTAGE	.EQU	5

; aux var's
PADCHR		.EQU	FPWTmp7		; char per padding ('0' o ' ')
PFXCHR		.EQU	FPWTmp7+1	; prefisso stringa (segno o '$')
NPADL		.EQU	FPWTmp8		; numero char padding a sinistra
NPADR		.EQU	FPWTmp8+1	; numero char padding a destra
STRLEN		.EQU	FPWTmp		; lunghezza stringa da copiare
PFXFLAG		.EQU	FPWTmp+1	; flag prefix segno/'$'
FMTTMP		.EQU	FPFlag		; temporaneo durante scanning formato


;---------------------------------------------------------------------------
; code segment
;---------------------------------------------------------------------------

	.CODEF8
	.EXTERN _Int2Str, _feFAC2S, _MovMem2Fa
	
	.GLOBAL sys_bprintf, sys_bsprintf, sys_bsprintf2, sys_bvprintf

;---------------------------------------------------------------------------
; system call interface
;---------------------------------------------------------------------------

; _bprintf:	format string and send to text default device,usually console
;		this function act like c-function printf()
;
; prototype:	_bprintf(count, format, ...)
;
;		count	=bytes count of all passed param's (count too)
;		format	=long pointer to format string
;		...	=param's list (like in format string)
;
;		param's passed in reverse order in stack
;
; 		all register's are preserved
sys_bprintf:
	lda	STKPRMS,s	; count bytes params
	sta	STKCNT,s
	CPU16CLC		; 16 bit + CLC
	tsc
	adc	#STKPRMS+1
	tax			; frame stack parametri
	lda	#DP02ADDR
	tcd			; imposta DPR => pagina 02
	stz	FmtGMax		; non usa stringa globale
	stz	FmtLPPut	; non usa putter globale
	lda	!0,x		; carica puntatore stringa di formato
	sta	FmtLPFmt
	inx
	inx	
	ACC08
	stz	FmtLPPut+2	; non usa putter globale			
	lda	!0,x
	sta	FmtLPFmt+2
	inx
	stx	FmtParm		; salva frame stack parametri	
	CPU08
	lda	#$80		; call putter locale 
	jsr	_vprinter
	clc
	rtl

; _bsprintf2:	format string and put result in a local string
;		this function act like c-function sprintf()
;
; prototype:	_bsprintf2(count, format, ...)
;
;		count	=bytes count of all passed param's (count too)
;		format	=long pointer to format string
;		...	=param's list (like in format string)
;
;		param's passed in reverse order in stack
;
; out:		C	=offset of result local string
;		X	=bank that hold result local string
;
;		local string is limited to 255 bytes
sys_bsprintf2:
	lda	STKPRMS,s	; count bytes params
	sta	STKCNT,s
	CPU16CLC		; 16 bit + CLC
	tsc
	adc	#STKPRMS+1
	tax			; frame stack parametri
	lda	#DP02ADDR
	tcd			; imposta DPR => pagina 02
	stz	FmtGMax		; non usa stringa globale
	stz	FmtLPPut	; non usa putter globale
	lda	!0,x		; carica puntatore stringa di formato
	sta	FmtLPFmt
	inx
	inx	
	ACC08
	stz	FmtLPPut+2	; non usa putter globale			
	lda	!0,x
	sta	FmtLPFmt+2
	inx
	stx	FmtParm		; salva frame stack parametri	
	CPU08
	lda	#$00		; local string 
	jsr	_vprinter
	ACC16
	lda	#LVPBUF		; buffer locale
	sta	STKCR,s
	lda	#^LVPBUF
	sta	STKXR,s
	ACC08
	clc
	rtl

; _bsprintf:	format string and put result in a global string
;		this function act like c-function sprintf()
;
; prototype:	_bsprintf(count, format, dest, size, ...)
;
;		count	=bytes count of all passed param's (count too)
;		format	=long pointer to format string
;		dest	=long pointer to destination string
;		size    = size of destination buffer
;		...	=param's list (like in format string)
;
;		param's passed in reverse order in stack
;
; out:		C	=count of bytes needs for destination string not
;			 including the #0 terminator
;
;		if dest=NULL, function just compute the needs number of bytes
;		dest must have room at least for C+1 bytes
sys_bsprintf:
	lda	STKPRMS,s	; count bytes params
	sta	STKCNT,s
	CPU16CLC		; 16 bit + CLC
	tsc
	adc	#STKPRMS+1
	tax			; frame stack parametri
	lda	#DP02ADDR
	tcd			; imposta DPR => pagina 02
	lda	!0,x		; carica puntatore stringa di formato
	sta	FmtLPFmt
	inx
	inx	
	ACC08
	lda	!0,x
	sta	FmtLPFmt+2
	inx
	ACC16
	lda	!0,x		; carica puntatore stringa destinazione
	sta	FmtLPPut
	inx
	inx
	ACC08
	lda	!0,x
	sta	FmtLPPut+2
	inx
	ACC16
	lda	!0,x
	sta	FmtGMax		; carica size stringa destinazione
	inx
	inx	
	stx	FmtParm		; salva frame stack parametri
	CPU08
	lda	#$40		; stringa globale 
	jsr	_vprinter
	ACC16
	lda	FmtCnt		; numero caratteri
	sta	STKCR,s
	ACC08
	clc
	rtl

; _bvprintf:	format string and send result to an user-defined "putter"
;		function; this function act like c-function fprintf()
;
; prototype:	_bvprintf(count, format, putter...)
;
;		count	=bytes count of all passed param's (count too)
;		format	=long pointer to format string
;		putter	=long pointer to user-dfined putter function
;		...	=param's list (like in format string)
;
;		param's passed in reverse order in stack
;
; out:		C	=count of bytes needs sent to user-defined putter, not
;			 including the #0 terminator
;
;		the user-defined putter will end with an "rtl" instruction
sys_bvprintf:
	lda	STKPRMS,s	; count bytes params
	sta	STKCNT,s
	CPU16CLC		; 16 bit + CLC
	tsc
	adc	#STKPRMS+1
	tax			; frame stack parametri
	lda	#DP02ADDR
	tcd			; imposta DPR => pagina 02
	stz	FmtGMax
	lda	!0,x		; carica puntatore stringa di formato
	sta	FmtLPFmt
	inx
	inx	
	ACC08
	lda	!0,x
	sta	FmtLPFmt+2
	inx
	ACC16
	lda	!0,x		; carica puntatore stringa destinazione
	sta	FmtLPPut
	inx
	inx
	ACC08
	lda	!0,x
	sta	FmtLPPut+2
	inx
	stx	FmtParm		; salva frame stack parametri
	CPU08
	lda	#$C0		; putter globale 
	jsr	_vprinter
	ACC16
	lda	FmtCnt		; numero caratteri
	sta	STKCR,s
	ACC08
	clc
	rtl

; funzione locale di formattazione stringa																																																																																																																																																																																																																												
; In	- FmtLPFmt = long ptr stringa formato (szFmt)
;	  FmtParm = ptr array parametri nel banco 0																																																																				
;	  FmtLPPut = long ptr alla funzione 'putter' o a stringa globale
;	  FmtGMax = max. lunghezza stringa globale (escluso NULL)
;	  A => flag funzione	<7> => 0 -> store stringa / 1 -> store putter
;				<6> => 0 -> locale / globale 
;	  DBR = 0 (banco 0)
;	  DPR = 2 (pagina 02)
;
; Out	- store locale => LVPBUF stringa formattata max. 255 car. (szDst)
;	  store globale => stringa FmtLPPut formattata (max $FFFF car.)
;	  putter locale => formattazione inviata al device TEXT corrente
;	  putter globale => formattazione inviata a funzione putter
_vprinter:
	.LONGA	off
	.LONGI	off
	
	ldy	#0			; azzera contatore caratteri
	sty	FmtCnt
	sty	FmtCnt+1
	sty	FmtDst			; azzera indice stringa locale
	sta	VStoreF			; flag store
	
	; LOOP scanning formato
?10:	lda	[FmtLPFmt],y		; next char
	beq	?60			; fine stringa - EXIT
	;bne	?11
	;jmp	?60
?11:	cmp	#'%'
	beq	?14			; specificatore formato
?12:	iny				; ptr next char
	jsr	_store			; store char in szDst
	bra	?10			; loop
?14:	sty	FmtSpec			; indice carattere '%' (per ABANDON)
	iny
	lda	[FmtLPFmt],y		; next char dopo specificatore
	cmp	#'%'
	beq	?12			; '%%' -> store '%' e loop

	; analizza specificatore di formato dopo '%'
	lda	#0			; A = stage = 0 (flag stage)
	sta	FmtSgn
	sta	FmtFlag
	sta	FmtMod	
	sta	FmtWidth
	sta	FmtPrec
	bra	?21			; reload
	
	; LOOP scanning specificatore formato
?20:	iny

	; A = stage - Y = ptr corrente formato
?21:	xba				; B = stage
	lda	[FmtLPFmt],y		; A = next char dopo specificatore
	beq	?50			; fine stringa - ABANDON
	sta	FMTTMP			; salva char
	sec
	sbc	#' '			; scala char to 0..5F
	bcc	?50			; char illegale a questo stadio
	cmp	#(128 - ' ')
	bcs	?50			; char illegale a questo stadio
	tax				; X = indice 0..5F
	lda	>CLASSCHR,x		; A = class char
	asl	a
	tax				; indice funzione
	lda	FMTTMP			; cur char
	xba				; B = cur char, A = stage, ZF valido
	jsr	(SCANTBL,x)		; scanning formato
?20a:	bcs	?50			; errore formato - ABANDON
	bvc	?20			; loop - scanning formato corrente
	
	; a questo punto si assume che lo specificatore di formato sia stato
	; scannerizzato correttamente e che X contenga l'indice alla tabella
	; di conversione CVTTBL
	phy				; salva Y
	jsr	(CVTTBL,x)		; funzione di conversione
?20b:	ply				; ripristina Y
	iny				; next char
	bra	?10			; loop - next formato
	
	; ABANDON - causa errore specificatore di formato si suppone che i
	; parametri non siano piu' allineati con il formato
	; viene quindi copiato letteralmente il resto di szFmt a partire
	; dal carattere '%' dove si e' verificato il disallineamento, fino
	; alla fine della stringa szFmt
?50:	ldy	FmtSpec			; indice al carattere '%'
?52:	lda	[FmtLPFmt],y		; copia da szFmt a szDst
	beq	?60			; fino a fine stringa
	iny
	jsr	_store
	bra	?52
	
	; EXIT
	; la funzione memorizza il terminatore #0 per indicare la fine
	; o chiama il putter per effettuare il flush del buffer
?60:	bit	VStoreF
	bmi	?64			; chiama flusher locale/globale
	bvs	?62			; termina stringa globale
	ldx	FmtDst			; indice stringa locale	
	stz	!LVPBUF,x		; store NULL at X = FmtDst
	rts
?62:	INDEX16
	ldy	FmtCnt
	lda	#0
	sta	[FmtLPPut],y		; store NULL in stringa globale
	INDEX08
	rts
?64:	bvc	_lflush			; termina putter locale
	lda	#0			; termina putter globale
	bra	_gput
	
;----------------------------------------------------------------------------
; FUNZIONI STORE/FLUSH
;----------------------------------------------------------------------------

; termina putter locale
_lflush:
	ldx	FmtDst			; indice attuale buffer locale

; invia buffer locale al device selezionato	
_lflush2:
	stz	!LVPBUF,x		; termina stringa
	txa
	beq	?02			; buffer vuoto
	ACC16
	lda	#LVPBUF			; CX = long ptr al buffer
	ACC08
	ldx	#0			; banco 0
	TXTSTROUT		; preserva A,X,Y
?02:	rts	

; funzione putter locale/globale
_putter:
	bvc	_lput			; putter locale
	
; putter globale -- char NULL indica END-OF-FILE
_gput:
	sta	FPDCnt
	sty	FPIndx
	phk
	pea	#?02-1			; salva ind. return => ?02
	
	pha
	ACC16
	lda	!GLPPUT
	sta	>?04+1
	ACC08
	lda	!GLPPUT+2
	sta	>?04+3
	pla
	cmp	#0
	bne	?04
	nop
	nop
?04:	jml	$80FF00
	
	jml	[GLPPUT]		; deve terminare con RTL
?02:	ldy	FPIndx
	lda	FPDCnt
	rts
		
; funzione putter locale
_lput:
	jsr	_lstor			; bufferizza in stringa locale
	bne	?02			; buffer non pieno
	sta	FPDCnt			; salva char	
	jsr	_lflush2		; flush buffer
	ldx	#0
	lda	FPDCnt			; memorizza char attuale
	sta	!LVPBUF,x		; store at X = 0
	inx
	stx	FmtDst
?02:	rts

; store char in stringa locale/globale o invia carattere al 'putter'
_store:
	ACC16
	inc	FmtCnt			; update numero bytes (16 bit)
	ACC08
	bit	VStoreF
	bmi	_putter			; chiama putter locale/globale
	bvs	_gstor			; store in stringa globale

; store char in stringa locale
_lstor:
	ldx	FmtDst			; indice stringa locale
	cpx	#$FF			; test overflow
	beq	?02			; no store (FmtDst = $FF)
	sta	!LVPBUF,x		; store at X = FmtDst
	inx
	stx	FmtDst
?02:	rts

; store char in stringa globale
_gstor:
	sty	FPIndx			; salva Y
	INDEX16
	ldy	FmtCnt
	dey				; indice attuale stringa globale
	cpy	FmtGMax			; overflow stringa ?
	bcs	?02			; si -- no store
	sta	[FmtLPPut],y		; store in stringa globale
?02:	INDEX08
	ldy	FPIndx
	rts

;----------------------------------------------------------------------------
; FUNZIONI DI CONVERSIONE (FORMATTAZIONE)
;----------------------------------------------------------------------------

; preleva integer dai parametri e lo copia in ACM
getint:
	ldy	#1
	lda	FmtMod
	bit	#F2_BYTE
	bne	?02			; byte -> size = 1
	iny
	bit	#F2_SHORT
	bne	?02			; word -> size = 2
	ldy	#8
	bit	#F2_LONG
	bne	?02			; quad -> size = 8
	ldy	#4			; dword -> size = 4
?02:	sty	ACMSize			; size
	stz	ACMSgn			; unsigned per il momento
	INDEX16
	ldx	FmtParm			; indice parametri
	lda	!0,x			; get byte
	sta	<ACM
	inx
	dey
	beq	?06			; size = 1 (byte)
	lda	!0,x			; get byte
	sta	<ACM+1
	inx
	dey
	beq	?06			; size = 2 (word)
	ACC16
	lda	!0,x			; get word
	sta	<ACM+2
	inx
	inx
	dey
	dey
	beq	?06			; size = 4 (dword)
	lda	!0,x			; get word
	sta	<ACM+4
	inx
	inx
	lda	!0,x			; get word
	sta	<ACM+6			; size = 8 (quad)
	inx
	inx	
?06:	stx	FmtParm			; aggiorna indice array parametri
	CPU08
	rts

DoFloat:
	lda	FmtFlag
	bit	#F_PREC
	bne	?01			; specificata la precision
	ldx	#FMINPREC		; imposta minima precisione
	lda	XVFlag
	bit	#$04
	beq	?00			; no formato a/A
	ldx	#0			; minima precisone formato a/A = 0
?00:	stx	FmtPrec
?01:	lda	FmtMod
	bit	#F2_CAPS
	beq	?02
	lda	#$08			; flag CAPS
	tsb	XVFlag
?02:	bit	FmtFlag
	bpl	?04			; no formato alt.
	lda	#$10			; flag formato alt
	tsb	XVFlag
?04:	ldx	FmtSgn			; segno ?
	beq	?08			; no
	lda	#$01			; flag segno
	cpx	#'+'
	beq	?06			; usa '+'
	ora	#$02			; usa ' '
?06:	ora	XVFlag
	sta	XVFlag
?08:	INDEX16
	ldx	FmtParm			; indice parametri
	lda	!0,x			; get long pointer
	sta	PTR1
	inx
	lda	!0,x
	sta	PTR1+1
	inx
	lda	!0,x
	sta	PTR1+2
	inx	
	stx	FmtParm			; aggiorna indice array parametri
	INDEX08
	lda	#F_ALTFMT
	trb	FmtFlag			; ignora formato alt. nel futuro
	bit	FmtMod
	bvc	?20			; binary float
	jsr	_MovMem2Fa
	lda	XVFlag
	ldx	FmtPrec
	jsr	_feFAC2S		; BCD float
	jmp	DoNum
?20:	jsr	_MovMem2Fa		; binary float	
	lda	XVFlag
	ldx	FmtPrec
	jsr	_feFAC2S
	jmp	DoNum

; formattazione long pointer p,P
DoPtr:
	ldy	#3			
	sty	ACMSize			; size = 3
	stz	ACMSgn			; unsigned
	INDEX16
	ldx	FmtParm			; indice parametri
	lda	!0,x			; get byte
	sta	<ACM
	inx
	lda	!0,x			; get byte
	sta	<ACM+1
	inx
	lda	!0,x			; get byte
	sta	<ACM+2
	inx
	stx	FmtParm			; aggiorna indice array parametri
	INDEX08
	ldy	#0			; flag hex minuscolo
	lda	FmtMod
	bit	#F2_CAPS
	beq	?02
	ldy	#$40			; flag hex maiuscolo
?02:	tya
	ora	#$10			; flag formato ptr
	ldx	#6			; precision = 6
	bra	DoInt3	
	
; formattazione interi decimali i,d,u
DoInt:
	jsr	getint			; preleva integer
	ldy	#$80			; flag decimale
	lda	FmtFlag
	bit	#F_GROUP
	beq	?08			; no group
	ldy	#$C0			; flag decimale + group
?08:	lda	FmtMod
	lsr	a			; CF = 1 => signed
	bcc	DoInt2			; unsigned
	ldx	<ACMSize
	dex
	lda	<ACM,x			; MSB
	and	#$80			; maschera segno
	ora	#$40			; tipo signed
	sta	ACMSgn
	ldx	FmtSgn			; segno ?
	beq	DoInt2			; no
	iny				; flag segno
	cpx	#'+'
	beq	DoInt2			; usa '+'
	iny
	iny				; usa ' '
	bra	DoInt2
	
; formattazione interi esadecimali x,X
DoHex:
	jsr	getint			; preleva integer
	ldy	#0			; flag hex minuscolo
	lda	FmtMod
	bit	#F2_CAPS
	beq	DoInt2
	ldy	#$40			; flag hex maiuscolo

DoInt2:
	ldx	#1			; default precision
	lda	FmtFlag
	bit	#F_PREC
	beq	?14
	ldx	FmtPrec
	bne	?14
	inx				; minima precision = 1
?14:	tya				; A = flag

DoInt3:
	jsr	_Int2Str		; A = Y = len -- X = ptr stringa
	
DoNum:
	stz	PFXFLAG			; flag store segno / '$'
	bit	FmtFlag
	bpl	?04			; solo p/P/x/X ammettono prefisso
	lda	#'$'			; prefisso alt. form.
	bra	?08
?04:	lda	FmtMod
	bit	#F2_SIGNED
	beq	?20			; unsigned -- no strip segno
	lda	<0,x			; segno prima del digit
	cmp	#'+'
	beq	?06
	cmp	#'-'
	beq	?06
	cmp	#' '
	bne	?20			; nessun segno prima del digit
?06:	inx				; sposta ptr. al primo digit
	dey				; lunghezza stringa originale - 1
?08:	sta	PFXCHR			; segno numero o prefisso '$'
	
	; ora il segno del numero e' stato strippato o il prefisso e' stato
	; fissato ed occorre stabilire se esso va emesso come primo
	; carattere oppure no.
	; Se e' stato specificato il campo width con allineamento di default
	; a destra ed il padding a sinistra viene effettuato con '0' allora
	; il segno/prefisso va emesso come primo carattere, altrimenti va
	; emesso prima della stringa numerica. Dato che il padding a destra
	; avviene sempre con blank, il prefisso/segno va emesso come primo
	; carattere se F_FILLZ e F_WIDTH sono entrambi settati.
	
	lda	FmtFlag
	and	#(F_FILLZ.OR.F_WIDTH)	; specificato width con padding '0' ?
	cmp	#(F_FILLZ.OR.F_WIDTH)
	beq	?10			; si - emette segno/prefisso per primo
	lda	#$40			; no - emette segno/prefisso dopo
	bne	?12
?10:	lda	#$80			; per primo
?12:	sta	PFXFLAG
?20:	stx	PTR1			; store puntatore
	lda	#>DP02ADDR
	sta	PTR1+1
	stz	PTR1+2			; banco 0
	jmp	fmtstr	
	
; formattazione '%c' -- tipo char
DoChar:
	INDEX16
	ldx	FmtParm			; indice parametri
	lda	!0,x			; get char
	inx
	stx	FmtParm			; aggiorna indice array parametri
	INDEX08
	sta	FPUStr			; stringa temporanea
	stz	FPUStr+1		; di 1 carattere
	lda	#FPUStr
	sta	PTR1			; ptr stringa da copiare
	lda	#>DP02ADDR
	sta	PTR1+1
	stz	PTR1+2			; banco 0
	ldy	#1			; lunghezza = 1
	bra	clrpfx
	
; formattazione '%s' -- tipo stringa
DoStr:
	ldy	#0
	INDEX16
	ldx	FmtParm			; indice parametri
	lda	!0,x			; get long ptr
	sta	PTR1
	inx
	lda	!0,x			; get long ptr
	sta	PTR1+1
	inx
	lda	!0,x			; get long ptr
	sta	PTR1+2
	inx
	stx	FmtParm			; aggiorna indice array parametri
	ora	PTR1+1			; test ptr NULL
	ora	PTR1
	bne	?00			; ptr != NULL
	lda	#<NULLSTR		; copia stringa di default "(null)"
	sta	PTR1
	lda	#>NULLSTR
	sta	PTR1+1
	lda	#^NULLSTR		; oppure .HIGH8.
	sta	PTR1+2
	bra	?01
?00:	lda	FmtMod
	bit	#F2_PSSTRING
	beq	?01			; stringa stile C
	lda	[PTR1]			; lunghezza stringa pascal/basic
	tay
	ACC16				; char stringa pascal dopo lunghezza
	inc	PTR1
	ACC08
	bne	?03
	inc	PTR1+2
	bra	?03
?01:	lda	[PTR1],y		; calcola lunghezza stringa stile C
	beq	?02
	iny
	bne	?01
?02:	cpy	#$0100			; max. 255 caratteri
	bcc	?03
	ldy	#$00FF
?03:	INDEX08				; Y = lunghezza stringa
	lda	FmtFlag			; test 'precision'
	bit	#F_PREC
	beq	clrpfx			; no precision specificata
	cpy	FmtPrec			; Len >= precision ?
	bcc	clrpfx			; NO
	ldy	FmtPrec			; SI -> Len = precision
					; precision puo' troncare la stringa	
clrpfx:
	stz	PFXFLAG
	
; formatta stringa
fmtstr:
	sty	STRLEN			; salva lunghezza stringa da copiare
	stz	NPADL			; no padding a sinistra
	stz	NPADR			; no padding a destra
	lda	PFXFLAG			; occorre memorizzare prefix/segno ?
	and	#$C0
	beq	?02			; NO
	iny				; conteggia 1 byte in +
?02:	sty	FPTmp4			; lunghezza reale stringa
	lda	FmtFlag
	bit	#F_WIDTH
	beq	?18			; no padding sx/dx
	lda	FmtWidth
	beq	?18			; width = 0 => no padding 
	sec
	sbc	FPTmp4
	beq	?18			; width = len => no padding
	bcc	?18			; width < len => no padding
	tax				; X = width - len => padding
	lda	FmtFlag
	bit	#F_LEFTJ
	beq	?04			; allineameno a destra (default)
	stx	NPADR			; allinea a sinistra => padding a dx 			
	bra	?06			; padding a dx sempre con blank
?04:	stx	NPADL			; allinea a dx => padding a sx
	ldx	#'0'			; pad con '0' ?
	bit	#F_FILLZ
	bne	?08			; SI
?06:	ldx	#' '			; NO -- padding con blank
?08:	stx	PADCHR
?18:	bit	PFXFLAG			; segno o prefisso '$' come primo ?
	bpl	?20			; NO
	lda	PFXCHR			; store segno/prefix
	jsr	_store
?20:	ldy	NPADL			; numero char padding a sinistra
	beq	?24			; no padding a sinistra
	lda	PADCHR			; carattere di padding
?22:	jsr	_store
	dey
	bne	?22
?24:	bit	PFXFLAG			; segno o prefisso '$' adesso ?
	bvc	?26			; NO
	lda	PFXCHR			; store segno/prefix
	jsr	_store
?26:	ldy	STRLEN			; copia la stringa 
	beq	?30			; non ci sono char da copiare
	ldy	#0
?28:	lda	[PTR1],y		; copia byte a byte
	jsr	_store
	iny
	dec	STRLEN
	bne	?28
?30:	ldy	NPADR			; numero char padding a destra
	beq	?34			; no padding a destra
	lda	PADCHR			; carattere di padding
?32:	jsr	_store
	dey
	bne	?32
?34:	rts
	
;----------------------------------------------------------------------------
; FUNZIONI SCANNING SPECIFICATORE DI FORMATO 
; 
; In	- B = cur char, A = stage (e flag N,Z)
;
; Out	- CF = 0 -> OK, A = stage, VF = 0 -> continua scanning
;	       VF = 1 -> fine scanning, X -> indice funzione conversione
;	  CF = 1 -> formato errato
;----------------------------------------------------------------------------

; case ' ' or '+' -- valido solo nello stato stage = 0
case_si:
	bne	case_dc			; abandon
	ldx	FmtSgn
	cpx	#'+'
	beq	?04			; ignora ' ' se '+' gia' specificato
	xba				; A = cur char - B = stage
	sta	FmtSgn
	xba				; A = stage
?04:
retok:	clc				; OK
	clv				; continua scanning
	rts

; case digit '0' - padding a sinistra con '0' invece di ' '
; '0' o e' il primo digit del campo width, oppure e' un digit successivo
; che fa parte del campo width o del campo precision
case_fz:
	; il flag Z e' valido -- indica se stage = 0
	bne	case_nu			; digit fa parte di width/precision ?
	tax				; X = stage
	lda	#F_LEFTJ
	bit	FmtFlag			; ignora padding '0' se left just.
	beq	?01
	txa				; A = stage
	bne	retok
?01:	lda	#F_FILLZ		; padding '0'
	inx				; X = 1 -> padding stage
	bra	setflag			; dopo deve seguire campo width

; case '-'  -- valido solo nello stato stage = 0
case_lj:
	bne	case_dc			; abandon
	tax				; X = stage
	lda	#F_LEFTJ		; flag allineamento a sx
	bra	setflag
	
; case '#' -- valido solo nello stato stage = 0
case_af:
	bne	case_dc			; abandon
	tax				; X = stage
	lda	#F_ALTFMT		; flag formato alt. 
setflag:
	tsb	FmtFlag
	txa				; A = stage
	clc				; OK
	clv				; continua scanning
	rts

; case ',' -- valido solo nello stato stage = 0
case_tg:
	bne	case_dc			; abandon
	tax				; X = stage
	lda	#F_GROUP		; flag group&thousand 
	bra	setflag
	
; case '*' - argomento per width o precision
case_ar:
	xba				; salva stage
	INDEX16
	ldx	FmtParm			; indice parametro byte
	lda	!0,x
	inx
	stx	FmtParm			; next param
	INDEX08
	tax				; X = param
	xba				; A = stage
	cmp	#WIDESTAGE
	bcs	?04			; campo precision ?
	stx	FmtWidth		; store width
	lda	#F_WIDTH
	tsb	FmtFlag			; imposta flag campo 'width'
	lda	#DOTSTAGE		; next: campo precisione
	clc				; OK
	clv				; continua scanning
	rts
?04:	cmp	#PRECSTAGE
	bne	case_dc			; abandon
	stx	FmtPrec			; store precision
	tax				; X = stage
	lda	#F_PREC
	tsb	FmtFlag			; imposta flag campo 'precision'
	txa				; A = stage (mantiene stesso stage)
	clc				; OK
	clv				; continua scanning
	rts	
	
; caratteri illegali - abbandona conversione
case_dc:
	sec				; flag abandon
	rts

; case '0'..'9' - digit che fanno parte o di 'width' o di 'precision'
; FmtWidth e FmtPrec sono inizializzati a ZERO quindi si somma il digit
; direttamente al valore corrente
case_nu:
	tax				; X = stage
	xba				; A = char '0'..'9'
	sec
	sbc	#'0'			; scala a 0..9
	cpx	#WIDESTAGE+1		; parte di specifica width ?
	bcs	?10			; no - parte di specifica precision
	tax				; X = number 0..9
	lda	FmtWidth
	stx	FmtWidth
	beq	?04			; old width = 0 
	asl	a			; width * 2
	bcs	?20			; overflow
	sta	FMTTMP
	asl	a
	bcs	?20			; overflow
	asl	a			; width * 8
	bcs	?20			; overflow
	adc	FMTTMP			; width * 10
	bcs	?20			; overflow
	adc	FmtWidth		; (width * 10) + number
	bcs	?20			; overflow
	sta	FmtWidth
?04:	lda	#F_WIDTH
	tsb	FmtFlag			; imposta flag campo 'width'
	lda	#WIDESTAGE		; current stage: campo numerico 'width'
	clc				; OK
	clv				; continua scanning
	rts
?10:	cpx	#PRECSTAGE		; stage precision ? 
	bne	case_dc			; no -- abandon
	tax				; X = number 0..9
	lda	FmtPrec
	stx	FmtPrec
	beq	?14			; old prec = 0 
	asl	a			; prec * 2
	bcs	?20			; overflow
	sta	FMTTMP
	asl	a
	bcs	?20			; overflow
	asl	a			; prec * 8
	bcs	?20			; overflow
	adc	FMTTMP			; prec * 10
	bcs	?20			; overflow
	adc	FmtPrec			; (prec * 10) + number
	bcs	?20			; overflow
	sta	FmtPrec
?14:	xba				; mantiene stesso stadio
	clc				; OK
	clv				; continua scanning
?20:	rts

; case '.' -- valido solo se stato corrente < precstage
case_pr:
	cmp	#PRECSTAGE
	bcs	?01			; abandon
	lda	#F_PREC
	tsb	FmtFlag			; imposta flag campo precision
	lda	#PRECSTAGE		; imposta stato corrente = precstage
	clc				; OK
	clv				; continua scanning
?01:	rts

;============================================================================
; modificatori -- impostano lo stato MODSTAGE (dopo sono ammessi solo i tipi)

; case 'l' -- modificatore 'long'
case_lo:
	ldx	#F2_LONG
	bra	setmod

; case 'h' -- modificatore 'short'
case_sh:
	ldx	#F2_SHORT
	bra	setmod

; case 'b' -- modificatore 'byte'
case_SH:
	ldx	#F2_BYTE
	bra	setmod

; case 'D' -- modificatore BCD float number
case_bc:
	ldx	#F2_BCD
	bra	setmod
	
; case 'B' -- modificatore stringa pascal/basic
case_PS:
	ldx	#F2_PSSTRING

setmod:
	cmp	#MODSTAGE
	bcs	?10			; abandon
	txa
	tsb	FmtMod	
	lda	#MODSTAGE		; stadio finale specificatore
	clc	; OK
	clv	; continua scanning
?10:	rts
	
;============================================================================
; specificatori di tipo -- qualsiasi tipo a qualsiasi stadio termina
; lo scanning 

; case 'P' -- pointer HEX
case_PT:
	lda	#F2_CAPS		; set CAPS
	tsb	FmtMod

; case 'p' -- pointer hex
case_pt:
	lda	#F2_SIGNED
	trb	FmtMod			; clear flag signed
	lda	#F_FILLZ		; ignora padding a sx. con '0'
	trb	FmtFlag
	ldx	#T_PTR			; indice funzione conversione
	lda	#MASKPTR		; maschera modificatori non ammessi
	bra	tsttype
	
; case 'X' -- unsigned HEX
case_HE:
	lda	#F2_CAPS
	tsb	FmtMod
	
; case 'x' -- unsigned hex
case_he:
	lda	#F2_SIGNED
	trb	FmtMod			; clear flag signed
	lda	FmtFlag
	bit	#F_PREC
	beq	?02
	lda	#F_FILLZ		; ignora padding a sx. con '0'
	trb	FmtFlag			; se specificata precision
?02:	ldx	#T_HEX			; indice funzione conversione
	lda	#MASKINTG		; maschera modificatori non ammessi
	bra	tsttype

; case 'd', 'i' -- signed integer
case_de:
	lda	#F2_SIGNED
	tsb	FmtMod			; set flag signed
	bra	case_i

; case 'u' -- unsigned integer
case_un:
	lda	#F2_SIGNED
	trb	FmtMod			; clear flag signed

; parte comune per gli interi decimali
case_i:
	lda	FmtFlag
	bit	#F_PREC
	beq	?02
	lda	#F_FILLZ		; ignora padding a sx. con '0'
	trb	FmtFlag			; se specificata precision
?02:	lda	#F_ALTFMT
	trb	FmtFlag			; decimale ignora formato alternativo
	ldx	#T_INT			; indice funzione conversione
	lda	#MASKINTG		; maschera modificatori non ammessi
	bra	tsttype

; case 'c' -- char
case_ch:
	lda	#(F_ALTFMT.OR.F_PREC)	; ignora precision e formato alt.
	trb	FmtFlag
	ldx	#T_CHAR			; indice funzione conversione
	lda	#MASKCHAR		; maschera modificatori non ammessi
	bra	tsttype
	
; case 's' -- string
case_st:
	lda	#F_ALTFMT		; ignora formato alt.
	trb	FmtFlag
	ldx	#T_STR			; indice funzione conversione
	lda	#MASKSTR		; maschera modificatori non ammessi
	bra	tsttype

; case 'E' -- float formato EXP
case_FE:
	lda	#F2_CAPS
	tsb	FmtMod

; case 'e' -- float formato exp
case_fe:
	lda	#$00			; flag formato e/E
	bra	case_f
	
; case 'G' -- float formato G
case_FG:
	lda	#F2_CAPS
	tsb	FmtMod

; case 'g' -- float formato g
case_fg:
	lda	#$40			; flag formato g/G
	bra	case_f
	
; case 'A' -- float formato A
case_XF:
	lda	#F2_CAPS
	tsb	FmtMod

; case 'a' -- float formato a
case_xf:
	lda	#F_FILLZ		; formato a/A no padding a sx con '0'
	trb	FmtFlag
	lda	#$04			; flag formato a/A
	bra	case_f

; case 'f', 'F' -- float formato decimale
case_ff:
	lda	#$20			; flag formato f/F

; parte comune ai numeri float
case_f:
	; i numeri float possono avere padding a sx con '0' anche se viene
	; specificata la precision
	sta	XVFlag			; flag formattazione
	lda	#F2_SIGNED
	tsb	FmtMod			; set flag signed	
	ldx	#T_FLOAT		; indice funzione conversione	
	lda	#MASKFLOAT		; maschera modificatori non ammessi

; testa modificatore tipo
tsttype:
	bit	FmtMod			; test modificatori non ammessi
	beq	?04			; OK
	sec				; abandon
	rts
?04:	sep	#PVFLAG			; set V = 1 -- fine scanning 
	clc				; OK  -> conversione
	rts
		
;----------------------------------------------------------------------------
; COSTANTI
;----------------------------------------------------------------------------

; tabella classi caratteri di conversione (carattere >= 32 e < 128)
CLASSCHR:

;            SP   !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /
	.DB _si,_dc,_dc,_af,_dc,_dc,_dc,_dc,_dc,_dc,_ar,_si,_tg,_lj,_pr,_dc

;	      0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?
	.DB _fz,_nu,_nu,_nu,_nu,_nu,_nu,_nu,_nu,_nu,_dc,_dc,_dc,_dc,_dc,_dc

;             _   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
	.DB _dc,_XF,_PS,_dc,_bc,_FE,_ff,_FG,_dc,_dc,_dc,_dc,_dc,_dc,_dc,_dc

;             P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _ 
	.DB _PT,_dc,_dc,_dc,_dc,_dc,_dc,_dc,_HE,_dc,_dc,_dc,_dc,_dc,_dc,_dc

;             `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o
	.DB _dc,_xf,_SH,_ch,_de,_fe,_ff,_fg,_sh,_de,_dc,_dc,_lo,_dc,_dc,_dc

;             p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~ DEL 
	.DB _pt,_dc,_dc,_st,_dc,_un,_dc,_dc,_he,_dc,_dc,_dc,_dc,_dc,_dc,_dc

; tabella funzioni scanning formato
SCANTBL:
	.DW	case_dc, case_si, case_af, case_ar
	.DW	case_lj, case_pr, case_nu, case_lo
	.DW	case_sh, case_SH, case_fz, case_de
	.DW	case_un, case_he, case_pt, case_HE
	.DW	case_PT, case_ch, case_st, case_ff
	.DW	case_fe, case_fg, case_FE, case_FG
	.DW	case_tg, case_bc, case_PS, case_xf
	.DW	case_XF
	
; tabella funzioni di conversione
CVTTBL:
	.DW	DoChar, DoStr, DoPtr, DoInt, DoHex, DoFloat
	
NULLSTR:
	.DB	"(null)", 0
	