;----------------------------------------------------------
; ACIA.ASM
; PROGETTO: B1601
;
; Gestione ACIA
;----------------------------------------------------------

_ACIA_INC_	.SET	1


	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc


	.CODEF9
	.GLOBAL _ACIA1Get, _ACIA1Put, _ACIA2Get, _ACIA2Put
	
	.LONGA	off
	.LONGI	off

; se CF = 0 -> A = byte prelevato dalla coda
; se CF = 1 -> A = codice errore RX (se = $00 -> coda vuota)
_ACIA1Get:
	phd				; salva DP
	clc				; suppone no error
	phb				; salva DBR
	php				; salva status
	pea	#0			; imposta DP = $00
	pld
	ACC08
	lda	#0			; imposta DBR = $00
	pha
	plb
	INDEX16				; salva X a 16 bit
	phx				; salva X
	sei				; disabilita IRQ
	lda	<ACIArxe_1		; test errori RX
	bmi	?10			; error -- esce
?00:	ldx	<ACIAICnt_1		; numero bytes in buffer input
	bne	?01			; coda RX non vuota
	; coda RX vuota - test disconnessione remota
	bra	?08
	;lda	<ACIAmsk_1		; test DSR o DCD
	;and	<ACIAdsrl_1		; test DSR = 1 o DCD = 1
	;beq	?08			; no disconnessione remota
	;lda	#$A0			; update errore (bit 5 -> disconnessione)
	;tsb	<ACIArxe_1
	;bra	?08
?01:	dex				; update count
	stx	<ACIAICnt_1
	ldx	<ACIAIHead_1		; ptr. testa buffer input
	lda	>ACIAINBUF1,x		; preleva carattere
	sta	<ACIATmp		; salva byte
	inx				; update ptr
	cpx	#ACIAIBUFLEN		; coda circolare
	bcc	?03
	ldx	#0
?03:	stx	<ACIAIHead_1
	
	; nel caso di handshake occorre cancellare la pausa
	; remota se il buffer di input ha posto libero
	bit	<ACIAPause_1		; pausa remota attiva ?
	bpl	?14			; no
	bit	<ACIAhsk_1		; hsk attivo ?
	bpl	?14			; no handshake
	
	;bra	?14
	
	ldx	<ACIAICnt_1		; X = count bytes in coda
	cpx	#NGUARD3		; coda con posto disponibile ?
	bcs	?14			; no
	bvs	?06			; hsk hardware	
	lda	#SPXON			; cancella pausa remota
	bit	!ACIA+ACIACSR1		; flag TUR (TX underrun) ?
	bvc	?04			; no -- invia con TX IRQ
	sta	!ACIA+ACIATDR1		; invia XON
	bra	?07
?04:	sta	<ACIAOut_1		; invia XON
	bra	?07
?06:		
	lda	<ACIAfr_1
	and	#$FE
	ora	#$80
	sta	<ACIAfr_1
	sta	!ACIA+ACIAFR1

?07:	lda	#$80			; cancella flag pausa remota
	trb	<ACIAPause_1
	
	inc	!$ff
	
	bra	?14			; esce con CF = 0	
	
	; coda RX vuota -- test errori RX pendenti
?08:	lda	<ACIArxe_1		; errori RX (= $00 no error)
?10:	sta	<ACIATmp		; se = $00 -> coda RX vuota	
	lda	$03,s			; P nello stack
	ora	#PCFLAG			; imposta C -- errore
	sta	$03,s		 	
?14:	lda	<ACIATmp		; return byte o codice errore
	plx				; restore X
	plp				; restore status
	plb
	pld				; restore DP
	rtl

; CF = 0 -> A = byte inviato
; CF = 1 -> A = codice errore, B = byte in out
;	A = $FF -> disconnessione remota
;	A = $7F -> overflow buffer TX (pausa locale ?)
_ACIA1Put:
	phd				; salva DP
	clc				; suppone no error
	phb				; salva DBR
	php				; salva status
	pea	#0			; imposta DP = 0
	pld
	ACC08
	sta	<ACIATmp		; salva byte in out
	lda	#0			; imposta DBR = $00
	pha
	plb
	INDEX16
	phx				; salva X a 16 bit
	sei
	; test disconnessione remota
	;lda	<ACIAmsk_1		; test DSR o DCD
	;and	<ACIAdsrl_1		; test DSR = 1 o DCD = 1
	;bne	?12			; disconnessione remota
	lda	<ACIATmp		; byte in out
	ldx	<ACIAOCnt_1
	cpx	#ACIAOBUFLEN
	bcc	?10			; OK --  room disponibile in buffer TX
	bit	<ACIAPause_1		; test pausa locale
	bvs	?13			; overflow TX in pausa locale
	bit	!ACIA+ACIACSR1		; flag TUR (TX underrun) ?
	bvc	?13			; no -- overflow buffer TX 
	lda	#$01			; segnala condizione overflow
	tsb	<ACIAPause_1
	bra	?11a			; libera 1 posto in coda TX
?10:	inx				; update count
	stx	<ACIAOCnt_1
	ldx	<ACIAOTail_1
	sta	>ACIAOUTBUF1,x		; store byte in coda TX
	inx				; update Tail
	cpx	#ACIAOBUFLEN		; coda circolare
	bcc	?11
	ldx	#0	
?11:	stx	<ACIAOTail_1	
	bit	!ACIA+ACIACSR1		; flag TUR (TX underrun) ?
	bvc	?16			; no -- invia con TX IRQ (coda TX)
	bit	<ACIAPause_1		; test pausa locale
	bvs	?16			; TX in pausa locale -- non invia
?11a:	ldx	<ACIAOHead_1		; indice testa buffer
	lda	>ACIAOUTBUF1,x
	sta	!ACIA+ACIATDR1		; invia dato
	inx				; aggiorna puntatore
	ACC16
	txa
	and	#(ACIAOBUFLEN-1)	; coda circolare
	sta	<ACIAOHead_1
	dec	<ACIAOCnt_1
	ACC08
	lda	#$01			; test overflow
	trb	<ACIAPause_1
	bne	?13			; condizione di overflow
	lda	<ACIATmp		; A = byte inviato
	bra	?16
	; imposta errore -- A = flag errore
?12:	lda	#$FF			; $FF -> disconnessione remota
	bra	?14
?13:	lda	#$7F			; $7F -> overflow buffer TX
?14:	xba				; B = flag errore	
	lda	$03,s			; P nello stack
	ora	#PCFLAG			; imposta C -- errore
	sta	$03,s
	lda	<ACIATmp		; byte in out
	xba				; A = flag errore - B = byte in out
?16:	plx				; restore X
	plp				; restore status
	plb				; restore DBR
	pld				; restore DP
	rtl

; se CF = 0 -> A = byte prelevato dalla coda
; se CF = 1 -> A = codice errore RX (se = $00 -> coda vuota)
_ACIA2Get:
	phd				; salva DP
	clc				; suppone no error
	phb				; salva DBR
	php				; salva status
	pea	#0			; imposta DP = $00
	pld
	ACC08
	lda	#0			; imposta DBR = $00
	pha
	plb
	INDEX16				; salva X a 16 bit
	phx				; salva X
	sei				; disabilita IRQ
	lda	<ACIArxe_2		; test errori RX
	bmi	?10			; error -- esce
?00:	ldx	<ACIAICnt_2		; numero bytes in buffer input
	bne	?01			; coda RX non vuota
	; coda RX vuota - test disconnessione remota
	;lda	<ACIAmsk_2		; test DSR o DCD
	;and	<ACIAdsrl_2		; test DSR = 1 o DCD = 1
	;beq	?08			; no disconnessione remota
	lda	#$A0			; update errore (bit 5 -> disconnessione)
	tsb	<ACIArxe_2
	bra	?08
?01:	dex				; update count
	stx	<ACIAICnt_2
	ldx	<ACIAIHead_2		; ptr. testa buffer input
	lda	>ACIAINBUF2,x		; preleva carattere
	sta	<ACIATmp		; salva byte
	inx				; update ptr
	cpx	#ACIAIBUFLEN		; coda circolare
	bcc	?03
	ldx	#0
?03:	stx	<ACIAIHead_2
	
	; nel caso di handshake occorre cancellare la pausa
	; remota se il buffer di input ha posto libero
	bit	<ACIAPause_2		; pausa remota attiva ?
	bpl	?14			; no
	bit	<ACIAhsk_2		; hsk attivo ?
	bpl	?14			; no handshake
	ldx	<ACIAICnt_2		; X = count bytes in coda
	cpx	#NGUARD3		; coda con posto disponibile ?
	bcs	?14			; no
	bvs	?06			; hsk hardware	
	lda	#SPXON			; cancella pausa remota
	bit	!ACIA+ACIACSR2		; flag TUR (TX underrun) ?
	bvc	?04			; no -- invia con TX IRQ
	sta	!ACIA+ACIATDR2		; invia XON
	bra	?07
?04:	sta	<ACIAOut_2		; invia XON
	bra	?07
?06:	;lda	<ACIArem_2
	;eor	#$FF			; inverte i bit OR
	;and	<ACIAfr_2		; RTS = 0 oppure DTR = 0 
	;sta	<ACIAfr_2
	;sta	!ACIA+ACIAFR2
?07:	lda	#$80			; cancella flag pausa remota
	trb	<ACIAPause_2
	bra	?14			; esce con CF = 0	
	
	; coda RX vuota -- test errori RX pendenti
?08:	lda	<ACIArxe_2		; errori RX (= $00 no error)
?10:	sta	<ACIATmp		; se = $00 -> coda RX vuota	
	lda	$03,s			; P nello stack
	ora	#PCFLAG			; imposta C -- errore
	sta	$03,s		 	
?14:	lda	<ACIATmp		; return byte o codice errore
	plx				; restore X
	plp				; restore status
	plb
	pld				; restore DP
	rtl

; CF = 0 -> A = byte inviato
; CF = 1 -> A = codice errore, B = byte in out
;	A = $FF -> disconnessione remota
;	A = $7F -> overflow buffer TX (pausa locale ?)
_ACIA2Put:
	phd				; salva DP
	clc				; suppone no error
	phb				; salva DBR
	php				; salva status
	pea	#0			; imposta DP = 0
	pld
	ACC08
	sta	<ACIATmp		; salva byte in out
	lda	#0			; imposta DBR = $00
	pha
	plb
	INDEX16
	phx				; salva X a 16 bit
	sei
	; test disconnessione remota
	;lda	<ACIAmsk_2		; test DSR o DCD
	;and	<ACIAdsrl_2		; test DSR = 1 o DCD = 1
	;bne	?12			; disconnessione remota
	lda	<ACIATmp		; byte in out
	ldx	<ACIAOCnt_2
	cpx	#ACIAOBUFLEN
	bcc	?10			; OK --  room disponibile in buffer TX
	bit	<ACIAPause_2		; test pausa locale
	bvs	?13			; overflow TX in pausa locale
	bit	!ACIA+ACIACSR2		; flag TUR (TX underrun) ?
	bvc	?13			; no -- overflow buffer TX 
	lda	#$01			; segnala condizione overflow
	tsb	<ACIAPause_2
	bra	?11a			; libera 1 posto in coda TX
?10:	inx				; update count
	stx	<ACIAOCnt_2
	ldx	<ACIAOTail_2
	sta	>ACIAOUTBUF2,x		; store byte in coda TX
	inx				; update Tail
	cpx	#ACIAOBUFLEN		; coda circolare
	bcc	?11
	ldx	#0	
?11:	stx	<ACIAOTail_2	
	bit	!ACIA+ACIACSR2		; flag TUR (TX underrun) ?
	bvc	?16			; no -- invia con TX IRQ (coda TX)
	bit	<ACIAPause_2		; test pausa locale
	bvs	?16			; TX in pausa locale -- non invia
?11a:	ldx	<ACIAOHead_2		; indice testa buffer
	lda	>ACIAOUTBUF2,x
	sta	!ACIA+ACIATDR2		; invia dato
	inx				; aggiorna puntatore
	ACC16
	txa
	and	#(ACIAOBUFLEN-1)	; coda circolare
	sta	<ACIAOHead_2
	dec	<ACIAOCnt_2
	ACC08
	lda	#$01			; test overflow
	trb	<ACIAPause_2
	bne	?13			; condizione di overflow
	lda	<ACIATmp		; A = byte inviato
	bra	?16
	; imposta errore -- A = flag errore
?12:	lda	#$FF			; $FF -> disconnessione remota
	bra	?14
?13:	lda	#$7F			; $7F -> overflow buffer TX
?14:	xba				; B = flag errore	
	lda	$03,s			; P nello stack
	ora	#PCFLAG			; imposta C -- errore
	sta	$03,s
	lda	<ACIATmp		; byte in out
	xba				; A = flag errore - B = byte in out
?16:	plx				; restore X
	plp				; restore status
	plb				; restore DBR
	pld				; restore DP
	rtl
	