;----------------------------------------------------------
; BIOSMEM.ASM
; PROGETTO: B1601
;
; Gestione memoria dinamica BIOS
;----------------------------------------------------------


	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc

_last	.EQU	$0000
_rover	.EQU	$0002

bsize	.EQU	$0000
bprev	.EQU	$0002
fprev	.EQU	$0004
fnext	.EQU	$0006

HEAPSIZ	.EQU	4
FREESIZ	.EQU	8
USEDSIZ	.EQU	4
PFIRST	.EQU	(HEAPSIZ + USEDSIZ)

	.CODEF9

	.LONGA	on
	.LONGI	on

; alloc: X = size, Y = heap offset
getmem:
	txa
	beq	?06		; null size
	clc
	adc	#USEDSIZ+1
	bcs	?08		; ovwerflow
	and	#$FFFE		; word aligned
	cmp	#FREESIZ
	bcs	?02
	lda	#FREESIZ
?02:	ldx	!_rover,y	; any free block?
	beq	?06
	phy
?04:	cmp	!bsize,x
	beq	?14		; exact size
	bcc	?10		; ok
	txy
	ldx	!fnext,y
	bne	?04
	ply
?06:	sec
?08:	rts
?10:	pha			; size
	lda	!bsize,x
	sec
	sbc	$01,s
	cmp	#FREESIZ
	bcc	?12		; no split
	sta	!bsize,x	; new free block size
	phx
	clc
	adc	$01,s
	tax			; pointer to splitted block
	pla			
	sta	!bprev,x
	pla
	inc	a		; mark used block
	sta	!bsize,x
	txa
	ply
	bra	?22
?12:	pla			; discard size
?14:	lda	!fprev,x
	ldy	!fnext,x
	beq	?16
	sta	!fprev,y
?16:	tay
	beq	?18
	lda	!fnext,x
	sta	!fnext,y
?18:	ply
	txa
	cmp	!_rover,y
	bne	?20
	lda	!fnext,x
	sta	!_rover,y
	txa
?20:	inc	!bsize,x
?22:	cmp	!_last,y
	bcc	?24
	sta	!_last,y
	clc
?24:	inx
	inx
	inx
	inx
_?26:	rts

; free: X = pointer, Y = heap offset
freemem:
	jsr	blksiz
	bcs	_?26

freemem2:
	sta	!bsize,x
	phy
	lda	!_last,y
	pha
	lda	#$FFFF
	sta	!fprev,x	; invalid pointer
	ldy	!bprev,x
	beq	?06		; no left block
	lda	!bsize,y
	lsr	a
	bcs	?06		; left block no free
	txa
	cmp	$01,s		; last pointer
	bcc	?02
	tya			; Y become last pointer
	sta	$01,s
	clc
	bra	?04
?02:	adc	!bsize,x
	tay
	lda	!bprev,x
	sta	!bprev,y
	tay			; Y = previous
?04:	lda	!bsize,y
	adc	!bsize,x
	sta	!bsize,y
	tyx
?06:	txa
	cmp	$01,s
	bcc	?16		; no right block
	clc
	adc	!bsize,x
	tay			; Y = next
	lda	!bsize,y
	lsr	a
	bcs	?16		; right block not free
	tya
	cmp	$01,s
	bcc	?08		; not the last pointer
	txa
	sta	$01,s		; X = last pointer
	bra	?10
?08:	adc	!bsize,y
	phy
	tay
	txa
	sta	!bprev,y
	ply
?10:	clc
	lda	!bsize,x
	adc	!bsize,y
	sta	!bsize,x
	phx			; unlink Y from free list
	lda	!fprev,y
	ldx	!fnext,y
	beq	?12
	sta	!fprev,x
?12:	txa
	beq	?14
	lda	!fnext,y
	sta	!fprev,x
?14:	plx
?16:	pla			; last pointer
	ply
	sta	!_last,y
	lda	!fprev,x
	inc	a
	bne	?18		; already in free list
	lda	!_rover,y
	sta	!fnext,x
	stz	!fprev,x
	txa
	sta	!_rover,y
?18:	rts

; verf pointer: X = pointer, Y = heap ofs

hlst	.SET	$01
hptr	.SET	$03
nsiz	.SET	$05
blksiz:
	phy			; save heap offset
	pha			; local for nsiz
	phx			; local for hptr
	lda	!_last,y
	pha			; local for hlst
	clc
	tya
	adc	#PFIRST		; first valid pointer
	sta	nsiz,s
	txa
	cmp	nsiz,s
	bcc	?10		; error
	dex
	dex
	dex
	dex
	txa
	sta	hptr,s
	cmp	hlst,s
	beq	?02
	bcs	?12		; error (> last pointer)
?02:	clc	
	lda	!bprev,x
	bne	?04
	tya			; check if first pointer
	adc	#HEAPSIZ
	cmp	hptr,s
	bne	?10		; error
	beq	?06		; check next block
?04:	tay			; previous block
	sta	nsiz,y		; here C = 0
	lda	!bsize,y
	and	#$FFFE
	adc	nsiz,s
	cmp	hptr,s
	bne	?10		; error
?06:	lda	!bsize,x
	lsr	a
	bcc	?10		; block is free
	asl	a		; here C = 0
	sta	nsiz,s		; size of block
	txa
	cmp	hlst,s	
	beq	?08
	adc	hptr,s		; here C = 0
	tay
	lda	!bprev,y
	cmp	hptr,s
	bne	?10		; error
?08:	clc
	.DB	$24		; fake 'bit $38'
?10:	sec
?12:	pla			; discard hlst
	pla			; discard hptr
	pla			; size of block
	ply
	rts
