;;
;; Copyright (c) 2016 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; name: fatdir.asm	
;; rev.: 2016/22/10
;; o.s. 65C816 version v1.0

;; fat directory handler

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	;.INCLUDE inc\sys_fs.inc
	.INCLUDE inc\tossys.inc

.LIST on

;---------------------------------------------------------------------------
; code segment -- bank $F9
;---------------------------------------------------------------------------

	.CODEF9
	.EXTERN oserror, getldt
	.EXTERN mxdrv, fndhcb, fndbuf, cls2lba			; utils.asm
	
	.PUBLIC prslbl

	.LONGA	off
	.LONGI	off

;---------------------------------------------------------------------------
; path parse & build
;---------------------------------------------------------------------------

; pathdrv, pathcnt & pathnxt functions
; run down the path and parse final name into fcbs
; set directory cluster to last but one path name
;
;	entry:	pathp  = long pointer to path string
;
;	exit:	dircls = start cluster of the last but one path name
;			 if CF=0 (no error)
;
;	if any error, CF=1 and A=error code
;	error code:	ED_FILE		(invalid filename/file not found)	
;			ED_PATH 	(path not found)
;			ED_ACCESS	(access denied)
;			ED_FORMAT	(invalid format)
;			ED_DRIVE	(invalid drive)
;			ED_NOTRDY	(drive not ready)
;	use: all
;

;-------
setpchk:	; special entry for directory handler function's
;-------
	jsr	setpath		; parse path string & build path
	jsr	nodev		; devices not allowed
	jsr	nodot		; no subdirs entries
	jsr	nowild		; wild cards not allowed
	rts

;-------
setpath:
;-------
	jsr	pathdrv
	bcc	phx
	jmp	oserror

;-------
pathdrv:	; get drive number from path or use default drive
;-------
	jsr	getpdrv		; return drive number in A
	bcc	pathcnt
	lda	#ED_DRIVE	; invalid drive error
phx:	rts
	
;-------
pathcnt:	; use drive in A
;-------
	jsr	mxdrv		; select drive in A
	ldy	strix		; path string index
	lda	[pathp],y	; get first character
	iny
	sty	strix
	jsr	chkslh		; if '\' or '/' then start at root...
	bne	?20		; ...else select current working directory
	jsr	fakefcb		; fake a '.' entry for the root
	stx	dircls		; start at fake cluster 0
	stx	dircls+1
	lda	[pathp],y	; get next character
	beq	?10		; if just a '\' or '/' then stop now
	jsr	chkslh		; detected '//' or '\\' ?
	bne	pathnxt		; no, so start processing from root dir.
	lda	#ED_ACCESS	; report 'access denied' error
	sec
	rts
?10:	clc			; no error
	rts
?20:	dey			; forget about char we looked at
	sty	strix
	jsr	mvtobp		; start processing from cur. dir.

;-------
pathnxt:	; parse&build path
;-------
	ldy	strix
?10:	lda	[pathp],y	; can't have trailing '/' or '\'
	beq	?15		; error
	jsr	prspath		; parse path & set up fcb
	bcs	?15		; error
	tax			; A=delimiter
	beq	?20		; we are at the end
	ACC16
	lda	fcbs
	cmp	#' .'		; if chdir(".")...
	ACC08
	beq	?10		; ...stay where we are
	lda	haswld		; no wilds cards in path's
	bne	?15		; skip if wild cards found
	jsr	fndfcbf		; locate the directory entry
	bcs	?15		; missing directory in path: path not found
	lda	dfcbatt		; test entry attribute
	bit	#DA_DIR
	beq	?15		; error: not a directory
	jsr	setdir		; set dir. cluster
	bcc	?10		; next path name
?15:	lda	#ED_PATH	; 'path not found' error
	sec
	rts
?20:	lda	fcbs		; is it '.' or '..' ?
	cmp	#'.'		; if so get its full name
	bne	?35		; exit, no error
	ACC16
	lda	dircls		; we are at root?
	beq	?30		; yes, exit, no error
	ldy	bplvl
	beq	?30		; no error
	dey			; go up one level
	lda	fcbs
	cmp	#'..'		; parent?
	bne	?25		; no, stay here
	dey			; go up two level's
	bmi	?40		; we are destined for the root
?25:	sty	bplvl
	phb			; save DBR
	ldx	#WKB
	phx
	plb
	ldx	#0
	INDEX16
	tya
	asl	a		; get HPL pointer
	asl	a
	asl	a
	asl	a		; here CF=0
	;adc	#cbpl
	tay			; HPL index access
	lda	HPL_CLS,y
	sta	dircls		; set up parent cluster
	lda	HPL_IX,y
	sta	dcnt		; set up dir. index
?28:	lda	HPL_FCB,y	; copy parental name
	sta	fcbs,x
	iny
	inx
	cpx	#11
	bcc	?28
	plb
?30:	CPU08
?35:	clc
	rts
	
?40:	stz	dircls		; root dir	
	CPU08
	jsr	fakefcb
	clc
	rts
	
; parse path name into fcb
;
;	entry:	Y = pointer to current name in path
;		pathp = long pointer to path
;
;	exit:	A  = last char parsed
;		Y  = pointer to next name in path
;		CF = 1 if parsing error
;
;-------
prspath:
;-------
	jsr	clrfcb		; initialise fcb to blanks
	stz	haswld		; clear wild cards flag
	lda	[pathp],y
	cmp	#'.'		; if name start with '.'...
	bne	?20		; ...then we parse it differently
	sta	fcbs		; move '.' to fcb
	iny
	lda	[pathp],y
	iny
	cmp	#'.'		; if name start with '..'...
	bne	?15		; ...then we parse it differently
	sta	fcbs+1		; move '..' to fcb
?10:	lda	[pathp],y
	iny
?15:	cmp	#' '		; skip all subsequent blanks
	beq	?10
	bra	?50		; check delimiter
?20:	jsr	chkdlm		; filename begins with a legal char?
	beq	?40		; no, then only allow '/' or '\'
	stz	wild		; yes, then parse it normally
	ldx	#0		; fcb name field ptr
	lda	#8		; length of name field
	jsr	prsone		; parse just name field
	bcs	?90		; error
	ldx	wild
	sta	haswld		; save name wildcard flag
?30:	cmp	#'.'
	bne	?50		; skip if not extension
	stz	wild
	ldx	#8		; fcb extension field ptr
	lda	#3		; length of extension field
	jsr	prsone		; parse just extension field
	bcs	?90		; error
	sta	tmpx
	lda	wild		; combine wild cards flag
	ora	haswld
	sta	haswld
	lda	tmpx
	dey
?40:	iny			; point to next character
?50:	tax			; last parsed delimiter
	beq	?90		; end of path: ok
	jsr	chkslh		; if delimiter is not '\' nor '/'...
	bne	?90		; ...error (CF=1)
?60:	lda	[pathp],y	; get next character
	iny
	jsr	chkdlm		; here we expect a normal legal character
	bne	?80		; exit if we've got one
	jsr	chkslh		; swallow '\'s at this point and leave...
	beq	?60		; ...other delimiters for next time
	cmp	#'.'		; trailing '\.' ?
	bne	?75		; no
	tyx			; remember position of '.'
?70:	lda	[pathp],y	; get next character
	beq	?90		; stop at ending of string
	iny
	cmp	#' '		; now discard trailing spaces
	beq	?70		; keep going until we lose all spaces
	jsr	chkslh		; if it's a '\' or '/' try again
	beq	?60
	txy			; retract to the '.'
?75:	lda	#'/'		; return '/' as the delimiter...
	clc			; ...and exit with no problems
?80:	dey			; retract a byte
?90:	rts

; parse a single name or extension
;
;	entry:	A = field size
;		X = local fcb field pointer
;		Y = pointer to current name in path
;		pathp = long pointer to path
;
;	exit:	A  = last char parsed
;		Y  = pointer to next name in path
;		CF = 1 if parsing error
;
;	use:	A,X,Y
;
;------
prsone:
;------
	sta	fsiz		; char's counter
	stx	fptr		; field start
?10:	lda	[pathp],y	; get next path char
	iny
	jsr	chkdlm
	beq	?20		; if it is a delimiter then return
	cmp	#'*'
	beq	?25		; special case if star
	cmp	#'?'		; is a wildcard?
	bne	?15		; no
	lda	#$80
	tsb	wild		; set wildcard flag
	lda	#'?'
?15:	sta	fcbs,x		; move char to fcb field
	inx
	dec	fsiz
	bne	?10		; loop: next path char
	lda	[pathp],y	; get next path char
	iny
	jsr	chkdlm		; here expected a delimiter
	beq	?20
	sec			; error flag
?20:	rts
?25:	bit	wild		; already parsed a star?
	bvs	?38		; yes, so error	
	lda	#$C0
	tsb	wild		; set wildcard+star flag
	cpx	fptr		; star is first char of the field?
	beq	?40		; yes
?28:	lda	#'?'		; if not, fill rest of field with '?'
?30:	sta	fcbs,x
	inx
	dec	fsiz
	bne	?30
?35:	lda	[pathp],y	; get next path char
	iny
	jsr	chkdlm		; here expected a delimiter...
	beq	?20
	cmp	#'*'		; ...or just star...
	beq	?35
?38:	sec			; ...else error
	rts
?40:	lda	[pathp],y	; get next path char
	iny
	cmp	#'*'		; skips all subsequent star's
	beq	?40
	jsr	chkdlm		; is a delimiter?
	bne	?45		; no
	dey			; forgot last char
	bra	?28		; fill rest of field with '?'
?45:	lda	#'*'		; first char of field is a star
	sta	fcbs,x
	inx
	dec	fsiz
	dey			; forgot last char
	bra	?10		; continue normal parsing

; parse a single label name (volume name)
;
;	entry:	pathp = long pointer to label name
;
;	exit:	CF = 1 if parsing error
;		ZF = 1 if null string
;
;	use:	A,X,Y
;
;------
prslbl:
;------
	jsr	clrfcb		; set up a clean fcb
	lda	#11
	sta	fsiz		; char's counter
	ldx	#0		; field start
	txy
	clc
	lda	[pathp]		; get first char
	beq	?30		; null string
	cmp	#' '		; reject if first char is blank
	beq	?15
	bra	?12
?10:	lda	[pathp],y	; get next name char
?12:	iny
	jsr	chkdlm
	bne	?20		; is not a delimiter
	cmp	#' '		; intermediate blanks are ok
	beq	?25
	cmp	#0		; end of string?
	beq	?28		; yes
?15:	sec			; reject delimiter
	rts
?20:	cmp	#'*'
	beq	?15		; reject wildcard
	cmp	#'?'
	beq	?15
?25:	sta	fcbs,x		; move char to fcb field
	inx
	dec	fsiz
	bne	?10		; loop: next path char
?28:	rep	#PZFLAG.OR.PCFLAG
	rts			; exit with ZF=0,CF=0
?30:	sep	#PZFLAG		; null string: exit with ZF=1,CF=0
	rts

; go down (or up) one level in directory
;
;	entry:	fcbs = local name
;		fcbp = long pointer to directory entry
;		bplvl = current path level
;
;	exit:	dircls = cluster of selected directory
;		bplvl = path level
;		CF = 1 if failure
;
;	use:	C,X
;
;------
setdir:
;------
	sty	strix		; save Y
	ldy	#11
	lda	[fcbp],y	; entry attribute
	bit	#DA_DIR		; check if directory
	sec
	beq	?30		; fail if this is a file
	lda	#WKB
	phb			; save current DBR
	pha			; set DBR bank
	plb
	ACC16
	ldy	#26
	lda	[fcbp],y	; first cluster of directory
	sta	wtmp
	ldx	bplvl		; current path level
	lda	fcbs
	cmp	#'..'		; watch out if going up a level
	beq	?10		; we go up one level
	txa
	inx			; increment level
	cpx	#HPLMAX+1	; level overflow?
	bcs	?25		; yes, return failure
	stx	bplvl
	ldy	#0		; fcb name pointer	
	CPU16
	asl	a		; mult. x 16 to get index
	asl	a
	asl	a
	asl	a		; here always CF=0
	adc	#cbpl
	tax			; point to HPL
	ACC08
	stz	!hpl_fcb,x
?05:	lda	[fcbp],y	; move fcb name
	sta	!hpl_fcb,x
	inx
	iny
	cpy	#11
	bcc	?05
	stz	!hpl_fcb,x	; last unused byte
	CPU16CLC
	lda	dircls		; set parent cluster
	sta	!hpl_cls,x
	lda	wtmp
	sta	dircls		; set current directory cluster
	lda	dcnt
	sta	!hpl_ix,x	; set index entry on parent cluster
	INDEX08
	bra	?25		; done (CF=0)
?10:	sec	
	dex			; decrement level
	bmi	?25		; already at level 0 (root) so failure
	stx	bplvl
	CPU16
	txa
	asl	a		; mult. x 16 to get index
	asl	a
	asl	a
	asl	a		; here always CF=0
	adc	#cbpl
	tax			; point to HPL of current dir.
	lda	wtmp		; parent cluster
	cmp	!hpl_cls,x	; will match...
	sec
	bne	?25		; ...otherwise return failure
	sta	dircls		; set cluster of selected directory
	clc
?25:	CPU08
	plb			; restore DBR
?30:	ldy	strix	
	rts

; check for a path name delimiter
;
; 	entry:	A  = path character
;
;	exit:	A  = upper case path character
;		CF = 0
;		ZF = 1 if character is a delimiter
;		ZF = 0 if character is legal in file name
;
;	use:	A
;
;------
chkdlm:
;------
	cmp	#$7F
	bcs	?40		; discard if ascii code > 126 (ZF=1)
	cmp	#' '+1
	bcc	?40		; discard blank & no-printable char (ZF=1)
	stx	tmpx		; save X
	ldx	#DLMSTR-1
?10:	cmp	>dlmstr,x	; match A against list
	beq	?50		; ZF=1
	dex
	bpl	?10
	ldx	tmpx		; restore X
	cmp	#'a'		; make upper case
	bcc	?20
	cmp	#'z'+1
	bcs	?20
	and	#$DF
?20:	rep	#PZFLAG.OR.PCFLAG	; ZF=0, CF=0 (no delim char)
	rts
?30:	ldx	tmpx		; restore X
?40:	sep	#PZFLAG		; ZF=1 (delim char match)
?50:	clc			; always return CF=0
	rts

; check for legal delimiter: '/' or '\'
;
;	entry:	A  = delimiter
;
;	exit:	A  = delimiter
;		ZF = 1, CF = 0 if legal delimiter '/' or '\'
;		ZF = 0, CF = 1 if not
;
;	use:	A
;
;------
chkslh:
;------
	cmp	#'/'
	beq	?10
	cmp	#'\'
	beq	?10
	sec			; error: ZF=0, CF=1
	rts
?10:	clc			; ok: ZF=1, CF=0
	rts
	
; sets up a clean fcb for later use
;
;	entry:	nothing
;
;	exit:	blank fcb
;		X = 0
;
;	use:	A,X
;
;------
clrfcb:
;------
	ldx	#10
	lda	#' '
?10:	sta	fcbs,x
	dex
	bpl	?10
	inx
	rts

; sets up a fake 'root' fcb for later use
;
;	entry:	nothing
;
;	exit:	fake '.' fcb
;		X = 0
;
;	use:	A,X
;
;-------
fakefcb:
;-------
	lda	#'.'		; start name with '.'
	sta	fcbs
	ldx	#10		; fill with blanks
	lda	#' '
?10:	sta	fcbs,x
	dex
	bne	?10
	rts

; move current working dir. list to build path list
;
;	entry:	ldtcdlp  = pointer to cur. dir. list
;		ldtcdcls = starting cluster of cur. dir.
;		ldtcdlvl = cur. dir. level
;
;	exit:	dircls = starting cluster of cur. dir.
;		bplvl  = build path level
;
;	use:	C,X,Y
;
;------
mvtobp:
;------
	ldx	curdrv
	ldy	!P0LDT+ldtp,x	; LDT pointer	
	ACC16
	ldx	!P0LDT+ldt_cdlvl,y
	stx	bplvl
	lda	!P0LDT+ldt_cdcls
	sta	dircls		; starting directory to build path
	beq	?10		; done: ii is the root dir.
	INDEX16
	txa			; count of list struct's
	asl	a		; mult. x 16 (size of HPL struct's)
	asl	a
	asl	a
	asl	a
	dec	a		; count of bytes to move - 1
	ldx	!P0LDT+ldt_cdlp,y
	;;ldx	ldtcdlp		; start source address
	ldy	#cbpl		; start dest address
	phb			; save DBR
	mvn	#WKB,#WKB	; move
	plb
?10:	CPU08
	rts

mvpath:
	ACC16
	lda	#cbpl		; offset of build path
	ldy	bplvl
	bcc	?10
	ldy	!P0LDT+ldtp,x	; LDT pointer
	lda	!P0LDT+ldt_cdlp,y
	ldx	!P0LDT+ldt_cdlvl,y
	txy
	clc
?10:	adc	#hpl_fcb	
	sta	tmpp		
	ACC08
	lda	#.SEG.cbpl
	sta	tmpp+2		; set HPL pointer
	sty	tmpx		; subdir level
	ldx	#0
	tya
	beq	?30		; is root
?20:	jsr	unparse
	dec	tmpx
	beq	?25
	dex
	lda	#'/'
	sta	!P0LDT+bpath,x
	inx
	ACC16
	lda	tmpp
	adc	#HPLSIZE
	sta	tmpp
	ACC08
	bra	?20
?25:	dex
?30:	stz	!P0LDT+bpath,x
	inx
	rts
	
; unparse fcb name into canonical asciiz string
unparse:
	ldy	#0		; unparse name
	lda	#8		; string limit
	jsr	?20		; unparse field
	lda	[tmpp],y
	cmp	#' '
	beq	?10		; no extension
	lda	#'.'
	sta	!P0LDT+bpath,x
	inx
	lda	#11		; string limit
	jsr	?20		; unparse extension
?10:	stz	!P0LDT+bpath,x
	rts
?20:	sta	fsiz
?25:	lda	[tmpp],y
	cmp	#' '
	beq	?30
	sta	!P0LDT+bpath,x
	inx
?30:	iny
	cpy	fsiz
	bcc	?25
	rts

; get drive number from path string
;
;	entry:	pathp	= long pointer to path string
;
;	exit:	A	= drive number
;		strix	= current path string index
;		CF	= 1 if illegal drive name in path string
;
;	use:	A,X
;
;-------
getpdrv:
;-------
	stz	strix		; clear current path string index
	lda	[pathp]		; check if string is empty...
	beq	?30		; ...which isn't O.K. at all
	tax
	ldy	#1
	lda	[pathp],y	; if the second char is ':'...
	beq	?10		; ...then drive is in pathname
	lda	defdrv		; take the default (selected) drive
	bra	?25		; return default drive
?10:	txa			; grab the drive designator
	cmp	#'a'		; make sure it is upper case
	bcc	?20
	cmp	#'z'+1
	bcs	?35		; invalid drive name
	and	#$DF
?20:	sbc	#'A'-1		; note: here always CF=0
	bcc	?30		; invalid drive number
	cmp	#MAXDRV
	bcs	?35		; invalid drive number
	iny			; bump path string pointer...
	sty	strix		; ...past ':'
?25:	clc
	rts
?30:	sec			; error flag
?35:	rts

; reject devices names -- never return to the caller if error
nodev:
	jsr	chkdev		; is this a device ?
	bcs	_eaccess	; yes
	rts
	
; reject if is directory or volume label
; never return to the caller if error
; A=fcb attribute
nodirvol:
	bit	#DA_DIR+DA_VOLUME
	bne	_eaccess
	rts

; reject "." and ".." names -- never return to the caller if error
nodot:
	lda	fcbs
	cmp	#'.'
	beq	chkx
_eaccess:
	lda	#ED_ACCESS	; access denied error
	jmp	oserror

; reject names with wild cards -- never return to the caller if error
nowild:
	lda	haswld
	beq	chkx
	lda	#ED_FILE	; invalid filename
	jmp	oserror

; check if fcb name = "."
; return ZF=1 if search for root (or '.' in root)
chkroot:
	lda	dircls		; are we in the root?
	ora	dircls+1
	bne	chkx		; no, no further checks required
	lda	fcbs
	cmp	#'.'		; is it a '.' entry ?
	bne	chkx
	ldx	#10
	lda	#' '		; check for 10 blanks
?10:	cmp	fcbs,x
	bne	chkx
	dex
	bne	?10
chkx:	rts

; check if fcb name is a character device name
; return:
;	CF=0 if not device name
;	CF=1 if is a device name, A=index of device name
; use: A,X,Y
chkdev:
	ACC16
	lda	#devlst		; set pointer to device names list
	sta	tmpp
	ACC08
	lda	#^devlst
	sta	tmpp+2
	stz	tmpx		; device index
?10:	ldx	#7		; loop match name
	txy
?20:	lda	fcbs,x		; compare fcb name...
	cmp	[tmpp],y	; ...against device name
	bne	?30		; unmatch
	dey
	dex
	bpl	?20
	sec			; match
	lda	tmpx		; A=index
	rts
?30:	lda	#8
	adc	tmpp		; next device name
	sta	tmpp
	bcc	?32
	inc	tmpp+1
?32:	inc	tmpx		; increment index
	lda	[tmpp]		; end of list?
	bne	?10		; no, so check next name
	clc			; unmatch
	rts

;---------------------------------------------------------------------------
; directory search
;---------------------------------------------------------------------------

fndfcbf:

; hashed directory search
; we try  to optimize  directory  searches using a dynamic hash table (HCB)
; 
;	entry:	dircls	= starting cluster of directory
;		curdrv	= current logical drive
;		fhash	= hash code of searched fcb entry (=0 if search empty)
;		dfcbtyp	= search type (1=empty entry, 2=fcbs entry)
;		fcbs	= local name to search
;
;	exit:	CF	= 1 if entry not found
;		CF	= 0 if entry was found
;		chdcls	= cluster of found entry
;		entofs	= entry offset number (0..15)
;		secofs	= sector's offset of found entry
;		fcbp	= long pointer to found entry
;		xbuf	= buffer that old sector of found entry
;		lba	= 24 bits address of sector
;		dfcbatt	= attribute of entry
;
;	use:	C,X,Y
;
;	note:	this routine should be called with 8 bit cpu mode
;
;-----
hfind:
;-----
	phb			; save current DBR
	lda	#^hcb_data
	pha
	plb			; set DBR to HCB struct's
	ldx	curdrv		; set drive
	stx	hcbdrv
	ACC16
	ldx	ldtrsiz		; root dir. size (in sec)
	lda	dircls		; starting directory cluster
	sta	chdcls		; current chainded dir. cluster
	beq	?02		; is root
	ldx	ldtcsiz		; subdir. size (in sec.)
?02:	stx	dsiz		; dir. size in sector's
	CPU08
	txa
	ldx	hlog2		; shift count
	beq	?06
	clc
	adc	hsiz		; round up to HCB size
	dec	a
?04:	lsr	a
	dex
	bne	?04
?06:	sta	maxseq		; how many sequnces
	ACC16
	lda	chdcls		; start from here
?10:	sta	hcbcls		; re-entry from this cluster
	sta	chdcls
	ldx	#0		; start with sequence=0
?15:	stx	hcbseq		; re-entry from this sequence
	ldy	#0		; current entry index
	CPU16
	sec			; we want an HCB even if it's recycled
	jsr	fndhcb		; find&fill an HCB (here return always CF=0) 
	ldx	hcbroot		; X=HCB ptr -- start search
	lda	!hcb_cnt,x	; we have this many entries hashed
	and	#$7FFF		; strip off bit 15
	asl	a
	adc	hcbroot
	sta	hlast		; point past to the last hash code
	lda	fhash		; look for this hash code
?20:	cmp	!hcb_data,x	; try to find a match
	beq	?30		; hash code match
?25:	iny			; re-entry after check hash code
	inx			; bump pointer
	inx
	cpx	hlast
	bcc	?20		; loop until last code
	ldx	hcbroot
	bit	!hcb_cnt,x	; test sign
	bmi	?40		; no more entries: exit with CF=1 (not found)
	INDEX08
	ldx	hcbseq		; will check more entries
	inx
	cpx	maxseq
	bcc	?15		; ok, try next sequence
	lda	hcbcls		; we are at root dir.?
	beq	?40		; yes, exit with CF=1 (not found)
	jsr	getncls		; C=next cluster in chain
	cmp	ldteoc		; have we fallen off the end of the chain?
	bcc	?10		; no, try next cluster
	bcs	?40		; exit with CF=1 (not found)
?30:	sty	hcby		; save current index
	stx	hcbx		; save current pointer
	tya			; hash matching index
	lsr	a		; div. by 16
	lsr	a
	lsr	a
	lsr	a
	tax			; sector's offset
	tya
	and	#$000F
	tay
	asl	a
	asl	a
	asl	a
	asl	a
	asl	a
	sta	fcbofs		; entry offset
	CPU08
	stx	secofs
	sty	entofs
	lda	hcbseq
	beq	?34		; offset=0
	ldy	hlog2		; shift count
	beq	?34		; no shift
?32:	asl	a
	dey
	bne	?32
?34:	clc
	adc	secofs
	sta	secofs
	tay			; sector's offset
	ldx	curdrv		; current logical drive
	ACC16
	lda	hcbcls		; current cluster
	sec			; preread buffer from disk
	jsr	fndbuf		; find & load buffer
	clc
	lda	fcbofs
	adc	fcbp
	sta	fcbp
	CPU08
	jsr	fcbcmp		; compare entry vs. local FCB
	bcc	?45		; match!
	CPU16
	ldx	hcbx		; restore pointer&index
	ldy	hcby
	lda	fhash		; hash code to find
	bra	?25		; restart search
?40:	CPU08
?45:	plb
	rts	

	.LONGA	off		; for assembler
	.LONGI	off

; compare directory entry with local FCB
fcbcmp:
	lda	dfcbtyp		; what kind of search?
	dec	a		; 1: wanted an empty entry?
	bne	?20		; no
	lda	[fcbp]
	beq	?30		; ok, virgin entry
	cmp	#$E5
	beq	?30		; ok, empty entry
?10:	sec			; return failure
	rts
?20:	ldx	#10		; match on all characters
	txy
?22:	lda	fcbs,x
	cmp	[fcbp],y
	bne	?10		; unmatch
	dey
	dex
	bpl	?22		; loop
	ldy	#11
	lda	[fcbp],y	; get file attribute
	sta	dfcbatt
?30:	clc			; return match
?35:	rts

; compare directory entry with local FCB
dfcbcmp:
	lda	dfcbtyp		; what kind of search?
	dec	a		; 1: wanted an empty entry?
	bne	?20		; no
	lda	[fcbp]
	beq	?30		; ok, virgin entry
	cmp	#$E5
	beq	?30		; ok, empty entry
?10:	sec			; return failure
	rts
?20:	dec	a		; 2: wanted match entry ('?' too)?
	bne	?40		; no
	ldx	#10		; match on all characters
	txy
?22:	lda	fcbs,x
	cmp	[fcbp],y
	beq	?24		; match
	cmp	#'?'		; this wild card match any character
	bne	?10		; failure: no match
?24:	dey
	dex
	bpl	?22		; loop
?26:	ldy	#11
	lda	[fcbp],y	; get file attribute
	sta	dfcbatt
?30:	clc			; return match
?35:	rts
?40:	ldx	#7		; caller wish a pattern match
	ldy	#0		; first time match name comnponent
	jsr	?60
	bcs	?50		; no match
	ldx	#8		; 2nd time match extension component
	ldy	#10
	jsr	?60
	bcc	?26		; ok, pattern match
?50:	rts			; no match
?60:	stx	xstart		; start of component
	sty	xend		; end of component
	lda	fcbs,x
	cmp	#'*'		; pattern start with star?
	beq	?70		; yes, so compare from end of string
	txy
	inc	xend		; one more for loop test
	bra	?64
?62:	lda	fcbs,x	
?64:	cmp	[fcbp],y	; compare
	beq	?66		; match
	cmp	#'?'		; wildcard match any character
	beq	?66
	cmp	#'*'		; star match any pattern
	beq	?68
	bne	?10		; no match
?66:	inx
	iny
	cpx	xend		; loop until end of string
	bcc	?62
?68:	clc			; match
	rts
?70:	ldy	xend		; compare from end of string
	inc	xstart		; skip beginning star
?72:	lda	[fcbp],y
	cmp	#' '		; skip padding blanks
	bne	?74
	dey
	cpy	xstart
	bcs	?72
	rts			; here was just one star
?74:	ldx	xend
?76:	lda	fcbs,x
	cmp	#' '		; skip padding blanks
	bne	?78
	dex
	cpx	xstart
	bcs	?76
	rts			; here was just one star
?78:	lda	fcbs,x
	cmp	#'?'
	beq	?80
	cmp	[fcbp],y
	bne	?80
	dey
	dex
	cpx	xstart
	bcs	?78
	rts			; match
?80:	sec
	rts


	.LONGA	off		; for assembler
	.LONGI	off
	
; compute hash code of local FCB name (fcbs)
; 
;	entry:	fcbs	= local name to hash
;		fhash	= hash code of searched fcb entry (=0 if search empty)
;
;	exit:	fhash	= hash code of local fcb name (=0 if empty)
;
;	use:	C,X,Y
;
;	note:	this routine should be called with 8 bit cpu mode
;
;-----
mkhsh:
;-----
	stz	fhash
	stz	fhash+1
	lda	fcbs		; first char of file name
	beq	?20		; empty entry
	cmp	#$E5
	beq	?20		; deleted entry
	and	#$7F
	sta	fhash+1
	ldy	#$0A		; hash all char's
	ldx	#0
?10:	ACC16
	lda	fhash
	asl	a
	adc	#$0000
	sta	fhash
	ACC08
	inx
	lda	fcbs,x
	and	#$7F
	eor	fhash
	sta	fhash
	dey
	bne	?10
	ACC16
	lda	fhash
	bne	?15		; hash=0 reserved for empty entry
	inc	a
	sta	fhash
?15:	ACC08
?20:	rts

;---------------------------------------------------------------------------
; utilities
;---------------------------------------------------------------------------

; in: dcnt, dircls, chdcls, curdrv
getdfcb:
getdbuf:
	CPU16
	lda	dircls		; first cluster of directory
	ldx	dcnt		
	inx
	stx	dcnt		; current directory index
	beq	?18		; first entry of dir. 	
	cpx	#$FFFF		; invalid next dir. index ?
	beq	?16		; yes, set end of directory condition (CF=1)
	txa			; C=dcnt
	lsr	a		; div. by 16 (16 entries per sector)
	lsr	a
	lsr	a
	lsr	a
	sta	wtmp		; save sector's count
	tay			; Y=sector's offset
	txa			; C=dcnt
	and	#$000F		; entry number
	tax			; X=entry offset
	INDEX08
	sty	secofs		; low sector's offset (valid for root dir only)
	stx	entofs		; entry offset
	CPU16
	lda	dircls		; we are at root dir.?
	bne	?05		; no
	ldx	dcnt
	cpx	ldtrent		; end of the root directory?
	bcs	?15		; yes, set end of directory condition (CF=1)
	bra	?20		; get buffer
?05:	INDEX08
	lda	wtmp		; sector's offset
	tay
	ldx	ldtcshf		; sector per cluster shift count
	beq	?08		; no shift (1 sec. per cluster)
?06:	lsr	a
	dex
	bne	?06
?08:	sta	clsofs		; cluster's count to skip from the first
	ACC08
	tya
	and	ldtmcls		; sector's offset within cluster
	sta	secofs
	CPU16
	lda	chdcls		; do we already know where we are?
	beq	?10		; if not trace from start of chain
	ldy	entofs		; entofs+secofs: null if we moved onto...
				; ...next cluster
	bne	?24		; we stay at the same cluster
	iny			; move on the next cluster in the chain...
	sty	clsofs		; ...starting from chdcls
	bra	?12
?10:	lda	dircls		; start cluster in the chain
?12:	ldx	clsofs		; skip along chain until we arrive...
	beq	?20		; ...at the destination cluster
	dec	clsofs		; decrement cluster offset
	jsr	getncls		; C=next cluster in chain
	cmp	ldteoc		; have we fallen off the end of the chain?
	bcc	?12		; no, loop again
				; here, end of cluster's chain
?15:	lda	#$FFFF		; set end of directory condition
	sta	dcnt
?16:	stz	chdcls		; forget where we are
	bra	?30		; exit with CF=1
?18:	stz	clsofs		; we are at first entry: clear offset's
	stz	entofs		; entofs+secofs
?20:	sta	chdcls		; remember this cluster for next time
?24:	CPU08			; C=cluster (=0 if root)
	ldx	curdrv		; current logical drive
	ldy	secofs		; sector's offset
	sec			; preread buffer from disk
	jsr	fndbuf		; find & load buffer 
?30:	CPU08
	rts

; get next cluster number
; in:	C = cluster
; out:	C = next chain cluster
; ldtfp preset to floppy fat table
getncls:
	php
	CPU08
	bit	ldtfg1		; bit 6=0 if floppy
	bvc	?10		; FAT12 media
	xba			; A=sector (one sec. = 256 cluster #)
	tax
	lsr	a		; divide by 16 to get x-mem buffer bank
	lsr	a
	lsr	a
	lsr	a
	clc
	adc	ldtfbuf		; add fat table base base bank
	sta	>PIA0+PIAPRA	; set bank
	txa
	xba			; restore cluster number in C
	CPU16
	and	#$0FFF		; mask 4096 cluster's
	asl	a
	tax			; index to 8k fat table in x-mem window
	lda	>XMWIN,x	; get chained cluster
	plp
	rts
?10:	CPU16
	pha			; cluster #
	lsr	a		; div. by 2
	clc
	adc	$01,s		; mult. x 1.5
	tay			; index to fat table
	pla
	lsr	a		; even or odd cluster?
	lda	[ldtfp],y
	bcs	?15		; odd cluster
	and	#$0FFF		; even cluster
	bra	?20
?15:	lsr	a		; odd cluster: 4 bit's shift		
	lsr	a
	lsr	a
	lsr	a
?20:	plp
	rts

	.LONGA	off		; for assembler
	.LONGI	off
	
; get pointer to directory entry given by dcnt
; in:	dcnt=directory entry counter 
getsec:
	ACC16
	lda	dcnt		; compute sector count
	lsr	a		; divide by 16 (16 entries per sector)
	lsr	a
	lsr	a
	lsr	a
	pha			; sector count
	ldx	ldtcshf		; sector per cluster shift count
	beq	?06		; no shift (1 sec. per cluster)
?04:	lsr	a
	dex
	bne	?04
?06:	sta	clsofs		; cluster's count to skip from the first
	pla
	ACC08
	and	ldtmcls		; cluster sector offset
	sta	secofs
	lda	dcnt
	and	#$0F
	asl	a
	sta	entofs
	rts
	
;---------------------------------------------------------------------------
; tables
;---------------------------------------------------------------------------

dlmstr	.DB	':.;,=+\<>|/"[]'	; path delimeters
DLMSTR	.EQU	$-dlmstr

devlst:
	.DB	'CON     '	; Keyboard and display
	.DB	'KEYBD$  '	; Keyboard
	.DB	'SCREEN$ '	; Display
	.DB	'CONSOLE$'
	.DB	'PRN     '	; System list device, usually a parallel port
	.DB	'LST     '	; System list device, usually a parallel port
	.DB	'LPT1    '	; First parallel printer port
	.DB	'LPT2    '	; Second parallel printer port
	.DB	'LPT3    '	; Third parallel printer port
	.DB	'AUX     '	; Auxiliary device, usually a serial port
	.DB	'COM1    '	; First serial communications port
	.DB	'COM2    '	; Second serial communications port
	.DB	'COM3    '	; Third serial communications port
	.DB	'COM4    '	; Fourth serial communications port
	.DB	'NUL     '	; Bit-bucket device
	.DB	'$IDLE$  '	; Bit-bucket device
	.DB	'CLOCK$  '	; System real-time clock
	.DB	'CLOCK   '	; System real-time clock
	.DB	'CONFIG$ '
	.DB	0
