;;
;; Copyright (c) 2016 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; name: fsint.asm	
;; rev.: 2016/02/26
;; o.s. 65C816 version v1.0

;; file system initialitation sub-system

; include RTC definition
_RTC_INC_	.SET	1

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\dirp01.inc
	.INCLUDE inc\tossys.inc
	.INCLUDE src\F9\fstmp.inc

.LIST on

; default hidden sectors
HIDSEC		.EQU	$3F
; default hashed sector's
DEFHSIZ		.EQU	1
DEFHLOG2	.EQU	0

; max. clusters count + 1
MAXF16		.EQU	$FFF5
MAXF12		.EQU	$0FF5
; End-of-Chain marker (EOC marker)
F16EOC		.EQU	$FFF8
F12EOC		.EQU	$0FF8
	
;---------------------------------------------------------------------------
; code segment -- bank $F9
;---------------------------------------------------------------------------

	.CODEF9
	.PUBLIC fssetup, gctime16, bldldt, lhdfmt, shpart, lfdisk	
	.EXTERN cbmfs_cmd

	.LONGA	off
	.LONGI	off

;---------------------------------------------------------------------------
; setup 
;---------------------------------------------------------------------------

fssetup:
	phd
	pea	#P0LDT
	pld			; set dp page
	ldx	#RTCSHID	; default hidden sector's (for fdisk)
	stx	RTCALE		; from RTC RAM in bank 0 & 1
	lda	RTCDATA		; read saved value
	cmp	#$40		; invalid?
	bcc	?01b		; no
?01:	lda	#HIDSEC		; get default value
	sta	RTCDATA		; save default value
?01b:	sta	defhsec
	cmp	#0
	bne	?01c
	lda	#HIDSEC		; get default value
	sta	RTCDATA		; save default value
	sta	defhsec
?01c:	ldx	#RTCHUSE	; flag use hash block's
	stx	RTCALE
	lda	RTCDATA		; read saved value
	cmp	#$FF
	bne	?02
	lda	#$80		; default value (use hash)
	sta	RTCDATA		; save default value
?02:	sta	!P0FAT+hcbuse
	ldx	#RTCHLOG2	; sector's per hcb block
	stx	RTCALE
	lda	RTCDATA		; read saved value
	cmp	#$FF
	beq	?03
	cmp	#7		; valid value: 0..6
	bcc	?04
?03:	lda	#$04		; default value (16 sector's per block)
	sta	RTCDATA		; save default value
?04:	sta	!P0FAT+hlog2
	tax
	lda	>CLSSIZE,x	; get block size in sector's
	sta	!P0FAT+hsiz
	ACC16
	lda	#16		; get max. entries per block
	txy
	beq	?04b		; no shift
?04a:	asl	a		; mult * 16 to get entries #
	dex
	bne	?04a
?04b:	asl	a		; 2 bytes per entry
	adc	#HCBSIZE	; plus HCB header
	sta	hcbsiz		; effective HCB size
	sec
	lda	#0
	sbc	#hcbstart
	sta	hcbrem		; free room in bank
	ldy	#0
	CPU16
	ldx	#TOS0SIZ-2	; clear TOS work area
	tya
?05:	sta	>_TOS0_START,x
	dex
	dex
	bpl	?05
	tyx
?10:	sta	>hcb_bnk,x	; clear HCB bank
	inx
	inx
	bne	?10
	CPU08
	phb			; save dbr
	lda	#WKB
	pha
	plb			; set dbr to fat work area
	CPU16
	lda	#BCBXBNK	; init BCB struct's for at drive's buffer's
	sta	bcbbnk
	lda	#bcbstart	
	sta	>P0FAT+bcbroot
	tax	
?15:	lda	bcbbnk
	sta	!bcb_buf,x
	asl	a
	asl	a
	asl	a
	asl	a		; x-mem bank
	tay
	and	#$FF00		; low byte = buffer hi ptr
	inc	a		; bank 1
	sta	!bcb_xbnk,x	; store buffer high ptr + x-mem bank
	tya
	xba			; C=0000,1000,2000,...,E000,F000
	and	#$F000
	lsr	a
	lsr	a
	lsr	a		; C=0000,0200,0400,...,1C00,1E00
	ora	#XMSTART	; C=2000,2200,2400,...,3C00,3E00
	sta	!bcb_ptr,x
	lda	#$FFFF
	sta	!bcb_drv,x	; drive+sec = $FFFF -> unused
	inc	bcbbnk
	txa
	clc
	adc	#BCBSIZE
	cmp	#bcbend
	bcs	?20
	sta	!bcb_next,x
	tax
	bra	?15
?20:	stz	!bcb_next,x	; last struct
	txa
	sta	>P0FAT+bcblst	
	CPU08
	lda	#BCBDMA		; init BCB's struct's in dma buffer's
	sta	bcbbnk+1	; dma buffer # in high byte
	jsr	?100		; get dma bank
	sty	bcbbnk
	CPU16
	lda	#dmastart	
	sta	>P0FAT+dmaroot
	tax
?25:	lda	#$FFFF		; invalidate buffer
	sta	!bcb_drv,x	; drive+sec = $FFFF -> unused
	lda	bcbbnk
	sta	!bcb_dma,x	; store buffer high ptr + dma buffer #
	tay
	and	#$7F00		; strip off bit 15
	asl	a
	sta	!bcb_ptr,x	; buffer pointer
	tya
	xba
	inc	a		; increment bank
	xba
	sta	bcbbnk
	txa
	clc
	adc	#BCBSIZE
	cmp	#dmaend
	bcs	?30
	sta	!bcb_next,x	; pointer to next BCB
	tax
	bra	?25
?30:	stz	!bcb_next,x	; last struct
	txa
	sta	>P0FAT+dmalst	
	ldx	#dma0s
	lda	#$FFFF		; invalidate buffer
	sta	!bcb_drv,x	; drive+sec = $FFFF -> unused
	ldx	#dma1s
	sta	!bcb_drv,x	; drive+sec = $FFFF -> unused	
	CPU08
	lda	#HCBBNK		; init HCB struct's
	pha
	plb
	CPU16
	lda	#hcbstart
	sta	>P0FAT+hcbroot
	tax
?35:	lda	#$FFFF
	sta	!hcb_seq,x	; unused HCB
	lda	hcbrem		; free room
	sec
	sbc	hcbsiz
	bcc	?40		; no more room
	cmp	hcbsiz
	bcc	?40		; no more room
	sta	hcbrem		; save new free room
	txa
	clc
	adc	hcbsiz
	sta	!hcb_next,x
	tax			; next HCB
	bra	?35
?40:	stz	!hcb_next,x	; last HCB
	txa
	sta	>P0FAT+hcblst
	CPU08
	plb			; restore dbr	
	ldx	#ldt0		; set up LDT ptr's
	stx	ldtp		; logical drive #0
	lda	!fdcctl		; phisycal drive 0 ready flag
	and	#$80
	sta	ldt_fg1,x
	ldx	#ldt1
	stx	ldtp+1		; logical drive #1
	lda	#$81		; virtual drive 1 always ready
	sta	ldt_fg1,x
	ldx	#ldt2
	stx	ldtp+2		; logical drive #2
	lda	!atadev		; phisycal ata device #0
	and	#$80
	ora	#$40		; ata device #0
	sta	ldt_fg1,x
	ldx	#ldt3
	stx	ldtp+3		; logical drive #3
	lda	!atadev+1	; phisycal ata device #1
	and	#$80
	ora	#$41		; ata device #1
	sta	ldt_fg1,x
	lda	#DR0SYS
	sta	sysbuf		; drive 0 system track cache
	tax
	inx
	stx	ldt0+ldt_fbuf
	jsr	?100
	sta	ldt0+ldt_pbr+1	; set PBR pointer
	sty	ldt0+ldt_pbr+2
	stz	ldt0+ldt_pbr
	inc	a		; drive 0 fat table dma buffer
	inc	a
	sta	ldt0+ldt_fp+1	; set fat pointer
	sty	ldt0+ldt_fp+2
	stz	ldt0+ldt_fp
	lda	#DR1SYS
	sta	sysbuf+1	; drive 1 system track cache
	tax
	inx
	stx	ldt1+ldt_fbuf
	jsr	?100
	sta	ldt1+ldt_pbr+1	; set PBR pointer
	sty	ldt1+ldt_pbr+2
	stz	ldt1+ldt_pbr
	inc	a		; drive 1 fat table dma buffer	
	inc	a
	sta	ldt1+ldt_fp+1	; set fat pointer
	sty	ldt1+ldt_fp+2
	stz	ldt1+ldt_fp
	lda	#FAT0X
	sta	ldt2+ldt_fbuf	; fat cache buffer drive #2
	lda	#FAT1X
	sta	ldt3+ldt_fbuf	; fat cache buffer drive #3
	lda	#>XMSTART
	ldy	#XMBANK
	stz	ldt2+ldt_fp	; set fat drive # 2 pointer
	sta	ldt2+ldt_fp+1
	sty	ldt2+ldt_fp+2
	stz	ldt3+ldt_fp	; set fat drive # 3 pointer
	sta	ldt3+ldt_fp+1
	sty	ldt3+ldt_fp+2
	lda	>PBRPTR2
	sta	ldt2+ldt_pbr+2
	lda	>PBRPTR2+1
	sta	ldt3+ldt_pbr+2
	lda	#$FF
	sta	ldt0+ldt_fmt
	sta	ldt1+ldt_fmt
	sta	ldt2+ldt_fmt
	sta	ldt3+ldt_fmt	
	ACC16
	lda	>PBRPTR
	sta	ldt2+ldt_pbr
	lda	>PBRPTR+2
	sta	ldt3+ldt_pbr
	lda	#F12EOC
	sta	ldt0+ldt_eoc	; FAT12 end of chain marker
	sta	ldt1+ldt_eoc
	lda	#F16EOC
	sta	ldt2+ldt_eoc	; FAT16 end of chain marker
	sta	ldt3+ldt_eoc
	lda	#cwdl0
	sta	ldt0+ldt_cdlp	; current directory pointer drive #0
	lda	#cwdl1
	sta	ldt1+ldt_cdlp	; current directory pointer drive #1
	lda	#cwdl2
	sta	ldt2+ldt_cdlp	; current directory pointer drive #2
	lda	#cwdl3
	sta	ldt3+ldt_cdlp	; current directory pointer drive #3
	stz	ldt0+ldt_cdcls	; current directory cluster
	stz	ldt1+ldt_cdcls	
	stz	ldt2+ldt_cdcls	
	stz	ldt3+ldt_cdcls	
	ldy	#0
	sty	ldt0+ldt_cdlvl	; current directory level
	sty	ldt1+ldt_cdlvl
	sty	ldt2+ldt_cdlvl
	sty	ldt3+ldt_cdlvl
	stz	ldt0+ldt_free	; clear free cluster's count & next free clu.
	stz	ldt0+ldt_nxt
	stz	ldt1+ldt_free
	stz	ldt1+ldt_nxt
	stz	ldt2+ldt_free
	stz	ldt2+ldt_nxt
	stz	ldt3+ldt_free
	stz	ldt3+ldt_nxt
	ACC08	
	ldx	#1		; test virtual drive (drive #1)
	jsr	ldtfd
	bcc	?90		; ok
	lda	#$82		; force HD format
	sta	!vdrive
	lda	#2
	ldy	#0
	ldx	#1
	jsr	newf12		; make fat12 file system (no error!)
	ldx	#1
	jsr	ldtfd		; test again virtual drive
?90:	ldx	#2		; build LDT for ata device's
?95:	jsr	ldt16
	inx
	cpx	#MAXDRV
	bcc	?95
	pld
	rts

?100:	ldy	#DMABNK		; bank for buffer #00..#7F
	asl	a		; get high pointer
	bcc	?102
	iny			; next bank
?102:	rts

; X=logical, DBR=0, DPR=P0LDT
ldt16:
	stx	ldev		; logical drive
	txy
	ldx	ldtp,y
	stx	ldtcur		; current LDT ptr
	bit	ldt_fg1,x
	bpl	?15		; not ready
	bvc	?15		; floppy disk
	lda	ldt_fg1,x
	and	#1
	sta	pdev		; phisycal drive
	clc
	adc	#'C'		; ata device: 'C' or 'D'
	sta	lunit		; logical drive letter
	jsr	?20		; check message
	jsr	?30		; load PBR
	jsr	?40		; check PBR
	jsr	?50		; fat calc
	ldx	ldtcur
	ldy	ldt_fbuf,x
	sty	xbnk
	ACC16
	tya
	asl	a
	asl	a
	asl	a
	asl	a
	sta	xfatb
	lda	fat1
	sta	flba
	lda	ldt_fp,x
	sta	tp
	ldy	ldt_fp+2,x
	sty	tp+2
	ACC08
	lda	#0		; try to load fat1 table
	jsr	?60
	bcc	?10		; ok
	ACC16
	lda	fat2
	sta	flba
	ACC08
	lda	#$80		; try to read fat2 table (no return if error)
	jsr	?60
?10:	jsr	?80		; count free cluster's
	jsr	mvldt		; move data to LDT
	lda	#$E0
	sta	ldt_fg2,x
	jsr	?100		; print summary data
?15:	ldx	ldev	
	rts

?20:	lda	lunit		; print check message
	pha
	phk
	pea	#!?25
	lda	#5
	pha
	BPRINTF
	ldx	pdev	
	lda	!atadev,x	; check ata device flag
	and	#10111101B	; ready&ident&valid MBR partition
	cmp	#10111101B
	beq	?36		; ok
	SCNPRINT
	.DB	'unpartitioned disk or invalid partition ID.', 13, 0
	bra	?32		; no or invalid partition

?25:	.DB	'check <%c>: ', 0

?30:	ldx	pdev		; load PBR
	lda	#3
	ATACTL
	bcc	?36		; ok
	SCNPRINT
	.DB	'unable to load partition boot record (PBR).', 13, 0
?32:	pla			; discard return address
	pla
?34:	ldx	ldev
?36:	rts
	
?40:	jsr	chk16		; check PBR
	bcc	?36
?45:	SCNPRINT
	.DB	'invalid partition type (not FAT16).', 13, 0
	bra	?32

?50:	jsr	fcalc		; fat calc
	bcc	?36
	bra	?45

?60:	sta	lflag
	ldx	pdev
	phx			; phisical drive
	lda	#0
	pha			; lba high
	pei	(flba)		; lba low
	ldx	fatsiz
	phx			; sector's count
	pha			; use x-mem buffer
	pei	(xfatb)		; x-mem buffer
	pha			; flag=0
	ATAREAD
	bcc	?36		; ok no error
	bit	lflag
	bpl	?36		; try fat2 table
	SCNPRINT
	.DB	'fat table load error.', 13, 0
	bra	?32

?80:	ldx	CRXME		; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	lda	xbnk
	sta	!PIA0+PIAPRA	; set x-mem bank
	phb			; save DBR
	lda	tp+2
	pha
	plb			; set DBR to fat table cache
	ldx	#0		; cluster number
	CPU16
	txy
	stz	freecnt
	stz	freecls
?85:	lda	(tp),y		; scan fat table
	bne	?90		; not free
	inc	freecnt		; update free cluster's count
	lda	freecls		; already set?
	bne	?90		; yes
	stx	freecls		; set first free cluster
?90:	inx
	cpx	maxcls		; end of table?
	bcs	?95		; yes
	iny
	iny
	cpy	#XMWSIZ		; end of x-mem window?
	bcc	?85		; no
	ACC08
	lda	>PIA0+PIAPRA	; update x-mem bank
	inc	a
	sta	>PIA0+PIAPRA	
	ACC16
	ldy	#0
	bra	?85
?95:	CPU08
	plb			; restore DBR
	plx			; restore X mem setting
	sta	!CRXME,x
	rts

?100:	ACC16			; print summary	
	lda	maxcls
	dec	a
	dec	a
	sta	maxcls
	ACC08
	lda	clssiz
	pha
	pei	(freecnt)
	pei	(maxcls)
	phk
	pea	#!?110
	lda	#9
	pha
	BPRINTF
	rts

?110:	.DB	'%,hu cluster(s) (free: %,hu, 1 cluster=%bu sector(s)).', 13, 0

; build LDT for floppy disk
; X=logical drive
bldldt:
	lda	#0
	phb
	pha
	plb
	phd
	pea	#P0LDT
	pld
	jsr	ldtfd
	pld
	plb
	rts

testldt:
	ldx	#0
	jsr	bldldt
	brk
	brk

; build LDT for floppy disk media
; X=logical device, DPR=P0LDT
ldtfd:
	stx	ldev		; logical drive
	txy
	ldx	ldtp,y
	stx	ldtcur		; current LDT ptr
	bit	ldt_fg1,x
	bpl	?15		; not ready
	bvs	?15		; ata device
	stz	ldt_fg2,x	; reset flag 2
	lda	ldt_fg1,x
	and	#1
	sta	pdev		; phisycal drive
	jsr	?20		; check media format
	jsr	?30		; load PBR
	jsr	?40		; check PBR
	jsr	?50		; fat calc
	jsr	?60		; count free cluster's
	lda	fdfmt
	cmp	#1		; DD ibm disk?
	bne	?10		; no
	jsr	?90		; pre-read cluster #4
?10:	jsr	mvldt		; move data to LDT
	lda	#$E0		; update flag 2
	ora	fdfmt
	sta	ldt_fg2,x
	stz	ldt_cdlvl,x	; set to root directory
	jsr	?100		; setup BCB list for root, cluster #2,#3&#4
?15:	ldx	ldev
	clc
	rts

?20:	tax			; try to get media format
	FDCGETFMT
	bcs	?34		; unformatted disk?
	sty	fdfmt		; save media format
	cpy	#1		; cbm format?
	bne	?35		; no, so return
	lda	#'I'		; send init command to cbm emulator
	sta	bpath
	lda	pdev
	ora	#'0'		; drive
	sta	bpath+1
	stz	bpath+2
	lda	#bpath		; address of command string
	ldx	#>P0LDT
	ldy	#0
	jsr	cbmfs_cmd
	bcs	?34		; skip return address if error
	lda	#10100001B	; set flag 2
	ldx	ldtcur
	sta	ldt_fg2,x
	bra	?34		; skip return address & CF=0

?30:	ldx	pdev		; read system track (track 0)
	phx			; phsycal drive
	lda	#0
	pha			; track=0
	lda	sysbuf,x	; dma buffer #
	pha
	lda	fdfmt		; format
	pha
	FDCTKRD			; read whole track
	bcc	?35		; ok
?34:	pla			; discard return address
	pla
?35:	rts

?40:	jsr	chk12		; check floppy disk PBR
	bcc	?35		; ok
	bra	?34		; skip return address

?50:	jsr	fcalc		; fat calc
	bcc	?35
	bra	?34		; skip return address

?60:	ldx	ldtcur		; compute free cluster's
	ACC16
	lda	ldt_fp,x
	sta	tp
	ldy	ldt_fp+2,x
	phb
	phy
	plb
	ldx	#0		; cluster number
	CPU16
	txy			; Y=fat table index
	stz	freecnt
	stz	freecls
?65:	lda	(tp),y		; even index cluster
	and	#$0FFF
	bne	?70		; not free
	inc	freecnt		; update free cluster's count
	lda	freecls		; already set?
	bne	?70		; yes
	stx	freecls		; set first free cluster
?70:	inx
	cpx	maxcls		; end of table?
	bcs	?80		; yes
	iny			; odd index cluster
	lda	(tp),y
	lsr	a		; shift x 4
	lsr	a
	lsr	a
	lsr	a
	bne	?75
	inc	freecnt		; update free cluster's count
	lda	freecls		; already set?
	bne	?75		; yes
	stx	freecls		; set first free cluster
?75:	iny
	iny			; pointer to next cluster
	inx
	cpx	maxcls
	bcc	?65
?80:	CPU08
	plb
	rts

?90:	ldx	pdev		; pre-read cluster #4 DD disk
	phx			; phsycal drive
	pea	#$0100		; track=1, head=0
	pea	#$0102		; start sec=1, sec num=2
	lda	sysbuf,x	; dma buffer #
	clc			; using first buffer after...
	adc	#20		; ...system track 0
	pha
	lda	fdfmt		; format
	pha
	FDCTKRD			; read whole track
	bcc	?95		; ok
	pla			; discard return address
	pla
?95:	rts

?100:	lda	ldev		; setup BCB list for root
	sta	tbdrv
	stz	tbcls+1
	stz	tbhead+1
	stz	tblba+1
	ldx	pdev
	lda	#$07		; first DD root sec
	ldy	fdfmt
	beq	?105
	lda	#$13		; first HD root sec
?105:	clc
	adc	sysbuf,x
	tax
	ldy	#DMABNK		; bank for buffer #00..#7F
	asl	a		; get high pointer
	bcc	?110
	iny			; next bank
?110:	sty	tbdma
	stz	tbptr
	dec	a
	dec	a
	sta	tbptr+1
	dex
	stx	tbdma+1		; dma buffer #
	ldx	#0	
	lda	fdfmt
	beq	?115
	ldx	#?501
?115:	phb
	lda	#WKB
	pha
	plb
	INDEX16
	ldy	#dma0s
	lda	pdev
	beq	?120
	ldy	#dma1s
?120:	lda	>?500,x		; lba
	beq	?125		; end of list
	sta	tblba
	lda	>?510,x		; sec. offset
	sta	tbdrv+1
	lda	>?550,x		; cluster #
	sta	tbcls
	lda	>?530,x		; head #
	sta	tbhead
	lda	>?520,x		; track #
	sta	tbtrk
	lda	>?540,x		; sector #
	sta	tbtrk+1
	lda	tbptr+1		; update buffer ptr
	inc	a
	inc	a
	sta	tbptr+1
	inc	tbdma+1		; update dma buffer #
	ACC16CLC
	lda	tbdrv
	sta	bcb_drv,y
	lda	tbcls
	sta	bcb_cls,y
	lda	tblba
	sta	bcb_lba,y
	lda	tbhead
	sta	bcb_head,y
	lda	tbtrk
	sta	bcb_trk,y
	lda	tbptr
	sta	bcb_ptr,y
	lda	tbdma
	sta	bcb_dma,y
	tya
	adc	#BCBSIZE
	sta	bcb_next,y
	tay
	ACC08
	inx
	bra	?120
?125:	ACC16
	tya
	sec
	sbc	#BCBSIZE
	tax
	stz	!bcb_next,x
	CPU08
	plb
	rts

	; DD system track lba
?500:	.DB	$07, $08, $09, $0A, $0B, $0C, $0D, $0E, $0F, $10, $11, $12
	.DB	$13, $00
?501	.EQU	$-?500
	; HD system track lba
	.DB	$13, $14, $15, $16, $17, $18, $19, $1A, $1B, $1C, $1D, $1E
	.DB	$1F, $20, $21, $22, $23, $00

?510:	; DD sector offset
	.DB	0, 1, 2, 3, 4, 5, 6, 0, 1, 0, 1, 0, 1, 0
	; HD sector offset
	.DB	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 0, 0, 0, 0
	
?520:	; DD track
	.DB	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0
	; HD track
	.DB	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	
?530:	; DD head
	.DB	0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0
	; HD head
	.DB	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0
	
?540:	; DD sector #
	.DB	8, 9, 1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 0
	; HD sector #
	.DB	2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
	.DB	16, 17, 18, 0

?550:	; DD cluster #
	.DB	0, 0, 0, 0, 0, 0, 0, 2, 2, 3, 3, 4, 4, 0
	; HD cluster #
	.DB	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	.DB	2, 3, 4, 0

; check fat16 volume in pbr record
chk16:
	lda	#$FF
	sta	fdfmt
	lda	pdev		; phisycal drive
	asl	a
	asl	a
	tax
	ACC16
	lda	!atambr,x	; partition start
	sta	hidsec		; this is also hidden sectors count
	lda	!atambr+2,x
	and	#$00FF
	sta	hidsec+2	
	lda	!ataprt,x	; total partition sector's
	sta	totsec
	lda	!ataprt+2,x
	sta	totsec+2
	ldx	ldtcur		; LDT ptr
	lda	ldt_pbr,x
	sta	pbrp
	ldy	ldt_pbr+2,x
	sty	pbrp+2
	ldx	pbrp+1
	inx
	stx	pbrp+1
	lda	#MAXF16
	sta	maxcnt
	ldy	#$FE		; pointer to pbr signature
	lda	[pbrp],y
	cmp	#$AA55
	bne	?10		; signature error
	dex
	stx	pbrp+1
	ldy	#$0B		; pointer to bytes per sector
	lda	[pbrp],y
	cmp	#$0200		; must be always 512 bytes per sector
	bne	?10		; error
	ldy	#$1C		; pointer to hidden sector's count
	lda	[pbrp],y	; typical value: $3F
	cmp	hidsec		; must match
	bne	?10
	iny
	iny
	lda	[pbrp],y
	cmp	hidsec+2
	bne	?10
	cmp	#0
	bne	?10		; must be 0
	ldy	#$0E		; pointer to reserved sector's
	lda	[pbrp],y	; typical value: 1 (no mandatory)
	sta	rsvsec
	ldy	#$11		; pointer to root dir entry #
	lda	[pbrp],y
	beq	?10		; fat32?
	cmp	#$0200		; typical value: not mandatory...
	bne	?10		; ...but this will be
	sta	rootent
	lsr	a		; divide by 16...
	lsr	a
	lsr	a
	lsr	a		;...to get root dir size in sec's
	sta	rootsiz
	ldy	#$13		; pointer to tot. sec16
	lda	[pbrp],y
	beq	?06		; invalid: is valid field tot. sec32
	cmp	totsec
	bne	?10		; error
	lda	totsec+2	; must be 0
	bne	?10		; errore
	beq	?08		; ok
?06:	ldy	#$20		; pointer to tot. sec32
	lda	[pbrp],y
	cmp	totsec
	bne	?10		; error
	iny
	iny
	lda	[pbrp],y
	cmp	totsec+2
	bne	?10		; error
?08:	ldy	#$16		; pointer to fat size (FAT16)
	lda	[pbrp],y
	sta	fatsiz
	beq	?10		; hmmm... is fat32... not good
	cmp	#257		; fat16 table size limited to 256 sec. max
	bra	?12
?10:	sec
?12:	CPU08
	bcs	?18		; error
	ldy	#$15		; pointer to media byte
	lda	[pbrp],y	; must be $F8
	cmp	#$F8
	bne	?16
	ldy	#$10		; pointer to count of fat table's
	lda	[pbrp],y
	cmp	#2		; must be 2
	bne	?16
	ldy	#$0D		; pointer to sector(s) per cluster
	lda	[pbrp],y
	ldx	#CLSCNT-1
?14:	cmp	>CLSSIZE,x
	beq	?20
	dex
	bpl	?14
?16:	sec
?18:	rts
?20:	sta	clssiz
	dec	a
	sta	clsmod
	lda	>CLSSHIFT,x
	sta	clsshf
	clc
	rts

; check floppy disk PBR
chk12:
	;ldx	pdev		; phisycal drive
	;lda	!fdcdrv,x	; get format from bios
	;and	#00000011B
	;sta	fdfmt		; format to check (0 or 2)
	lda	fdfmt
	lsr	a		; index DD/HD
	sta	fdix
	asl	a
	sta	fd2ix
	ACC16
	lda	#MAXF12
	sta	maxcnt
	ldx	ldtcur		; LDT ptr
	lda	ldt_pbr,x
	sta	pbrp
	ldy	ldt_pbr+2,x
	sty	pbrp+2
	ldx	pbrp+1
	inx
	stx	pbrp+1
	ldy	#$FE		; pointer to pbr signature
	lda	[pbrp],y
	cmp	#$AA55
	bne	?10		; signature error
	dex
	stx	pbrp+1
	ldx	fd2ix		; X=DD/HD double index
	ldy	#$0B		; pointer to bytes per sector
	lda	[pbrp],y
	cmp	#$0200		; must be always 512 bytes per sector
	bne	?10		; error
	ldy	#$0E		; pointer to reserved sector's
	lda	[pbrp],y	; mandatory value: 1
	sta	rsvsec
	dec	a
	bne	?10		; error
	ldy	#$11		; pointer to root dir entry #
	lda	[pbrp],y
	cmp	>FDROOT,x	; mandatory
	bne	?10		; error
	sta	rootent
	lsr	a		; divide by 16...
	lsr	a
	lsr	a
	lsr	a		;...to get root dir size in sec's
	sta	rootsiz
	ldy	#$13		; pointer to tot. sec.
	lda	[pbrp],y
	cmp	>FDSEC,x
	bne	?10		; error
	sta	totsec
	ldy	#$16		; pointer to fat size (FAT12)
	lda	[pbrp],y
	cmp	>FDFAT,x
	bne	?10		; error
	sta	fatsiz
	ldy	#$18		; pointer to sector's per track
	lda	[pbrp],y
	cmp	>FDSPT,x
	bne	?10		; error
	ldy	#$1A		; pointer to sides count #
	lda	[pbrp],y
	cmp	#2		; must be 2
	bne	?10		; error
	ldy	#$1C		; pointer to hidden sec
	lda	[pbrp],y
	bne	?10		; must be 0
	ldy	#$1E
	lda	[pbrp],y
	bne	?10		; must be 0
	ldy	#$20		; tot sec32
	lda	[pbrp],y
	bne	?10		; must be 0
	sta	hidsec
	sta	hidsec+2
	sta	totsec+2
	ldy	#$22
	lda	[pbrp],y
	bne	?10		; must be 0
	ldy	#$24		; rsvd1 & rsvd2
	lda	[pbrp],y
	clc
	beq	?12		; must be 0
?10:	sec
?12:	ACC08
	bcs	?16		; exit&error
	ldx	fdix
	ldy	#$0D		; pointer to sector(s) per cluster
	lda	[pbrp],y
	cmp	>FDCLS,x
	bne	?14		; error
	sta	clssiz
	dec	a
	sta	clsmod
	lda	>FDSHF,x
	sta	clsshf
	ldy	#$10		; pointer to count of fat table's
	lda	[pbrp],y
	cmp	#2		; must be 2
	bne	?14
	ldy	#$15		; pointer to media byte
	lda	[pbrp],y
	cmp	>FDMEDIA,x
	beq	?18
?14:	sec
?16:	rts
?18:	sta	media
	clc
	rts

; compute fat param's
fcalc:
	ACC16CLC
	lda	hidsec
	adc	rsvsec
	bcs	?14		; error
	sta	fat1		; fat1 table lba
	adc	fatsiz
	sta	fat2		; fat2 table lba
	bcs	?14		; error
	adc	fatsiz
	bcs	?14		; error
	sta	root		; root dir. lba
	adc	rootsiz
	bcs	?14		; error
	sta	start		; first data sector lba
	sec
	sbc	hidsec
	sta	first		; first relative data sec
	sec
	lda	totsec		; tot. data sec
	sbc	first
	sta	clscnt		; cluster count
	lda	totsec+2
	sbc	#0
	ldy	clsshf
	beq	?12		; no shift
?10:	lsr	a
	ror	clscnt
	dey
	bne	?10
?12:	pha
	pla
	sec
	bne	?14		; error: must be 0
	lda	clscnt
	cmp	maxcnt
	bcc	?13
	lda	#MAXF16-1	; limit max. count
?13:	inc	a		; here CF=0
	inc	a		; max. usable cluster+1=N+2 (2..N+1)
	sta	maxcls
	ldx	clssiz
	txa			; B=0, A=clssiz
	asl	a		; 16 entries per sec...
	asl	a
	asl	a
	asl	a		; here CF=0
	sta	clsent		; entries in cluster
?14:	ACC08
	rts

; move computed fat var's to LDT
mvldt:
	ldx	ldtcur
	ACC16
	lda	root
	sta	ldt_root,x
	lda	fat1
	sta	ldt_fat1,x
	lda	fat2
	sta	ldt_fat2,x
	lda	start
	sta	ldt_cls,x
	lda	maxcls
	sta	ldt_max,x
	lda	rootent
	sta	ldt_rent,x
	lda	clsent
	sta	ldt_cent,x
	lda	freecnt
	sta	ldt_free,x
	lda	freecls
	sta	ldt_nxt,x
	ACC08
	lda	clssiz
	sta	ldt_csiz,x
	lda	clsshf
	sta	ldt_cshf,x
	lda	fatsiz
	sta	ldt_fsiz,x
	lda	rootsiz
	sta	ldt_rsiz,x
	lda	clsmod
	sta	ldt_mcls,x
	lda	fdfmt
	sta	ldt_fmt,x
	rts

;---------------------------------------------------------------------------
; new fat file system
;---------------------------------------------------------------------------

; make a new fresh fat12 file system on floppy disk
; in:	A=format (0 or 2)
;	X=logical drive
;	Y=$FF if label is specified
newf12:
	sta	fdfmt
	stx	ldev
	sty	lflag
	lsr	a
	sta	fdix		; DD/HD index
	txy
	ldx	ldtp,y
	stx	ldtcur
	lda	ldt_fg1,x
	and	#1
	sta	pdev		; phisycal drive
	bit	lflag
	bmi	?10		; label specified
	ldx	#10		; move standard label
?05:	lda	>PBRLBL,x
	sta	!P0FAT+fcbs,x
	dex
	bpl	?05
?10:	ldx	fdix
	lda	>FDMEDIA,x
	sta	media		; media type
	ACC16
	ldx	ldtcur
	lda	ldt_pbr,x	; set PBR pointer
	sta	pbrp
	ldy	ldt_pbr+2,x
	sty	pbrp+2
	lda	ldt_fp,x	; set fat1 table pointer
	sta	tp
	ldy	ldt_fp+2,x
	sty	tp+2
	lda	#F12HDD		; boot sector header DD
	ldy	fdfmt
	beq	?12
	lda	#F12HHD		; boot sector header HD
?12:	sta	tlp
	ldy	#.SEG.F12HDD
	sty	tlp+2
	lda	#$2400		; bytes for DD whole track
	ldy	fdfmt
	beq	?15
	lda	#$4800		; bytes for HD whole track
?15:	ldx	#0
	INDEX16
	tay
	dey
	dey			; index
	txa			; clear C
?20:	sta	[pbrp],y	; clear track buffer
	dey
	dey
	bpl	?20
	ldy	#$01FE		; store signature in boot sector
	lda	#$AA55
	sta	[pbrp],y
	CPU08
	ldy	#F12HSIZ-1	; move boot sector header
?25:	lda	[tlp],y
	sta	[pbrp],y
	dey
	bpl	?25
	ldy	#$26		; extended boot signature flag
	lda	#$29
	sta	[pbrp],y
	ldy	#$2B		; volume label pointer
?30:	lda	!P0FAT+fcbs,x
	sta	[pbrp],y
	inx
	iny
	cpx	#11
	bcc	?30
	ldx	#0
?35:	lda	>F12NAM,x	; store system name	
	sta	[pbrp],y
	inx
	iny
	cpx	#8
	bcc	?35
	jsr	gettime		; get time stamp	
	ldy	#$27
	ACC16
	lda	ctime		; store volume serial number
	sta	[pbrp],y
	iny
	iny
	lda	cdate
	sta	[pbrp],y
	lda	#$0E00		; DD root directory offset
	ldy	fdfmt
	beq	?40
	lda	#$2600		; HD root directory offset
?40:	clc
	adc	pbrp
	sta	tlp		; root directory pointer
	ldy	pbrp+2
	sty	tlp+2
	lda	ctime		; store volume label write/creation time 
	ldy	#$0E		; index of creation
	sta	[tlp],y
	ldy	#$16		; index of write time
	sta	[tlp],y
	iny
	iny
	lda	cdate
	sta	[tlp],y
	ldy	#$10
	sta	[tlp],y	
	ACC08
	ldy	#13
	lda	sec100
	sta	[tlp],y		
	ldx	#0		; store volume label in root directory
	txy
?45:	lda	!P0FAT+fcbs,x
	sta	[tlp],y
	inx
	iny
	cpx	#11
	bcc	?45
	lda	#DA_VOLUME	; store label attribute
	sta	[tlp],y
	ACC16CLC
	lda	#$0600		; fat 2 offset
	ldy	fdfmt
	beq	?50
	lda	#$1200
?50:	adc	tp
	sta	tlp		; fat 2 pointer
	ldy	tp+2
	sty	tlp+2
	lda	#$FFFF		; store fake cluster 0&1
	ldy	#1
	sta	[tp],y
	sta	[tlp],y
	ACC08
	lda	media
	sta	[tp]
	sta	[tlp]
	ldx	pdev		; write whole track on disk
	phx			; drive #
	lda	#0		; track #0
	pha
	lda	sysbuf,x
	pha			; dma buffer #
	lda	fdfmt
	ora	#$80		; verf flag
	pha			; format
	FDCTKWR
nx:	rts			; return CF=1 if error

; create partition on fixed disk X and format it
; in:	X=logical drive
;	Y=$FF if label is specified
lfdisk:
	jsr	prenew		; prepare for partitioning
	bcs	nx
	lda	lunit
	pha
	phk
	pea	#!?08
	lda	#5
	pha
	BPRINTF
	jsr	newprt		; prepare for a new partition
	bcc	?10
	SCNPRINT
	.DB	'error computing a new partition.', 13, 0
	sec
?05:	rts

?08:	.DB	13, 'partitioning fixed disk on drive <%c>:', 13, 0

?10:	lda	#$80
	sta	lflag		; flag creation new pbr/mbr
	bra	lfmt2
	
; format main partition on fixed disk X
; in:	X=logical drive
;	Y=$FF if label is specified
lhdfmt:
	jsr	prefmt		; format pre-task
	bcs	nx
	
lfmt2:
	jsr	showprt
	jsr	askcont
	bcc	?01
	rts
?01:	SCN_CR
	bit	lflag
	bpl	?02
	jsr	newpbr		; create a new pbr/mbr
?02:	lda	fatbnk
	jsr	newfat
	lda	rootbnk
	jsr	newroot
	lda	clsbnk
	jsr	newcls
	jsr	askfmt
	bcc	?04		; quick format
	jsr	fmtcls		; write out cluster's
?04:	SCN_CR
	bit	lflag
	bpl	?06		; skip MBR&PBR write out
	jsr	wrmbr		; write out MBR
	bcs	?10		; error
	jsr	wrpbr		; write out PBR
	bcs	?10
?06:	lda	#0
	xba			; B=0
	lda	fatbnk
	ACC16
	asl	a
	asl	a
	asl	a
	asl	a
	sta	xfatb		; x-mem fat buffer
	lda	fat1
	sta	flba
	ACC08
	ldx	#1		; write out fat1
	jsr	wrfat
	bcs	?10
	lda	fat2
	sta	flba
	lda	fat2+1
	sta	flba+1
	ldx	#2		; write out fat1
	jsr	wrfat
	bcs	?10
	lda	#0
	xba			; B=0
	lda	rootbnk
	ACC16
	asl	a
	asl	a
	asl	a
	asl	a
	sta	xfatb		; x-mem fat buffer
	lda	root
	sta	flba
	ACC08
	jsr	wrroot
	bcs	?10
	ldx	pdev		; update drive flag
	lda	atadev,x
	and	#$E0
	ora	#$1D
	sta	atadev,x
	jsr	mvldt
	lda	#$E0
	sta	ldt_fg2,x
	stz	ldt_cdlvl,x
	stz	ldt_cdcls,x
	stz	ldt_cdcls+1,x	
	clc
?08:	rts
?10:	ldx	ldtcur		; invalidate logical drive if error
	stz	ldt_fg2,x
	ldx	pdev		; invalidate drive if error
	lda	atadev,x
	and	#$E0
	sta	atadev,x
	txa
	asl	a
	asl	a
	tax
	ldy	#0
	CPU16
	stz	atambr,x
	stz	atambr+2,x
	stz	ataprt,x
	stz	ataprt+2,x
	tya			; C=0
?15:	sta	[tlp],y		; clear MBR
	iny
	iny
	cpy	#$200
	bcc	?15
	CPU08
	lda	pdev		; write out an empty invalid MBR
	pha			; ata drive
	lda	#0
	pha
	pea	#0		; lba=0
	lda	#1
	pha			; write 1 sector
	lda	tlp+2
	pha			; buffer pointer
	pei	(tlp)
	lda	#$00		; no verify after write
	pha
	ATAWRITE
	sec			; return error status
?20:	rts

fmtcls:
	SCN_CR
	SCNPRINT
	.DB	'press <CTRL+BREAK> to stop.', 13, 0
	SCNGETPOS
	stx	fd2ix
	lda	#0
	sta	first+1		; cluster size 16 bit
	sta	rsvsec		; lba hi
	ldx	clssiz
	stx	first
	ldx	#XMBANK
	stx	tp+2
	xba			; B=0
	lda	clsbnk
	ACC16
	asl	a
	asl	a
	asl	a
	asl	a
	sta	xfatb		; x-mem fat buffer
	lda	start
	sta	flba		; cluster #2 lba
	lda	#2
	sta	clscnt		; start with cluster #2
	lda	maxcls
	dec	a
	sta	maxcnt
	stz	badcls
	ACC08
?10:	ldx	fd2ix		; set screen position
	ldy	#0
	SCNSETPOS
	pei	(maxcnt)
	pei	(clscnt)
	phk
	pea	#!?100
	lda	#8
	pha
	BPRINTF
	lda	pdev
	pha			; ata drive
	lda	rsvsec
	pha
	pei	(flba)		; fat table lba
	lda	first
	pha			; write n sector
	lda	#0
	pha			; A=0 -> x-mem buffer
	pei	(xfatb)
	lda	#$80		; verify after write
	pha
	ATAWRITE
	bcc	?20		; write ok
	brk
	jsr	?40		; mark bad cluster
?20:	ACC16
	lda	clscnt
	inc	a
	cmp	maxcls
	bcs	?30
	sta	clscnt
	lda	flba
	ldx	#0
	adc	first
	sta	flba
	bcc	?25
	inx
	clc
?25:	stx	rsvsec
?30:	ACC08
	bcs	?35
	lda	!CtrlBrk
	stz	!CtrlBrk
	cmp	#$80
	bne	?10
?35:	SCNPRINT
	.DB	'done.', 13, 0
	pei	(badcls)
	phk
	pea	#!?110
	lda	#6
	pha
	BPRINTF
	rts

?40:	ldx	CRXME		; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	lda	clscnt+1	; A=sector (one sec. = 256 cluster #)
	tax
	lsr	a		; divide by 16 to get x-mem buffer bank
	lsr	a
	lsr	a
	lsr	a
	clc
	adc	fatbnk		; add fat table base base bank
	sta	!PIA0+PIAPRA	; set bank
	txa
	xba
	lda	clscnt
	ACC16
	and	#$0FFF		; mask 4096 cluster's
	asl	a
	sta	tp
	lda	#$FFF7
	sta	[tp]
	inc	badcls
	ACC08
	plx			; restore X mem setting
	sta	!CRXME,x
	rts

?100:	.DB	'clear cluster %05hu of %05hu...', 0
?110:	.DB	'defective cluster(s): %hu', 0

; common pre task routine for format fixed disk
; in:	X=logical drive
;	Y=$FF if label is specified
prenew:
	stx	ldev
	sty	lflag
	txy
	ldx	ldtp,y
	stx	ldtcur
	bit	ldt_fg1,x
	bpl	?01		; not ready
	bvs	?03		; ok, ata device
?01:	tya
	clc
	adc	#'A'
	pha
	phk
	pea	#!?100
	lda	#5
	pha
	BPRINTF
	sec
	rts
?03:	lda	ldt_fg1,x
	and	#1
	sta	pdev		; phisycal drive
	clc
	adc	#'C'		; ata device: 'C' or 'D'
	sta	lunit		; logical drive letter
	bit	lflag
	bmi	?10		; label specified
	ldx	#10		; move standard label
?05:	lda	>NOLBL,x
	sta	!P0FAT+fcbs,x
	dex
	bpl	?05
?10:	stz	lflag	
	lda	pdev
	asl	a
	asl	a
	tax
	ACC16
	lda	!atasec,x	; max. disk usable sector's
	sta	maxsec
	lda	!atasec+2,x
	sta	maxsec+2
	lda	!atambr,x
	sta	hidsec
	lda	!atambr+2,x
	and	#$00FF
	sta	hidsec+2
	lda	!ataprt,x
	sta	totsec
	lda	!ataprt+2,x
	sta	totsec+2	
	ldx	ldtcur
	lda	ldt_pbr,x	; set PBR pointer
	sta	pbrp
	ldy	ldt_pbr+2,x
	sty	pbrp+2
	lda	ldt_fp,x	; set fat1 table pointer
	sta	tp
	ldy	ldt_fp+2,x
	sty	tp+2
	lda	#MBR0ATA	; set pointer for MBR
	ldy	pdev
	beq	?15
	lda	#MBR1ATA
?15:	sta	tlp
	ldy	#^MBR0ATA
	sty	tlp+2
	ldy	ldt_fbuf,x
	sty	fatbnk		; x-mem bank for fat cache buffer
	lda	ldt_root,x
	sta	root
	lda	ldt_fat1,x
	sta	fat1
	lda	ldt_fat2,x
	sta	fat2
	lda	ldt_cls,x
	sta	start
	lda	ldt_rent,x
	sta	rootent
	lda	ldt_cent,x
	sta	clsent
	lda	ldt_free,x
	sta	freecnt
	lda	ldt_nxt,x
	sta	freecls
	ldy	ldt_csiz,x
	sty	clssiz
	lda	ldt_max,x
	sta	maxcls
	dec	a
	dec	a
	sta	clscnt
	ldy	ldt_cshf,x
	sty	clsshf
	ldy	#$B4
	sty	rootbnk		; root dir. x-mem bank
	ldy	#$B8
	sty	clsbnk		; cluster x-mem bank	
	ACC08
	lda	ldt_fsiz,x
	sta	fatsiz
	stz	fatsiz+1
	lda	ldt_rsiz,x
	sta	rootsiz
	lda	ldt_mcls,x
	sta	clsmod
	lda	ldt_fmt,x
	sta	fdfmt
	lda	ldt_fg2,x
	stz	lflag		; invalid partition flag
	and	#$E0		; valid partition?
	cmp	#$E0
	bne	?50		; no
	sta	lflag		; valid partition flag
	ACC16
	lda	clscnt
	sta	remsec
	lda	#0
	ldy	clsshf
	beq	?25
?20:	asl	remsec
	rol	a
	dey
	bne	?20
?25:	sta	remsec+2
	clc
	lda	start
	adc	remsec
	sta	remsec
	bcc	?30
	inc	remsec+2
?30:	sec		
	lda	maxsec
	sbc	remsec
	sta	remsec
	lda	maxsec+2
	sbc	remsec+2
	sta	remsec+2
	ACC08
?50:	clc
	rts

?100:	.DB	13, 'error: drive <%c> not ready or not a fixed disk.', 13, 0

; set a new MBR/PBR for ata drive
newpbr:
	ldy	#0
	lda	pdev
	asl	a
	asl	a
	tax			; device index
	CPU16
	tya			; C=0
?10:	sta	[pbrp],y	; clear PBR...
	sta	[tlp],y		; ...and MBR
	iny
	iny
	cpy	#$01FE
	bcc	?10
	lda	#$AA55		; put signature
	sta	[pbrp],y
	sta	[tlp],y
	ldy	#$01BE		; ptr to first partition in MBR
	ACC08
	lda	#$80
	sta	[tlp],y		; first active partition
	iny			; $01BF
	lda	#1		; fake CHS
	sta	[tlp],y
	iny			; $01C0
	sta	[tlp],y
	iny			; $01C1
	lda	#0
	sta	[tlp],y
	iny			; $01C2
	lda	#6		; partition type = FAT16
	sta	[tlp],y
	iny			; $01C3
	lda	#$FE		; fake CHS
	sta	[tlp],y
	iny			; $01C4
	inc	a
	sta	[tlp],y
	iny			; $01C5
	sta	[tlp],y
	iny			; $01C6 -> partition start lba
	ACC16
	lda	hidsec
	sta	[tlp],y
	sta	!atambr,x
	iny
	iny
	lda	hidsec+2	; this is always null
	sta	[tlp],y
	ora	#$0600		; partition type
	sta	!atambr+2,x
	iny
	iny			; $01CA -> partition total sectors
	lda	totsec
	tax
	sta	[tlp],y
	iny
	iny
	lda	totsec+2
	sta	[tlp],y
	beq	?15		; now switch to PBR	
	ldy	#$22		; partition total sectors 32 high
	sta	[pbrp],y
	dey
	dey
	bra	?20
?15:	ldy	#$13		; partition total sectors 16
?20:	txa
	sta	[pbrp],y
	INDEX08
	ldy	#$1C		; hidden sectors
	lda	hidsec
	sta	[pbrp],y
	iny
	iny
	lda	hidsec+2
	sta	[pbrp],y
	ldy	#$0B		; bytes per sector
	lda	#$0200
	sta	[pbrp],y
	ldy	#$11		; root directory entries
	sta	[pbrp],y
	sta	rootent
	ldy	#$16		; fat table size
	lda	fatsiz
	sta	[pbrp],y
	ldy	#$0E		; reserved sectors count
	lda	#1
	sta	[pbrp],y
	inc	a		; C=2
	CPU08
	ldy	#$10		; fat's count = 2
	sta	[pbrp],y
	ldy	#$0D		; sector(s) per cluster
	lda	clssiz
	sta	[pbrp],y
	ldy	#$15		; media type = F8
	lda	#$F8
	sta	[pbrp],y
	ldy	#$26		; extended signature flag
	lda	#$29
	sta	[pbrp],y
	ldy	#3+8-1		; OEM name end
	ldx	#7
?25:	lda	>OEMNAM,x
	sta	[pbrp],y
	dey
	dex
	bpl	?25
	inx			; X=0
	ldy	#$2B		; volume label pointer
?30:	lda	!P0FAT+fcbs,x
	sta	[pbrp],y
	inx
	iny
	cpx	#11
	bcc	?30
	ldx	#0
?35:	lda	>F16NAM,x	; store system name string	
	sta	[pbrp],y
	inx
	iny
	cpx	#8
	bcc	?35
	lda	pdev
	asl	a
	asl	a
	tax			; device index
	ACC16
	lda	totsec
	sta	ataprt,x
	lda	totsec+2
	sta	ataprt+2,x
	ACC08
	jsr	gettime		; get time stamp
	lda	#$EB		; this signature in first 3 bytes...
	sta	[pbrp]		; ...only for microsoft!
	ldy	#1
	lda	#0
	sta	[pbrp],y
	iny
	lda	#$90
	sta	[pbrp],y	
	ldy	#$27
	ACC16
	lda	ctime		; store volume serial number
	sta	[pbrp],y
	iny
	iny
	lda	cdate
	sta	[pbrp],y
	ldx	clssiz
	txy
	txa			; B=0, A=clssiz
	asl	a		; 16 entries per sec...
	asl	a
	asl	a
	asl	a		; here CF=0
	sta	clsent		; entries in cluster
	dey
	sty	clsmod
	lda	clscnt
	sta	freecnt
	lda	#2
	sta	freecls	
	ACC08
	lda	#$FF
	sta	fdfmt
	rts

; new empty fat table
; in:	A=fat bank buffer
newfat:
	ldx	CRXME		; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	phb			; save DBR
	ldx	#XMBANK
	phx
	plb			; set DBR to fat table cache
	ldy	#16		; clear 16 8k-blocks
	jsr	clrxmem
	CPU16
	ldx	#XMSTART
	lda	#$FFF8		; set fake cluster 0&1
	sta	!0,x
	lda	#$FFFF
	sta	!2,x
	CPU08	
	plb			; restore DBR
	plx			; restore X mem setting
	sta	!CRXME,x
	rts

; new empty root directory
; in:	A=root directory bank buffer
newroot:
	ldx	CRXME		; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	phb			; save DBR
	ldx	#XMBANK
	phx
	plb			; set DBR to fat table cache
	ldy	#2		; clear 2 8k-blocks
	jsr	clrxmem
	ldx	#10
	txy			; set volume label
	lda	#DA_VOLUME
	sta	!XMSTART+1,y
?10:	lda	>P0FAT+fcbs,x
	sta	!XMSTART,y
	dey
	dex
	bpl	?10
	ACC16
	lda	ctime		; set creation time&date...
	sta	!XMSTART+14
	sta	!XMSTART+22	; ...and write time&date
	lda	cdate
	sta	!XMSTART+16
	sta	!XMSTART+24
	ACC08
	lda	sec100
	sta	!XMSTART+13	
	plb			; restore DBR
	plx			; restore X mem setting
	sta	!CRXME,x
	rts

; new empty cluster
; in:	A=cluster bank buffer
newcls:
	ldx	CRXME		; save X mem setting
	phx
	sta	!CRXMEON	; enable X mem in window $012000 - $013FFF
	phb			; save DBR
	ldx	#XMBANK
	phx
	plb			; set DBR to fat table cache
	ldy	#4		; clear 4 8k-blocks
	;jsr	clrxmem
	jsr	clrxmem2

	plb			; restore DBR
	plx			; restore X mem setting
	sta	!CRXME,x
	rts

; clear x-mem buffer's
; in:	A=starting bank
;	Y=bank count
;	DBR=x-mem bank
; out:	x-mem bank set to starting bank
clrxmem:
	sta	xbnk
	sta	>PIA0+PIAPRA	
?10:	CPU16
	ldx	#XMSTART	; start of x-mem window
?15:	stz	!0,x
	inx
	inx
	cpx	#XMSTOP
	bcc	?15
	CPU08
	lda	>PIA0+PIAPRA	; update x-mem bank
	inc	a
	sta	>PIA0+PIAPRA	
	dey
	bne	?10		; next block
	lda	xbnk
	sta	>PIA0+PIAPRA	
	rts

clrxmem2:
	sta	xbnk
	sta	>PIA0+PIAPRA	
?10:	INDEX16
	ldx	#XMSTART	; start of x-mem window
	lda	#0
?15:	sta	!0,x
	inx
	inc	a
	cpx	#XMSTOP
	bcc	?15
	CPU08
	lda	>PIA0+PIAPRA	; update x-mem bank
	inc	a
	sta	>PIA0+PIAPRA	
	dey
	bne	?10		; next block
	lda	xbnk
	sta	>PIA0+PIAPRA	
	rts
	
; write out root directory (xfatb=x-mem root dir. cache buffer)
wrroot:
	lda	lunit
	pha
	phk
	pea	#!?10
	lda	#5
	pha
	BPRINTF
	lda	pdev
	pha			; ata drive
	lda	#0
	pha
	pei	(root)		; root dir. lba
	ldy	#32
	phy			; write 32 sector
	pha			; A=0 -> x-mem buffer
	pei	(xfatb)
	lda	#$80		; verify after write
	pha
	ATAWRITE
	bcs	?05
	jmp	wrok
?05:	jmp	wrerr		; error

?10:	.DB	'write root directory on drive <%c>...', 0

; write out PBR (pointer: pbrp)
wrpbr:
	lda	lunit
	pha
	phk
	pea	#!?10
	lda	#5
	pha
	BPRINTF
	lda	pdev
	pha			; ata drive
	lda	#0
	pha
	pei	(hidsec)	; lba=partition start
	lda	#1
	pha			; write 1 sector
	lda	pbrp+2
	pha			; buffer pointer
	pei	(pbrp)
	lda	#$80		; verify after write
	pha
	ATAWRITE
	bcc	wrok
	bcs	wrerr		; error

?10:	.DB	'write PBR on drive <%c>...', 0

; write out MBR (pointer: tlp)
wrmbr:
	lda	lunit
	pha
	phk
	pea	#!?10
	lda	#5
	pha
	BPRINTF
	lda	pdev
	pha			; ata drive
	lda	#0
	pha
	pea	#0		; lba=0
	lda	#1
	pha			; write 1 sector
	lda	tlp+2
	pha			; buffer pointer
	pei	(tlp)
	lda	#$80		; verify after write
	pha
	ATAWRITE
	bcc	wrok
	bcs	wrerr		; error

?10:	.DB	'write MBR on drive <%c>...', 0
	
wrok:
	SCNPRINT
	.DB	'done.', 13, 0
	clc
	rts			; CF=1 if error

wrerr:
	phx			; bit error
	phy			; error code
	phk
	pea	#!?10
	lda	#6
	pha
	BPRINTF
	sec
	rts
	
?10:	.DB	'error %02bX (bits error: %02bX).', 13, 0

; write out fat table (xfatb=x-mem fat cache buffer, X=fat)
wrfat:
	lda	lunit
	pha
	phx
	phk
	pea	#!?10
	lda	#6
	pha
	BPRINTF
	lda	pdev
	pha			; ata drive
	lda	#0
	pha
	pei	(flba)		; fat table lba
	ldy	fatsiz
	phy			; write n sector
	pha			; A=0 -> x-mem buffer
	pei	(xfatb)
	lda	#$80		; verify after write
	pha
	ATAWRITE
	bcc	wrok
	bcs	wrerr		; error

?10:	.DB	'write FAT%bu table on drive <%c>...', 0

; A=drive
shpart:
	tax
	ldy	#0
	jsr	prenew		; move data of partition
	bcs	?30
	lda	lunit
	pha
	phk
	bit	lflag
	bmi	?20		; valid partition
	pea	#!?110
	lda	#5
	pha
	BPRINTF
	bra	?30	
?20:	pea	#!?100
	lda	#5
	pha
	BPRINTF
	jsr	showprt
?30:	rts

?100:	.DB	13, 'partition summary on drive <%c>:', 13, 0
?110:	.DB	13, 'no valid partition found on drive <%c>.', 13, 0

atanewp:
	.PUBLIC atanewp
	phd
	pea	#P0LDT
	pld
	ldy	#0
	jsr	lfdisk
	pld
	rtl

; fixed disk pre-format routine 
; in:	X=logical drive
;	Y=$FF if label is specified 
prefmt:
	jsr	prenew		; move data of partition
	bcs	?30		; no valid partition
	lda	lunit
	pha
	phk
	bit	lflag
	bmi	?20		; valid partition
	pea	#!?110
	lda	#5
	pha
	BPRINTF
	bra	?30	
?20:	pea	#!?100
	lda	#5
	pha
	BPRINTF
	jsr	showprt
	clc
?30:	rts

?100:	.DB	13, 'format main partition on drive <%c>:', 13, 0
?110:	.DB	13, 'no valid partition found on drive <%c>.', 13, 0
	
; print partition data
showprt:
	pei	(maxsec+2)
	pei	(maxsec)
	pei	(maxsec+2)
	pei	(maxsec)
	phk
	pea	#!?10
	lda	#12
	pha
	BPRINTF
	pei	(hidsec+2)
	pei	(hidsec)
	pei	(hidsec+2)
	pei	(hidsec)
	phk
	pea	#!?20
	lda	#12
	pha
	BPRINTF
	pei	(totsec+2)
	pei	(totsec)
	pei	(totsec+2)
	pei	(totsec)
	phk
	pea	#!?30
	lda	#12
	pha
	BPRINTF
	pei	(remsec+2)
	pei	(remsec)
	pei	(remsec+2)
	pei	(remsec)
	phk
	pea	#!?40
	lda	#12
	pha
	BPRINTF
	pei	(fatsiz)
	pei	(fatsiz)
	phk
	pea	#!?50
	lda	#8
	pha
	BPRINTF
	pei	(clscnt)
	pei	(clscnt)
	phk
	pea	#!?60
	lda	#8
	pha
	BPRINTF
	lda	clssiz
	pha
	pha
	phk
	pea	#!?70
	lda	#6
	pha
	BPRINTF
	pea	#0
	pei	(fat1)
	phk
	pea	#!?75
	lda	#8
	pha
	BPRINTF
	pea	#0
	pei	(fat2)
	phk
	pea	#!?80
	lda	#8
	pha
	BPRINTF
	pea	#0
	pei	(root)
	phk
	pea	#!?85
	lda	#8
	pha
	BPRINTF
	pea	#0
	pei	(start)
	phk
	pea	#!?90
	lda	#8
	pha
	BPRINTF	
	rts

?10:	.DB	'total sector(s) : %08X %, +10u', 13, 0
?20:	.DB	'partition start : %08X %, +10u', 13, 0
?30:	.DB	'partition size  : %08X %, +10u', 13, 0
?40:	.DB	'unused sector(s): %08X %, +10u', 13, 0
?50:	.DB	'fat table size  : % 8.4hX %, +10hu', 13, 0
?60:	.DB	'clusters count  : % 8.4hX %, +10hu', 13, 0
?70:	.DB	'cluster size    : % 8.4bX %, +10bu sector(s)', 13, 0
?75:	.DB	'fat1 table start: %08X', 13, 0
?80:	.DB	'fat2 table start: %08X', 13, 0
?85:	.DB	'root dir. start : %08X', 13, 0
?90:	.DB	'cluster #2 start: %08X', 13, 0

; compute a new disk partition
newprt:
	lda	pdev		; phisycal drive
	asl	a
	asl	a
	tax
	lda	defhsec
	sta	hidsec		; hidden sector's
	stz	hidsec+1
	ACC16
	stz	hidsec+2
	lda	!atasec,x	; max. disk usable sector's
	sta	maxsec
	lda	!atasec+2,x
	sta	maxsec+2
	lda	#1
	sta	rsvsec		; reserved sector for PBR
	lda	#32
	sta	rootsiz		; root directory sector's
	lda	hidsec
	inc	a		; leave one sector at partition end
	sta	remsec
	sec
	lda	maxsec
	sbc	remsec
	sta	totsec		; total partition sector's
	lda	maxsec+2
	sbc	hidsec+2	; this is null
	sta	totsec+2
	clc
	lda	hidsec		; compute total reserved sector's
	adc	rsvsec
	adc	rootsiz
	sta	totrsv
	lda	maxsec+2
	cmp	#$0040		; disk size > 2gb?
	bcc	?05		; no
	ldx	#$40		; yes, so...
	stx	clssiz		; ...cluser size=64 sector's...
	ldx	#6
	stx	clsshf
	lda	#$0100		; ...and fat size=max (256 sector's)...
	sta	fatsiz
	lda	#MAXF16-1
	sta	clscnt		; ...and cluster's count=max ($FFF4)...
	lda	#$FF30		; ...and set partition totsec=$3FFF30
	sta	totsec
	lda	#$003F
	sta	totsec+2
	jmp	?50		; done for disk > 2Gb
?05:	CPU08			; now disk <= 2Gb
	lda	maxsec+2
	ldx	#0
?10:	cmp	>CLSSIZE,x
	bcc	?15		; X=shift
	inx
	cpx	#7
	bcc	?10
	dex
?15:	lda	>CLSSIZE,x
	sta	clssiz		; cluster size
	stx	clsshf		; cluster shift
	CPU16
	sec			; compute approximate fat size
	lda	maxsec
	sbc	totrsv
	sta	maxcnt
	lda	maxsec+2
	sbc	#0
	txy
	beq	?25		; no shift
?20:	lsr	a		; divide by cluster size
	ror	maxcnt
	dex
	bne	?20
?25:	tax			; now C must be cleared
	bne	?48		; error... shoul'nt happen
	CPU08
	lda	maxcnt+1	; high byte is fat size!
	inc	a
	bne	?30
	inx			; fat size=256
?30:	sta	fatsiz
	stx	fatsiz+1
	ACC16CLC
?35:	lda	rsvsec		; now compute clusters count
	adc	fatsiz
	adc	fatsiz
	adc	rootsiz
	sta	clscnt
	sec
	lda	totsec
	sbc	clscnt
	sta	clscnt
	lda	totsec+2
	sbc	#0
	ldx	clsshf
	INDEX16
	beq	?45
?40:	lsr	a
	ror	clscnt
	dex
	bne	?40
?45:	tax			; now C must be cleared
	bne	?48		; error... shoul'nt happen
	INDEX08
	lda	clscnt
	cmp	#MAXF16
	bcc	?46
	lda	#MAXF16-1
	clc
?46:	inc	a		; add fake cluster 0&1
	inc	a
	adc	#$00FF		; round to fit sector
	and	#$FF00
	xba
	bcc	?47
	adc	#$0100
?47:	cmp	fatsiz
	beq	?50		; ok, fat size is right
	bcc	?50		; ok, fat size is right
	inc	fatsiz		; increment fat size
	clc
	bra	?35		; try again
?48:	sec
	bra	?55
?50:	INDEX08
	clc
	lda	totrsv
	adc	fatsiz
	adc	fatsiz
	sta	totrsv
	lda	clscnt
	sta	remsec
	inc	a
	inc	a
	sta	maxcls
	lda	#0
	ldx	clsshf
	beq	?52
?51:	asl	remsec
	rol	a
	dex
	bne	?51
?52:	sta	remsec+2
	lda	remsec
	clc
	adc	totrsv
	sta	remsec
	bcc	?53
	inc	remsec+2
?53:	lda	maxsec
	sec
	sbc	remsec
	sta	remsec
	lda	maxsec+2
	sbc	remsec+2
	sta	remsec+2
	clc
	lda	hidsec
	adc	rsvsec
	sta	fat1
	adc	fatsiz
	sta	fat2
	adc	fatsiz
	sta	root
	adc	rootsiz
	sta	start
?55:	CPU08
	rts

askfmt:
	SCNPRINT
	.DB	'continue with the quick format? <Y/N>?', 0
?10:	jsr	getkey
	bcs	?10		; control key
	tax
	xba
	tay
	bne	?10		; CTL, ALT, SHIFT
	cpx	#'y'
	clc
	beq	?20
	cpx	#'n'
	bne	?10
?20:	rts

; return CF=0 if YES, CF=1 if NO
askcont:
	SCNPRINT
	.DB	'WARNING! Proceeding, all data on the disk will be erased. Continue <Y/N>?', 0
?10:	jsr	getkey
	bcs	?10		; control key
	tax
	xba
	tay
	bne	?10		; CTL, ALT, SHIFT
	cpx	#'y'
	clc
	beq	?20
	cpx	#'n'
	bne	?10
?20:	rts
	
; wait a pressed key
getkey:
?02:	KBGETIN			; system call - get key
	cmp	#0
	beq	?02		; loop -- wait pressed key
	tax			; X=key code
	xba
	clc
	tay			; Y=control key flag
	bpl	?04
	sec			; flag control key
?04:	lda	>DP01ADDR+KbdFlag	
	xba			; B = flag ALT, CTL, SHIFT
	clc
	txa			; key code
	rts

;---------------------------------------------------------------------------
; compact time function's
;---------------------------------------------------------------------------

; get compact time for file timestamp
; return cpu in 16 bit mode
; C=low compact time, X=high compact time
gctime16:
	CPU08
	phd
	pea	#P0LDT
	pld			; set dp page
	jsr	getrtc
	jsr	cpctime
	CPU16
	lda	ctime
	ldx	cdate
	pld
	rts
	
	.LONGA	off
	.LONGI	off

; get compact time for file timestamp
getctime:
	phd
	pea	#P0LDT
	pld			; set dp page
	jsr	getrtc
	jsr	cpctime
	pld
	rts

; compact time
gettime:
	jsr	getrtc
cpctime:
	sec
	lda	rtcyear		; subctract 1980
	sbc	#$50		; without century...
	clc			; ...we assume century=20
	adc	#$64		; add 100
	xba			; B = 0.127 (till 2101)
	lda	rtcmonth	; shift month
	asl	a
	asl	a
	asl	a
	asl	a
	ACC16
	asl	a		; shift year+month
	ACC08
	ora	rtcday
	sta	cdate		; compact date
	xba
	sta	cdate+1
	lda	rtcsec
	lsr	a
	sta	ctime
	stz	sec100
	bcc	?10
	lda	#100
	sta	sec100		; millisecond stamp
?10:	lda	rtchour
	xba
	lda	rtcmin
	asl	a
	asl	a
	ACC16
	asl	a
	asl	a
	asl	a
	ACC08
	ora	ctime		; compact time
	sta	ctime
	xba
	sta	ctime+1
	rts
	
; get real time clock
getrtc:
	phb
	lda	#0
	pha
	plb
	php
	sei
	ldx	#RTCCTRLA	; set bank 1
	stx	RTCALE
	lda	#$10
	tsb	RTCDATA
	ldx	#RTCEXTCTRLA	; check INCR bit 6
	stx	RTCALE
?02:	bit	RTCDATA
	bvs	?02
	ldx	#0		; seconds
	stx	RTCALE
	lda	RTCDATA
	sta	rtcsec
	inx
	inx
	stx	RTCALE		; minutes
	lda	RTCDATA
	sta	rtcmin
	inx
	inx
	stx	RTCALE		; hours
	lda	RTCDATA
	sta	rtchour
	inx
	inx
	inx
	stx	RTCALE		; day
	lda	RTCDATA
	sta	rtcday
	inx
	stx	RTCALE		; month
	lda	RTCDATA
	sta	rtcmonth
	inx
	stx	RTCALE		; year
	lda	RTCDATA
	sta	rtcyear
	plp
	plb
	rts

;---------------------------------------------------------------------------
; tables
;---------------------------------------------------------------------------
	
; sector's per cluster table
CLSSIZE:
	.DB	$01, $02, $04, $08, $10, $20, $40

CLSCNT	.EQU	$-CLSSIZE

; sector's per cluster shift table   	
CLSSHIFT:
	.DB	$00, $01, $02, $03, $04, $05, $06
	
; costant's for fat formatted floppy disk (D.D. & H.D. format)
; root dir. entry count
FDROOT:
	.DW	$0070, $00E0
; total sector's 
FDSEC:
	.DW	$05A0, $0B40
; fat size
FDFAT:
	.DW	$0003, $0009
; sector's per track
FDSPT:
	.DW	9, 18
; sector per cluster
FDCLS:
	.DB	2, 1
; shift value
FDSHF:
	.DB	1, 0
; media value
FDMEDIA:
	.DB	$F9, $F0

; ata device PBR buffer pointer
PBRPTR:
	.DW	.LOW16.PBR0ATA, .LOW16.PBR1ATA
PBRPTR2:
	.DB	.SEG.PBR0ATA, .SEG.PBR1ATA

; standard label 
PBRLBL:
	.DB	'DOS DISK   '

NOLBL:
	.DB	'NO NAME    '

; standard fat12 system name
F12NAM:
	.DB	'FAT12   '

; standard fat16 system name
F16NAM:
	.DB	'FAT16   '

; boot sector header DD fat12
F12HDD:
	.DB	$EB, $34, $90
OEMNAM:
	.DB	'IBM  3.3'
	.DB	$00, $02, $02, $01, $00, $02
	.DB	$70, $00, $A0, $05, $F9, $03, $00
	.DB	$12, $00, $02, $00
	
F12HSIZ	.EQU	($ - F12HDD)

; boot sector header HD fat12
F12HHD:
	.DB	$EB, $34, $90
	.DB	'IBM  3.3'
	.DB	$00, $02, $01, $01, $00, $02
	.DB	$E0, $00, $40, $0B, $F0, $09, $00
	.DB	$12, $00, $02, $00
