;;
;; Copyright (c) 2016 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; name: toscmd.asm	
;; rev.: 2016/22/10
;; o.s. 65C816 version v1.0

;; built in o.s. command's 

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\tossys.inc
	.INCLUDE src\F9\fstmp.inc

.LIST on

;---------------------------------------------------------------------------
; code segment -- bank $F9
;---------------------------------------------------------------------------

	.CODEF9
	.EXTERN cbmfs_cmd, cbmfs_err	; fscbm.asm
	.EXTERN lhdfmt, shpart, lfdisk	; fsint.asm
	.EXTERN	prslbl			; fatdir.asm
	.EXTERN _getopt			; FA/getopt.asm

	.LONGA	off
	.LONGI	off

;---------------------------------------------------------------------------
; System interface -- built in command's
;---------------------------------------------------------------------------

lpCmd	.SET	STKPRMS

tos_cmd:
	.GLOBAL	tos_cmd

	lda	STKPBR,s	; bank where was executed cop instruction
	sta	COPPtr+2
	ACC16			; retrieve function number
	lda	STKPCL,s	; pointer to byte after signature
	sta	COPPtr
	inc	a		; update return address
	sta	STKPCL,s
	lda	[COPPtr]	; byte after signature: function number #
	tax			; X=function number
	lda	#P0LDT		; set dpr
	tcd
	tsc			; save stack pointer
	sta	cmdstk
	lda	lpCmd,s
	sta	tlp
	ACC08
	lda	lpCmd+2,s
	sta	tlp+2
	cpx	#MAXCMDS
	bcs	?10		; invalid function
	txa
	asl	a
	tax			; function index
	jsr	(CMDJMP,x)
	bra	?15
?10:	phx
	phk
	pea	#!?20
	lda	#5
	pha
	BPRINTF
	sec
?15:	lda	#1		; return always 1 if error
	rtl

?20:	.DB	'invalid shell function number (%bu).', 13, 0

; fdisk command
;
;	fdisk [-s] X:		; show partition on drive X
;	fdisk -d X:		; delete partition on drive X
;	fdisk -c [-llabel] X:	; create partition on drive X and format it
;
fdisk:
	jsr	fdskopt		; get option (A=drive)
	bcs	?10
	ldy	fdopt
	bne	?02
	jmp	shpart		; show partition
?02:	dey
	beq	?04		; delete partition
	ldy	fdlbl		; flag label
	tax			; drive
	jmp	lfdisk		; create & format
?04:
?10:	rts

; format logical drive
;
;	param's (5 bytes)
;
;		bDrive		logical drive number
;		bFmt		format (0,1 or 2) + quick flag (bit 7)
;		lpLabel		long ptr to label string (volume name)
;
; param's offset
lpLabel	.SET	STKPRMS + 2
bFmt	.SET	STKPRMS + 5
bDrive	.SET	STKPRMS + 6

format:
	lda	bDrive,s
	jsr	chkdrv		; valid drive number?
	sta	ldev		; logical drive
	tay
	lda	bFmt,s
	sta	lflag		; quick format flag
	and	#3
	cmp	#3
	bne	?04
	lda	#2		; force dos HD format
?04:	sta	fdfmt
	ldx	ldtp,y
	stx	ldtcur		; LDT ptr
	jsr	chkrdy		; drive is ready?	
	ACC16
	lda	lpLabel,s	; get label string pointer
	sta	!P0FAT+pathp
	sta	tlp
	ACC08
	ldy	#0
	lda	lpLabel+2,s
	sta	tlp+2		; test null ptr
	sta	!P0FAT+pathp+2
	ora	tlp+1
	ora	tlp
	beq	?06		; no label
	dey			; Y=$FF
?06:	sty	nulptr		; flag null ptr
	bit	ldt_fg1,x	; test fixed disk/floppy disk
	bvs	?12		; fixed disk
	jsr	tstfd		; test floppy disk format
	lda	getfmt		; here X=phisycal drive
	bmi	?08		; can't use quick format
	cmp	fdfmt		; compare with requested format
	beq	?10		; we can use quick format
?08:	stz	lflag		; standard full fd format
?10:	lda	fdfmt
	cmp	#1
	;bne	?12		; ibm-dos format
	beq	cbmfmt		; cbm format
?12:	tya
	beq	?16		; no volume name specified
	phd
	pea	#P0FAT
	pld
	jsr	prslbl		; parse label string
	pld
	bcc	?16
	beq	?14		; empty string
	jmp	invlbl		; invalid label
?14:	stz	nulptr		; like if not string given
?16:	ldx	ldtcur
	bit	ldt_fg1,x	; test fixed disk/floppy disk
	bvc	?18		; floppy disk
	ldx	pdev
	ldy	nulptr
	jmp	lhdfmt
?18:	rts

; format cbm diskette (X=phisycal drive) 
cbmfmt:
	lda	nulptr
	beq	nolbl		; cbm format need a volume name
	lda	#'Q'		; send command "Qx:...." (quick format)
	sta	bpath
	txa
	ora	#'0'
	sta	bpath+1
	lda	#':'
	sta	bpath+2
	ldy	#0
	ldx	#3
?02:	lda	[tlp],y
	sta	bpath,x
	beq	?04
	iny
	inx
	cpx	#32		; max. 32 char's
	bcc	?02
	stz	bpath,x		; trunc
?04:	lda	#bpath		; address of command string
	ldx	#>P0LDT
	ldy	#0
	jsr	cbmfs_cmd	; cbm quick format
	bcs	cbmerr
	rts
	
;---------------------------------------------------------------------------
; check routines & error handling
;---------------------------------------------------------------------------

; report cbm emulator error
cbmerr:
	jsr	cbmfs_err	; return error string pointer (C,X)
	SCNSTROUT		; print error string
	bra	cmderr

; no label/disk name specified error
nolbl:
	SCNPRINT
	.DB	'no label nor cbm disk name specified', 0
	bra	cmderr

; test floppy disk and get format (X=LDT pointer)
; return X=phisycal drive
tstfd:
	lda	ldt_fg1,x
	and	#1
	tax			; phisycal drive
	FDCTST			; check media inside drive
	bcc	?02		; media ready
	cpy	#FDC_CHANGE	; check error code
	bne	notrdy		; disk not ready
?02:	ldy	#$FF
	sty	getfmt		; invalid disk format
	FDCGETFMT		; try to get media format
	bcs	?04		; unformatted disk or unknow disk format
	sty	getfmt		; current disk format
?04:	rts

; check drive ready (X=LDT pinter)
chkrdy:
	bit	ldt_fg1,x
	bmi	noer
notrdy:	SCNPRINT
	.DB	'drive not ready', 0
	bra	cmderr

noer:	rts
	
; check drive number (A=drive #)
chkdrv:
	cmp	#MAXDRV
	bcc	noer
invdrv:	SCNPRINT
	.DB	'invalid drive specified', 0
cmderr:
	SCNPRINT
	.DB	'.', 13, 0
	ACC16CLC
	lda	cmdstk		; recover stack
	tcs
	ACC08
	lda	#1
	sec
	rtl			; return to caller

; invalid label
invlbl:
	SCNPRINT
	.DB	'invalid label', 0
	bra	cmderr

; missing drive
nodrv:
	SCNPRINT
	.DB	'missing drive', 0
	bra	cmderr

; parse command line for 'fdisk' command
fdskopt:
	stz	fdopt
	stz	fdarg
	stz	fdlbl
	ldx	#$80
	phx			; flag
	ldx	tlp+2		; command line
	phx
	pei	(tlp)
	phk			; option string
	pea	#!?100		
	pea	#0		; result
	pea	#0
?10:	jsl	_getopt		; get opttion's loop
	bcs	?20		; error
	tax			; option char
	beq	?20		; end of command line
	bmi	?18		; no option argument
	cmp	#'s'
	bne	?12
	stz	fdopt		; show
	bra	?10
?12:	cmp	#'d'
	bne	?14
	lda	#1
	sta	fdopt		; delete
	bra	?10
?14:	cmp	#'c'
	bne	?16
	lda	#2
	sta	fdopt		; create
	bra	?10
?16:	lda	$01,s		; argument for 'l' option
	sta	!P0FAT+pathp
	lda	$02,s		; store argument pointer
	sta	!P0FAT+pathp+1
	lda	$03,s
	sta	!P0FAT+pathp+2
	lda	#$FF
	sta	fdlbl	
	bra	?10
?18:	lda	$01,s		; argument without option
	sta	tlp
	lda	$02,s		; store argument pointer
	sta	tlp+1
	lda	$03,s
	sta	tlp+2
	lda	#$80
	sta	fdarg
	bra	?10
?20:	php			; clean stack
	plx			; save carry
	ACC16CLC
	tsc
	adc	#11
	tcs
	ACC08
	txa
	lsr	a		; test carry
	bcs	?30		; exit
	bit	fdarg
	bmi	?21
	jmp	nodrv
?21:	jsr	drvarg
	cmp	#ATADRV
	bcc	?40
	ldy	fdopt
	cpy	#2
	bne	?26
	bit	fdlbl
	bpl	?26
	sta	ldev
	phd
	pea	#P0FAT
	pld
	jsr	prslbl		; parse label string
	pld
	bcc	?24
	beq	?22		; empty string
	jmp	invlbl		; invalid label
?22:	stz	fdlbl		; like if not string given
?24:	lda	ldev	
?26:	clc
?30:	rts
?40:	adc	#'A'	
	pha
	phk
	pea	#!?110
	lda	#5
	pha
	BPRINTF
	jmp	cmderr

?100:	.DB	's c d l:', 0
?110:	.DB	'drive <%c> is not a fixed disk', 0

; check if drive argument is valid
drvarg:
	ldy	#1
	lda	[tlp],y
	cmp	#':'
	bne	?20
	iny
	lda	[tlp],y
	bne	?20
	lda	[tlp]
	cmp	#'a'
	bcc	?08
	cmp	#'z'+1
	bcs	?10
	and	#$DF
?08:	sec
?10:	sbc	#'A'
	bcc	?20
	jmp	chkdrv
?20:	jmp	invdrv

;---------------------------------------------------------------------------
; Tables
;---------------------------------------------------------------------------

; command's jump table
CMDJMP:
	.DW	fdisk, format

MAXCMDS		.EQU	($ - CMDJMP)/2
