;;
;; Copyright (c) 2016 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; name: utils.asm	
;; rev.: 2016/22/10
;; o.s. 65C816 version v1.0

;; utilities & helper's functions

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\tossys.inc

.LIST on

;---------------------------------------------------------------------------
; code segment -- bank $F9
;---------------------------------------------------------------------------

	.CODEF9
	.EXTERN oserror
	.EXTERN bldldt						; fsint.sm
	.PUBLIC mxdrv, fndhcb, fndbuf, cls2lba

	.LONGA	off
	.LONGI	off

; switch logical drive
;
;	entry:	A = logical drive number
;
;	exit:	curdrv = current logical drive number
;		pdrive = current phsycal drive
;		LDT struct of the drive moved in page 0
;
;	use:	A,X,Y
;
;	note:	if any error this function never return to the caller
;
;-----
mxdrv:
;-----
	cmp	#MAXDRV
	bcc	?02
	lda	#ED_DRIVE	; invalid drive
	jmp	oserror
?02:	tay			; drive to select
	ldx	!P0LDT+ldtp,y	; LDT pointer
	bit	!P0LDT+ldt_fg1,x	; drive flag
	bpl	?05		; drive not ready
	bvs	?10		; fixed disk: we can trust...
	sty	ldrv
	stx	tmpldt	
	lda	!P0LDT+ldt_fg1,x
	and	#1
	tax
	FDCTST			; check media inside drive
	bcs	?04		; media changed or not ready?
	ldx	tmpldt
	lda	!P0LDT+ldt_fg2,x
	bit	#$20		; test bit 5
	bne	?08		; ok, disk already checked
	beq	?06		; must check disk
?04:	phy			; save error code
	phx			; save phisycal drive
	ldx	ldrv		; discard any HCB block for this drive
	jsr	hdscdrv
	ldx	ldrv		; discard any buffer for this drive
	jsr	bdscdrv	
	plx			; restore phisycal drive
	ply			; restore fdc error code
	cpy	#FDC_CHANGE	; check error code
	bne	?05		; disk not ready
	FDCTST			; check again
	bcc	?06		; ok
?05:	lda	#ED_NOTRDY	; disk not ready
	jmp	oserror
?06:	lda	ldrv
	jsr	bldldt		; build LDT table for this drive
	bcs	?40
?08:	ldx	tmpldt
	ldy	ldrv
?10:	lda	!P0LDT+ldt_fg2,x
	and	#$A0		; bit 7&5 will be set
	cmp	#$A0
	bne	?34		; invalid format
	cpy	curdrv		; change drive?
	beq	?20		; no, use current table
	sty	curdrv		; set up curdrv
	txy			; LDT pointer
	ldx	#LDTSIZE-1	; move LDT table in current page 0
?12:	lda	!P0LDT+ldt0+LDTSIZE-1,y
	sta	ldtfg1,x
	dey
	dex
	bpl	?12
	and	#1
	sta	pdrive
	tay
	bit	ldtfg1
	bvs	?20		; fixed disk
	lda	#9		; 9 sec. per track
	ldx	ldtfmt
	beq	?14
	lda	#18		; 18 sec. per track
?14:	sta	fspt
	ACC16
	lda	#dma0s
	tyx
	beq	?18
	lda	#dma1s
?18:	sta	dmabcb
	ACC08
?20:	rts

?34:	lda	#ED_FORMAT
?40:	jmp	oserror

;---------------------------------------------------------------------------
; HCB helpers
;---------------------------------------------------------------------------
	
fndhcb2:
	phb			; save current DBR
	lda	#^hcb_data
	pha
	plb			; set DBR to HCB struct's
	CPU16
	jsr	fndhcb
	CPU08
	plb			; restore DBR
	rts

; find HCB for given drive,cluster and sequence
;
;	entry:	hcbcls  = cluster (=0 if root dir.)			
;		hcbseq  = sector offset sequence
;		hcbdrv  = logical drive
;		DBR	= bank that hold HCB struct's
;		CF	= 0 if exact match required
;		CF	= 1 if we want to recycle oldest HCB
;
;	exit:	CF	= 0 if HCB found
;		hcbroot	= HCB pointer if found
;		buffer loaded from disk
;		CF	= 1 if HCB was not found
;
;	use:	C,X,Y
;
;	note:	this routine should be called with 16 bit cpu mode
;
;------
fndhcb:
;------
	stz	hcbrec		; recycle flag
	bcc	?05		; no recycle
	dec	hcbrec		; =$FFFF, recycle
?05:	ldx	hcbroot		; start from head of list
	bra	?15
?10:	lda	!hcb_next,x	; onto the next entry
	beq	?20		; no one there...
	txy			; Y=previous HCB	
	tax			; X=current HCB
?15:	lda	!hcb_seq,x	; sequence+drive
	bmi	?20		; from here start the free list
	cmp	hcbseq		; does seq+drive match?
	bne	?10		; goto next if not
	lda	!hcb_cls,x
	cmp	hcbcls		; does cluster match?
	bne	?10		; goto next if not
	clc			; HCB found
	beq	?30		; we have a match: find & load buffer
?20:	sec			; we have been all along the chain w/no luck
	bit	hcbrec		; do we want to recycle oldest HCB?
	bpl	?40		; no, so HCB not found & exit with CF=1
	stz	hcb_cnt,x	; recycle oldest HCB...
	lda	hcbseq
	sta	!hcb_seq,x	; ...so mark as us, but with nothing in it
	lda	hcbcls
	sta	!hcb_cls,x
?30:	php			; save carry	
	cpx	hcbroot		; this HCB is the head of list?
	beq	?32		; yes, no need to move ahead
	lda	!hcb_next,x	; get link to the rest of chain
	sta	!hcb_next,y	; unlink ourselves from chain
	lda	hcbroot
	stx	hcbroot		; move current entry at the head...
	sta	!hcb_next,x	; ...and relink the rest of the chain
?32:	cpx	hcblst		; this HCB was the last one of list?
	bne	?35		; no
	sty	hcblst		; yes...so the previous become the last one
?35:	plp			; restore carry
	bcc	?40		; HCB found, not need to fill it
	jsr	hcbfill		; fill it
	clc
?40:	rts

	.LONGA	off		; for assembler
	.LONGI	off

; discard HCB for given drive, cluster and sequence
;
;	entry:	C = cluster
;		X = logical drive
;		Y = sequence #
;
;	out:	nothing
;
;	use:	all
;
; 	note:	
;-------
hdscblk:
;-------
	ACC16
	sta	hcbcls		; cluster to discard
	stx	hcbdrv		; drive
	sty	hcbseq		; sequence
	lda	#$FFFF
	sta	bdmask		; drive+sec mask
	sta	bcmask		; cluster mask
	bra	hdscrd		; discard

; discard HCB's for given drive & cluster
;
;	entry:	C = cluster
;		X = logical drive
;
;	out:	nothing
;
;	use:	all
;
; 	note:	
;-------
hdsccls:
;-------
	ACC16
	sta	hcbcls		; cluster to discard
	stx	hcbdrv		; drive
	ldy	#0
	sty	hcbseq		; all sector's
	lda	#$FFFF
	sta	bcmask		; cluster mask
	lda	#$FF00	
	sta	bdmask		; drive mask (all sector's)
	bra	hdscrd		; discard

; discard HCB's for given drive
;
;	entry:	X = logical drive
;
;	out:	nothing
;
;	use:	all
;
; 	note:	
;-------
hdscdrv:
;-------
	ACC16
	stz	hcbcls		; all cluster's
	stx	hcbdrv		; drive
	ldy	#0
	sty	hcbseq		; all sector's
	stz	bcmask		; cluster mask (all cluster's)
	lda	#$FF00
	sta	bdmask		; drive mask (all sector's)

; discard HCB's
hdscrd:
	phb			; save dbr
	ldx	#HCBBNK		; set working dbr
	phx
	plb
	CPU16
	ldx	hcbroot		; list root
	bra	?15
?10:	lda	!hcb_next,x	; next BCB
	beq	?30		; done
	txy			; Y=previous BCB
	tax			; X=current BCB
?15:	lda	!hcb_seq,x
	bmi	?30		; no more BCB...done
	and	bdmask		; mask drive+sec
	cmp	hcbseq		; match?
	bne	?10		; no...go to next BCB
	lda	!hcb_cls,x
	and	bcmask		; mask cluster
	cmp	bcbcls		; match?
	bne	?10		; no...go to next BCB
	lda	!hcb_next,x	; get link to the rest of chain
	cpx	hcbroot		; is the root of the list?
	bne	?20		; no
	sta	hcbroot		; unlink ourselves from chain
	bra	?25
?20:	sta	!hcb_next,y	; unlink ourselves from chain
?25:	pha			; next BCB to check
	lda	#$FFFF		; put this BCB in free list
	sta	!hcb_seq,x
	lda	hcblst		; the actual last one
	stx	hcblst		; this BCB become the last one
	stz	!hcb_next,x
	tax			; link the rest of list to the last one
	lda	hcblst
	sta	!hcb_next,x
	plx			; next BCB to check
	bne	?15
?30:	CPU08
	plb			; restore dbr
	rts

; fill an HCB with many hashed entries as much is possible
;
;	entry:	hcbroot = HCB pointer
;		hcbcls  = cluster (=0 if root dir.)			
;		hcbseq  = sector offset sequence
;		hcbdrv  = logical drive
;		DBR     = bank that hold HCB struct's
;
;	exit:	HCB is filled with hashed directory entries
;		hcb_cnt field is negative if no more entries
;
;	use:	C,X,Y
;
;	note:	this routine can be called in any cpu mode (8/16 bit)
;
;-------
hcbfill:
;-------
	php
	CPU08
	lda	hcbseq		; what sequence number?
	beq	?10		; first...start at offset zero
	ldx	hlog2		; shift count
	beq	?10		; no shift (HCB can contain 1 sector)
?05:	asl	a
	dex
	bne	?05
?10:	sta	secofs
	tay			; Y=sector's offset
	lda	dsiz		; directory size in sector's
	sec
	sbc	secofs		; how many sector's left to hash...
	cmp	hsiz		; do we support this many?
	bcc	?15
	lda	hsiz		; ...else limit it to this many
?15:	tax			; X=sector's count
	ACC16
	lda	hcbroot
	sta	hlast		; starting hash code pointer
?20:	sty	secofs		; Y=sector offset
	stx	seccnt		; X=sector's count
	lda	hcbcls		; C=cluster
	ldx	hcbdrv		; X=drive
	sec			; preread from disk
	jsr	fndbuf		; find & fill sector buffer (can return CF=1)
	jsr	hcbblk		; hash 16 entries (preserve CF)
	ldx	seccnt
	dex			; how many sector's left?
	beq	?40		; all sector's loaded&hashed
	bcs	?30		; no more entries after this sector...
	ldy	secofs
	iny			; next sector
	bra	?20		; load&hash next sector
?30:	CPU16	
	txa			
	asl	a
	asl	a
	asl	a
	asl	a
	tay			; we have this many null entries
	ldx	hlast		; last hash code ptr
?35:	stz	!hcb_data,x	; zap rest of cluster
	inx
	inx
	dey
	bne	?35
	stx	hlast		; update last ptr
	ldy	hcbroot
	clc
	adc	hcb_cnt,y
	bra	?45		; no more entries...
?40:	bcc	?50		; more entries...
	ldy	hcbroot
	lda	hcb_cnt,y
?45:	ora	#$8000		; negative count: no more entries after...
	sta	hcb_cnt,y
?50:	plp
	rts

	.LONGA	off		; for assembler
	.LONGI	off

; hash a directory block (16 entries)
;
;	entry:	hcbroot = HCB pointer
;		hlast   = start pointer where store hash codes  
;		fcbp    = long pointer to directory block
;		DBR     = bank that hold HCB struct's
;
;	exit:	hlast=pointer to next available free entry
;
;	use:	C,X,Y
;
;	note:	this routine can be called in any cpu mode (8/16 bit)
;
;------
hcbblk:
;------
	php			; save status
	CPU16
	ldx	#16		; 16 entries
?05:	phx			; save counter
	stz	thash		; reset hash code
	lda	[fcbp]
	and	#$00FF		; mask off 2nd character
	beq	?15		; empty entry so thash=0 
	cmp	#$00E5
	beq	?10		; deleted entry so thash=0
	tax			; save first character
	ldy	#11		; get attribute
	lda	[fcbp],y
	and	#$003F		; long file name entry?
	cmp	#$000F
	beq	?10		; yes so thash=0
	txa	
	CPU08
	and	#$7F
	sta	thash+1		; initialize hash code MSB
	ldx	#10		; involve other 10 characters... 
	ldy	#1		; ...starting from 2nd
?08:	ACC16
	lda	thash		; this simulate a 17 bit rotation
	asl	a		; CF=old A<15>
	adc	#$0000		; A<0>=CF=old A<15>
	sta	thash
	ACC08
	lda	[fcbp],y	; get next char
	and	#$7F
	eor	thash		; xor char into the hash code
	sta	thash
	iny
	dex
	bne	?08		; repeat for all characters
	CPU16CLC
	lda	thash
	bne	?15		; thash=0 reserved for empty/deleted/lfn entry
	inc	a
	bra	?15
?10:	lda	thash		; here thash=0
?15:	plx			; restore counter
	ldy	hlast		; Y=hash code data pointer
	sta	hcb_data,y
	iny			; bump ptr
	iny
	sty	hlast
	lda	fcbp		; here always CF=0
	adc	#32		; bump fcbp pointer
	sta	fcbp
	dex
	bne	?05		; next entry
	clc
	ldx	hcbroot
	lda	!hcb_cnt,x	; update entries count
	adc	#16
	sta	!hcb_cnt,x
	plp			; restore cpu status
	rts

	.LONGA	off		; for assembler
	.LONGI	off
	
;---------------------------------------------------------------------------
; disk buffer's handler
;---------------------------------------------------------------------------

; find buffer for given drive,cluster&sector's offset
;
;	entry:	C  = cluster (16 bit)
;		X  = logical drive
;		Y  = sector's offset from cluster start
;		CF = 1 if will preread buffer
;
;	exit:	fcbp=long pointer to sector, x-mem bank set to right value
;		xbuf=transfer buffer number
;		lba&lbah=sector address
;		BCB pointer moved at head of list (bcbroot)
;		buffer loaded from disk sector if caller wished it
;		CF=1 if this sector is the last useful sector of directory
;
;	use:	C,X,Y
;
;	note:	this routine can be called in any cpu mode (8/16 bit)
;		if disk error this routine never return to the caller
;
;------
fndbuf:
;------
	clc			; normal exit carry flag
	php			; save cpu status
	CPU08
	stz	bcbload		; preread flag
	bcc	?05		; no preread
	dec	bcbload		; preread
?05:	phb			; save current DBR
	stx	bcbdrv		; logical drive
	sty	bcbsec		; sector's offset
	ldx	#^bcbstart
	phx
	plb			; set DBR to BCB struct's
	bit	ldtfg1		; ata device or floppy?
	CPU16
	sta	bcbcls		; cluster #
	bvs	?08		; ata device
	jmp	?50		; floppy disk
?08:	ldx	bcbroot		; start from head of list
	bra	?15		; search BCB
?10:	lda	!bcb_next,x	; onto the next entry
	beq	?20		; no one there...
	txy			; Y=previous BCB	
	tax			; X=current BCB
?15:	lda	!bcb_drv,x
	bmi	?20		; from here start the free list
	cmp	bcbdrv		; does drive+sector match?
	bne	?10		; goto next if not
	lda	!bcb_cls,x
	cmp	bcbcls		; does cluster match?
	bne	?10		; goto next if not
	lda	!bcb_lba,x	; we have a match...
	sta	lba		; ...so get lba address of sector...
	lda	!bcb_lbah,x
	sta	lbah
	lda	!bcb_buf,x	; ...and get transfer buffer #
	sta	xbuf 
	bra	?30		; set buffer pointer
?20:	lda	bcbdrv		; or is the first one free BCB...
	sta	!bcb_drv,x	; ...or we recycle the oldest one
	lda	bcbcls		; ...and we use it
	sta	!bcb_cls,x
	lda	!bcb_buf,x	; get transfer buffer #
	sta	xbuf
	phx			; save current BCB ptr
	phy			; save previous BCB ptr
	lda	bcbcls		; ...and compute lba address of sector
	CPU08
	ldx	bcbsec		; sector offset within cluster
	jsr	cls2lba		; get lba of sector
	clc
	bit	bcbload		; need to preread buffer?
	bpl	?25		; no
	lda	ldtfg1
	and	#1
	pha			; phisycal device
	lda	lbah		; 24 bit lba only
	pha
	pei	(lba)		; lba address
	lda	#1
	pha			; read one sec
	lda	#0
	pha			; x-mem buffer
	pei	(xbuf)		; buffer#
	pha			; flag (ignored)
	ATAREAD			; read from ata device
	sty	ioerr		; save i/o error if any
	stx	ioerr2		; save bit error if any
?25:	CPU16
	ply			; restore previous BCB
	plx			; restore current BCB ptr
	bcc	?27		; no i/o error
	lda	#$FFFF		; invalidate buffer
	sta	!bcb_drv,x
	jmp	?200
?27:	lda	lba		; save lba address onto current BCB
	sta	!bcb_lba,x
	lda	lbah
	sta	!bcb_lbah,x
?30:	lda	!bcb_ptr,x	; set buffer pointer
	sta	fcbp
	lda	!bcb_xbnk,x
	sta	fcbp+2		; set hi ptr & x-mem bank
	cpx	bcbroot		; this BCB is at the head of list?
	beq	?35		; yes, not need to move
	lda	!bcb_next,x	; get link to the rest of chain
	sta	!bcb_next,y	; unlink ourselves from chain
	lda	bcbroot
	stx	bcbroot		; move current entry at the head...
	sta	!bcb_next,x	; ...and relink the rest of the chain
	cpx	bcblst		; this BCB was the last one of list?
	bne	?35		; no
	sty	bcblst		; yes...so the previous become the last one
?35:	ldy	#$01FE		; pointer to last directory entry	
	ACC08
	lda	fcbp+3
	sta	>PIA0+PIAPRA	; set x-mem bank
	plb			; restore DBR
	lda	[fcbp],y	; get first name character
	bne	?38		; not null entry, exit with CF=0
	lda	$01,s		; no more entry after this sector...
	ora	#PCFLAG		; ...so set CF in saved status
	sta	$01,s
?38:	plp
	rts

?50:	; find dma buffer (C=cluster)

	.LONGA	off		; remember... here cpu mode 16 bit
	.LONGI	off

	cmp	#5		; root dir, cluser #2,#3&#4 ijn separate list
	bcs	?55
	jmp	?100
?55:	ldx	dmaroot		; start from head of list
	bra	?65		; search BCB
?60:	lda	!bcb_next,x	; onto the next entry
	beq	?70		; no one there...
	txy			; Y=previous BCB	
	tax			; X=current BCB
?65:	lda	!bcb_drv,x
	bmi	?70		; from here start the free list
	cmp	bcbdrv		; does drive+sector match?
	bne	?60		; goto next if not
	lda	!bcb_cls,x
	cmp	bcbcls		; does cluster match?
	bne	?60		; goto next if not
	lda	!bcb_lba,x	; we have a match...
	sta	lba		; ...so get lba address of sector...
	ACC08
	lda	!bcb_head,x
	sta	fhead		; ...and head...
	ACC16
	lda	!bcb_trk,x
	sta	ftrack		; ...and track & sector
	lda	!bcb_ptr,x	; set buffer pointer
	sta	fcbp
	lda	!bcb_dma,x
	sta	fcbp+2		; set hi ptr & dma buffer
	bra	?80		; done
?70:	lda	bcbdrv		; or is the first one free BCB...
	sta	!bcb_drv,x	; ...or we recycle the oldest one
	lda	bcbcls		; ...and we use it
	sta	!bcb_cls,x
	lda	!bcb_ptr,x	; set buffer pointer
	sta	fcbp
	lda	!bcb_dma,x
	sta	fcbp+2		; set hi ptr & dma buffer	
	phx			; save current BCB ptr
	phy			; save previous BCB ptr
	lda	bcbcls		; ...and compute lba address of sector
	CPU08
	ldx	bcbsec		; sector offset within cluster
	jsr	cls2lba		; get lba of sector
	jsr	lba2hts		; get floppy tuplet H,T,S
	clc
	bit	bcbload		; need to preread buffer?
	bpl	?75		; no
	lda	ldtfg1		; read sector from floppy disk
	and	#1
	pha			; phisycal device
	lda	ftrack
	ldx	fhead
	ldy	fsec
	pha			; track
	phx			; head
	phy			; sector
	lda	#1		; read one sector
	pha
	lda	dmabuf
	pha			; dma buffer #
	lda	ldtfmt
	pha			; format
	FDCREAD			; fdc read
	sty	ioerr		; save i/o error if any
	stz	ioerr2		; no extended i/o error for fdc
?75:	CPU16
	ply			; restore previous BCB
	plx			; restore current BCB ptr
	bcc	?77		; ok, no read errror
	lda	#$FFFF		; invalidate buffer
	sta	!bcb_drv,x
	bra	?200		; i/o error
?77:	lda	lba		; store lba address onto current BCB
	sta	!bcb_lba,x
	lda	fhead		; store head
	sta	!bcb_head,x
	lda	ftrack		; store track&sector
	sta	!bcb_trk,x
?80:	cpx	dmaroot		; this BCB is at the head of list?
	beq	?85		; yes, not need to move
	lda	!bcb_next,x	; get link to the rest of chain
	sta	!bcb_next,y	; unlink ourselves from chain
	lda	dmaroot
	stx	dmaroot		; move current entry at the head...
	sta	!bcb_next,x	; ...and relink the rest of the chain
	cpx	dmalst		; this BCB was the last one of list?
	bne	?85		; no
	sty	dmalst		; yes...so the previous become the last one
?85:	ldy	#$01FE		; pointer to last directory entry	
	ACC08
	plb			; restore DBR
	lda	[fcbp],y	; get first name character
	bne	?90		; not null entry, exit with CF=0
	lda	$01,s		; no more entry after this sector...
	ora	#PCFLAG		; ...so set CF in saved status
	sta	$01,s
?90:	plp
	rts

; floppy disk root dir, cluster #2,#3,#4
?100:	ldx	dmabcb		; BCB list in systerm track (track 0)
	bra	?115
?110:	lda	!bcb_next,x	; onto the next entry
	beq	?120		; no one there...
	tax			; X=current BCB
?115:	lda	!bcb_drv,x
	bmi	?120		; from here start the free list
	cmp	bcbdrv		; does drive+sector match?
	bne	?110		; goto next if not
	lda	!bcb_cls,x
	cmp	bcbcls		; does cluster match?
	bne	?110		; goto next if not
	lda	!bcb_lba,x	; we have a match...
	sta	lba		; ...so get lba address of sector...
	ACC08
	lda	!bcb_head,x
	sta	fhead		; ...and head...
	ACC16
	lda	!bcb_trk,x
	sta	ftrack		; ...and track & sector
	lda	!bcb_ptr,x	; set buffer pointer
	sta	fcbp
	lda	!bcb_dma,x
	sta	fcbp+2		; set hi ptr & dma buffer
	bra	?85		; done
?120:	; no BCB...should'nt happen this

	;
	
?200:	; i/o error
	; load from disk error 
	CPU08
	plb			; restore DBR
	plp
	sec			; error
	rts

; discard buffer for given drive, cluster & sector offset
;
;	mode:	A/M/X/Y 8 bit
;
;	entry:	C = cluster
;		X = logical drive
;		Y = sector offset within cluster
;
;	out:	BCB's list affected		
;
;	use:	all
;
; 	note:	
;-------
bdscsec:
;-------
	ACC16
	sta	bcbcls		; cluster to discard
	stx	bcbdrv		; drive
	sty	bcbsec		; sector
	lda	#$FFFF
	sta	bdmask		; drive+sec mask
	sta	bcmask		; cluster mask
	bra	bdscrd		; discard

; discard buffer's of given drive & cluster
;
;	mode:	A/M/X/Y 8 bit
;
;	entry:	C = cluster
;		X = logical drive
;
;	out:	BCB's list affected
;
;	use:	all
;
; 	note:	
;-------
bdsccls:
;-------
	ACC16
	sta	bcbcls		; cluster to discard
	stx	bcbdrv		; drive
	ldy	#0
	sty	bcbsec		; all sector's
	lda	#$FFFF
	sta	bcmask		; cluster mask
	lda	#$00FF	
	sta	bdmask		; drive mask (all sector's)
	bra	bdscrd		; discard

; discard buffer's of given drive
;
;	mode:	A/M/X/Y 8 bit
;
;	entry:	X = logical drive
;
;	out:	BCB's list affected
;
;	use:	all
;
; 	note:	
;-------
bdscdrv:
;-------
	ACC16
	stz	bcbcls		; all cluster's
	stx	bcbdrv		; drive
	ldy	#0
	sty	bcbsec		; all sector's
	stz	bcmask		; cluster mask (all cluster's)
	lda	#$00FF	
	sta	bdmask		; drive mask (all sector's)

; discard buffer's
bdscrd:
	ldy	#0
	cpx	#ATADRV
	bcc	?02
	dey
?02:	sty	ldrv+1		; flag fdc/ata
	phb			; save dbr
	ldx	#WKB		; set working dbr
	phx
	plb
	CPU16
	ldx	bcbroot		; fixed disk list
	ldy	bcblst
	bit	ldrv
	bmi	?05
	ldx	dmaroot		; fd list
	ldy	dmalst
?05:	stx	troot		; head of list
	sty	tlst		; end of list
	bra	?15
?10:	lda	!bcb_next,x	; next BCB
	beq	?30		; done
	txy			; Y=previous BCB
	tax			; X=current BCB
?15:	lda	!bcb_drv,x
	bmi	?30		; no more BCB...done
	and	bdmask		; mask drive+sec
	cmp	bcbdrv		; match?
	bne	?10		; no...go to next BCB
	lda	!bcb_cls,x
	and	bcmask		; mask cluster
	cmp	bcbcls		; match?
	bne	?10		; no...go to next BCB
	lda	!bcb_next,x	; get link to the rest of chain
	cpx	troot		; is the root of the list?
	bne	?20		; no
	sta	troot		; unlink ourselves from chain
	bra	?25
?20:	sta	!bcb_next,y	; unlink ourselves from chain
?25:	pha			; next BCB to check
	lda	#$FFFF		; put this BCB in free list
	sta	!bcb_drv,x
	lda	tlst		; the actual last one
	stx	tlst		; this BCB become the last one
	stz	!bcb_next,x
	tax			; link the rest of list to the last one
	lda	tlst
	sta	!bcb_next,x
	plx			; next BCB to check
	bne	?15
?30:	ldx	troot
	ldy	troot
	bit	ldrv
	bmi	?35
	stx	dmaroot		; update fd list
	sty	dmalst
	lda	bcmask		; discard drive?
	bne	?40		; no
	ldx	#dma0s
	lda	bcbdrv
	and	#$00FF
	beq	?32		; drive #0
	ldx	#dma1s
?32:	lda	#$FFFF
	sta	!bcb_drv,x	; empty list
	stz	!bcb_next,x
	bra	?40
?35:	stx	bcbroot		; update fixed disk list
	sty	bcblst
?40:	CPU08
	plb			; restore dbr
	rts

;---------------------------------------------------------------------------
; utilities
;---------------------------------------------------------------------------

; compute lba address from a given cluster
;
;	entry:	C = cluster (16 bit) -- if C = 0 return lba of root directory
;		X = sector offset (8 bit)
;
;	exit:	lba and lbah filled with lba address
;
; 	use:	C,X,Y
;
;	note:	this routine should be called in 8 bit cpu mode
;
;-------
cls2lba:
;-------
	stx	wtmp		; sector's offset...
	stz	wtmp+1		;...extended to 16 bit
	ldx	ldtcshf		; shift count#
	CPU16
	stz	lbah
	tay			; compute lba for root dir.?
	beq	?20		; yes
	dec	a		; skip fake cluster#0&#1
	dec	a
	txy
	beq	?10		; no shift
?05:	asl	a		; compute lba base
	rol	lbah
	dex
	bne	?05
?10:	clc
	ldy	lbah		; Y=lbah
	adc	ldtcls		; add offset of first data sector
	bcc	?32
	iny
	bra	?30
?20:	lda	ldtroot		; lba address of root dir. (here Y=0)
?30:	clc
?32:	adc	wtmp		; add sector's offset
	sta	lba
	bcc	?40	
	iny			; update lbah
?40:	sty	lbah
	CPU08
	rts

	.LONGA	off		; for assembler
	.LONGI	off

; translate lba address into tuplet (H,T,S) for floppy disk operation
;
; remind:
;	H   = head number (0..1)
;	T   = track number (0..79)
;	S   = sector number (1..max) where max=9 or 18
;	SPT = sector per tack (18 or 36)
;
;	LBA = (H + 2*T)*SPT + (S - 1)
;
;	T   = LBA / (2 * SPT)
;	L   = LBA MOD (2 * SPT)
;	H   = L / P
;	S   = (L MOD P) + 1
;
;	entry:	lba = lba address
;		fspt = sector per track
;
;	exit:	fhead, ftrack, fsec
;
;	use:	all
;
;-------
lba2hts:
;-------
	lda	fspt
	asl	a		; 2*SPT
	sta	dvsor
	stz	dvsor+1		; divisor=2*SPT
	lda	lba+1
	xba
	lda	lba		; C=lba
	jsr	fudiv		; return C=quotient
	sta	ftrack		; store track, ignore B (should be null)
	ldy	#0		; Y=head
	ldx	dvsor		; remainder (L)
	txa
	sec
	sbc	fspt		; get sector & head
	bcc	?10		; L <  SPT so head=0, sec=L+1
	iny			; head=1
	tax			; L >= SPT so head=1, sec=L-SPT+1
?10:	inx			; X=sector
	sty	fhead
	stx	fsec
	rts

; fast unsigned division 16 bit
;
;	entry:	C = 16 bit dividend
;		dvsor = 16 bit divisor
;
;	exit:	C = quot = 16 bit quotient
;		dvsor = 16 bit remainder
;
;	use:	all
;
;	note:	no check for null divisor
;
;-----
fudiv:
;-----
	ldx	#1		; bit counter
	CPU16
	;ldy	lba		; Y=dividend
	tay
	stz	quot		; init quotient
	lda	dvsor		; C=divisor
?10:	asl	a		; shift divisor: get leftmost bit
	bcs	?20		; go to division
	inx
	cpx	#17		; test all divisor bit's
	bne	?10
?20:	ror	a		; put shifted-out bit back
	sta	dvsor
?30:	tya			; get dividend
	sec
	sbc	dvsor
	bcc	?40		; can't subctract, retain old dividend
	tay			; Y=new dividend
?40:	rol	quot		; shift carry into quotient (1 if division)
	lsr	dvsor		; shift right divisor for next subtract
	dex
	bne	?30
	sty	dvsor		; store remainder
	lda	quot		; C=quotient
	CPU08
	rts

ftest:
	phd
	pea	#P0FAT
	pld
	stx	fspt
	sta	lba
	xba
	sta	lba+1
	jsr	lba2hts
	lda	#0
	xba
	lda	ftrack
	ldx	fsec
	ldy	fhead	
	pld
	brk
	brk
	
;---------------------------------------------------------------------------
; tables
;---------------------------------------------------------------------------

dlmstr	.DB	':.;,=+\<>|/"[]'	; path delimeters
DLMSTR	.EQU	$-dlmstr

devlst:
	.DB	'CON     '	; Keyboard and display
	.DB	'KEYBD$  '	; Keyboard
	.DB	'SCREEN$ '	; Display
	.DB	'CONSOLE$'
	.DB	'PRN     '	; System list device, usually a parallel port
	.DB	'LST     '	; System list device, usually a parallel port
	.DB	'LPT1    '	; First parallel printer port
	.DB	'LPT2    '	; Second parallel printer port
	.DB	'LPT3    '	; Third parallel printer port
	.DB	'AUX     '	; Auxiliary device, usually a serial port
	.DB	'COM1    '	; First serial communications port
	.DB	'COM2    '	; Second serial communications port
	.DB	'COM3    '	; Third serial communications port
	.DB	'COM4    '	; Fourth serial communications port
	.DB	'NUL     '	; Bit-bucket device
	.DB	'$IDLE$  '	; Bit-bucket device
	.DB	'CLOCK$  '	; System real-time clock
	.DB	'CLOCK   '	; System real-time clock
	.DB	'CONFIG$ '
	.DB	0
