;;
;; Copyright (c) 2016 Marco Granati <mg@unet.bz>
;;
;; Permission to use, copy, modify, and distribute this software for any
;; purpose with or without fee is hereby granted, provided that the above
;; copyright notice and this permission notice appear in all copies.
;;
;; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;;

;; name: lib1.asm	
;; rev.: 2016/22/10
;; o.s. 65C816 version v1.0

;; shared library: getopt function

	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc

.LIST on

; page 0 local var's (declared in bios temp. work area)
_P0BTMP:	.SECTION page0, ref_only, offset bwrktmp	; DP Tmp. BIOS

cmdlin		LP		; long pointer to command line
optstr		LP		; long pointer to options mask
dwres		.LWORD		; result
slen		.DB		; string length
squot		.DB		; flag quotation marks
cmdy		.DB		; index
chopt		.DB		; option char
prsfg		.DB		; parsing flag
opttyp		.DB		; option type
intarg		.DS	4	; integer argument
tmpp		LP
argsgn		.DB

	.ENDS

;---------------------------------------------------------------------------
; code segment -- bank $FA
;---------------------------------------------------------------------------

	.CODEFA
	.EXTERN _lstrtoint

	.LONGA	off
	.LONGI	off

; getopt
;
; parse a command line and return option&argument
; command line is parsed basis on option string, a mask that specify option ch
; and argument type. Any option start with '-' character, followed by a single
; character; if option requires an argument, this argument follow (without
; blanks) the character option. String option can be quoted (with quotation
; marks, this need when argument contains blanks).
; Option string have this simple skeleton:
;
;	xtxtxtxt....
;
; where x is a single option character and t is the argument type
; recognized argument types:
;	t=	blank, mean option does not requires argument
;		':', '$': mean option requires a string argument
;			  type '$' return an up case string
;		'!', '^': mean option requires a single char argument
; 			  type '^' return an up case char
;		'%'	: byte argument required
;		'#'	: integer 16 bit argument required
;		'*'	: integer 24 bit argument required (ptr type)
;		'&'	: integer 32 bit argument required
;
; option string example: 'r s:t^'
; option '-r' without argument
; option '-s' requires a string
; option '-t' requires a single character, returned in up case
; an example of valid command line for above mask:
;	xxxxx  -szzzzz -r -tC  yyyyy
; string 'yyyyy' is an argument without option
;
; function return
; getopt does not clean stack frame when exit and parameter's will be modified
; sCmd change at any call and hold the residual command line string
; sRes change at any call and hold the argument (long pointer to string,
; a single character, a numeric argument)
; Y return the option type associated with the current parsed option
; A return: 0 if reached the end of command line, the current parsed option,
; or a negative value if return an argument without option: in this last case
; sRes return a string.
;
; calling getopt will be in a loop until A=0 or until error; the caller should
; clean stack when finish to parse command line
;
; param's:
;
;		sFlag	= <7>: show error message
;			  <6>: accept decimal numeric arguments
;			  <5>: wanted signed int arg. (only if bit 6 is set)
;			  <0>: called from monitor
;		sCmd	= long pointer to command line
;		sOpt	= long pointer to option's string
;		sRes	= 32 bit result
;
; exit:
;		sCmd	= residual command line string
;		sRes	= argument (depend on type)
;		A	= option character currently parsed
;			  or =0 when end of string,
;			  or negative when parsed argument without option
;		Y	= option type associated with option
;
; if any error, function return CF=1, otherwise CF=0
;
; param's stack offset
sRes	.SET	6
sOpt	.SET	10
sCmd	.SET	13
sFlag	.SET	16

;----------------------
_getopt:
	.PUBLIC _getopt
;----------------------
	phd
	pea	#0
	pld
	ACC16			; move param's in local var's
	lda	sOpt,s
	sta	optstr
	lda	sCmd,s
	sta	cmdlin
	stz	intarg
	stz	intarg+2
	ACC08
	lda	sOpt+2,s
	sta	optstr+2
	lda	sCmd+2,s
	sta	cmdlin+2
	lda	sFlag,s
	sta	prsfg
	ldy	#0
	sty	opttyp
	sty	argsgn
	jsr	?104		; get next char
	beq	?94		; return A=0
	cmp	#'-'		; option?
	beq	?05		; yes
	jsr	?110		; find end of argument
	bcs	?95		; error
	lda	#$80		; no option argument
	sta	chopt
	bra	?90
?05:	jsr	?130		; check option
	bcs	?95		; error
	jsr	?140		; test option argument
	bcs	?95		; error
	ldx	opttyp
	cpx	#' '		; option without argument?
	beq	?90		; yes...so exit
	cpx	#':'		; option requires a simple string argument?
	beq	?90		; yes...so exit
	cpx	#'$'		; option requires up case string argument?
	beq	?10		; yes
	cpx	#'!'		; option requires a single char argument?
	beq	?30		; yes
	cpx	#'^'		; option requires a single up case char arg.?
	bne	?40		; no
?10:	ldy	#0
?15:	lda	[dwres],y	; up case string
	beq	?25		; end of string
	cmp	#'a'
	bcc	?20
	cmp	#'z'+1
	bcs	?20
	and	#$DF
	sta	[dwres],y
?20:	iny
	bra	?15
?25:	cpx	#'$'		; if requires string exit
	beq	?90
?30:	jsr	?150		; check single char argument
	bcs	?95		; error
	bra	?90
?40:	jsr	?160		; others option type
	bcs	?95
?90:	ACC16			; return result to the caller...
	lda	cmdlin		; ...and residual command line string
	sta	sCmd,s
	lda	dwres
	sta	sRes,s
	lda	dwres+2
	sta	sRes+2,s
	ACC08
	lda	cmdlin+2
	sta	sCmd+2,s
	ldy	opttyp		; return Y=option type
	lda	chopt		; return A=option char
?94:	clc			; no error
?95:	pld
	rtl

?100:	iny			; get char from command line
?104:	lda	[cmdlin],y
	beq	?106
	cmp	#' '
	beq	?100
?106:	rts

?110:	ldx	#' '	
	stx	squot		; normal end of string char
	stz	slen		; size of argument
	cmp	#'"'		; search end of string
	bne	?112		; not quoted
	sta	squot		; quoted string
	iny
?112:	sty	cmdy		; save index	
?114:	lda	[cmdlin],y
	bne	?115
	ldx	squot
	cpx	#'"'
	beq	?122		; missing final quote
	bra	?117
?115:	cmp	squot		; first blank/quote is end of string
	beq	?116
	inc	slen
	iny
	bra	?114
?116:	lda	#0		; put string terminator
	sta	[cmdlin],y
	iny			; next token in command line
?117:	ldx	slen		; check string length
	beq	?126		; error
	ldx	cmdy
	ACC16CLC
	txa
	adc	cmdlin
	sta	dwres		; pointer to argument
	ACC08
	lda	cmdlin+2
	bcc	?118
	inc	a
?118:	sta	dwres+2
	stz	dwres+3
?119:	ACC16CLC
	tya
	adc	cmdlin
	sta	cmdlin		; bump command line ptr
	ACC08
	bcc	?120
	inc	cmdlin+2
	clc
?120:	rts
?122:	lda	prsfg		; test bit 7
	bpl	?124		; no show error message
	jsr	?128
	SCNPRINT
	.DB	'missig quotation marks in argument.', 13, 0 
?124:	sec			; error flag
	rts
?126:	lda	prsfg
	bpl	?124
	jsr	?128
	SCNPRINT
	.DB	'argument zero length.', 13, 0 
?127:	sec
	rts
?128:	lsr	a
	bcc	?129
	SCNPRINT
	.DB	'::', 0
?129:	rts

?130:	iny			; check option
	lda	[cmdlin],y
	beq	?135		; missing option
	cmp	#' '
	beq	?135		; missing option
	tyx
	sta	chopt		; store option char
	ldy	#0		; scan option string
?132:	lda	[optstr],y
	beq	?136		; option not found
	cmp	chopt
	beq	?134		; found
	iny
	iny
	bra	?132
?134:	iny
	lda	[optstr],y	; get option type
	sta	opttyp
	txy
	iny
	clc
	rts
	
?135:	lda	prsfg
	bpl	?127
	jsr	?128
	SCNPRINT
	.DB	'missing option.', 13, 0
	sec
	rts

?136:	lda	prsfg
	bpl	?127
	jsr	?128
	lda	chopt
	pha
	phk
	pea	#!?138
	lda	#5
	pha
	BPRINTF
	sec
	rts

?138:	.DB	'illegal option ''-%c''.', 13, 0

?140:	ldx	opttyp		; option type
	lda	[cmdlin],y	; get char after option
	beq	?142		; no argument
	cmp	#' '
	beq	?142		; no argument
	cpx	#' '		; option requires an argument?
	beq	?144		; no...so error
	jmp	?110		; get argument
?142:	cpx	#' '		; option requires an argument?
	bne	?148		; yes...so error
	jmp	?119		; bump command line ptr
	
?144:	lda	prsfg
	bpl	?145
	jsr	?128
	lda	chopt
	pha
	phk
	pea	#!?146
?144a:	lda	#5
	pha
	BPRINTF
?145:	sec
	rts

?146:	.DB	'option ''-%c'' does not requires an argument.', 13, 0

?148:	lda	prsfg
	bpl	?145
	jsr	?128
	lda	chopt
	pha
	phk
	pea	#!?149
	bra	?144a

?149:	.DB	'option ''-%c'' requires an argument.', 13, 0

?150:	lda	slen		; check single char argument
	cmp	#1
	bne	?152
	lda	[dwres]
	sta	dwres
	stz	dwres+1
	stz	dwres+2
	stz	dwres+3
	clc
	rts
?152:	lda	prsfg
	bpl	?154
	jsr	?128
	lda	chopt
	pha
	phk
	pea	#!?156
	lda	#5
	pha
	BPRINTF
?154:	sec
	rts

?156:	.DB	'option ''-%c'' requires a single character argument.', 13, 0

?160:	ldx	#3		; check others option's type
?162:	lda	>?200,x
	cmp	opttyp
	beq	?166
	dex
	bpl	?162
	lda	prsfg
	bpl	?164
	jsr	?128
	lda	opttyp
	pha
	lda	chopt
	pha
	phk
	pea	#!?165
	lda	#6
	pha
	BPRINTF
?164:	sec
	rts

?165:	.DB	'option ''-%c'': invalid argument type ''%c''.', 13, 0

?166:	inx			; wanted integer size
	lda	prsfg
	tay
	and	#$40		; mask autodetect flag
	beq	?166a		; hex string only
	sta	tmpp
	tya
	asl	a
	asl	a		; wanted signed flag
	ora	tmpp	
?166a:	ldy	dwres+2		; string ptr
	phy
	pei	(dwres)
	phx			; integer size
	pha			; flag
	pea	#0		; pointer to result
	lda	#intarg
	pha
	jsl	_lstrtoint
	bcs	?168
	bvs	?180
	ACC16
	sta	tmpp
	stx	tmpp+2
	sty	argsgn
	lda	intarg
	sta	dwres
	lda	intarg+2
	sta	dwres+2
	ACC08
	lda	[tmpp]		; check residual string
	bne	?167		; should be null	
	rts
?167:	clv			; invalid numeric string	
?168:	lda	prsfg
	bpl	?184
	php
	jsr	?128
	plp
	lda	chopt
	pha
	phk
	bvs	?170
	pea	#!?190
	bra	?182
?170:	pea	#!?192		
	bra	?182
?180:	lda	prsfg
	bpl	?184
	jsr	?128
	lda	chopt
	pha
	phk
	pea	#!?194
?182:	lda	#5
	pha
	BPRINTF
?184:	sec
	rts

?190:	.DB	'option ''-%c'': invalid numeric string argument.', 13, 0
?192:	.DB	'option ''-%c'': numeric argument overflow.', 13, 0
?194:	.DB	'option ''-%c'': signed/unsigned mismatch argument.', 13, 0

?200:	.DB	'%#*&'
